unit OpenGLContext;
{ OpenGLContext -    KOLOGL12 (OpenGL  1.2)

  by Vladimir Kladov, 2006
  version: 0.6
  Stage: Beta (tested mostly)

}

interface

//{$DEFINE GLCONTEXT_MULTITHREADS}
//{$DEFINE SUPPORT_JPEG}
//{$DEFINE SUPPORT_PNG}

//{$DEFINE FAST_GLFONT}
//{$DEFINE ITALIC_SMOOTH}

{$IFNDEF NOT_CHECKING_STATE}
  {$DEFINE CHECKING_STATE}
{$ENDIF}

{$IFNDEF NO_MULTITEXTURE}
  {$DEFINE MULTITEXTURING}
{$ENDIF}

uses Windows, KOL, KOLOGL12 {$IFDEF SUPPORT_JPEG}, JpegObj{$ENDIF}
                            {$IFDEF SUPPORT_PNG}, KOLPng{$ENDIF};

const
  MinFontHeight: Integer = 6;
  MaxFontHeight: Integer = 40;

type TColor4 = packed record
       r, g, b, a: Single;
     end;
     PColor4 = ^TColor4;

const Color0: TColor4 = ( r:0; g:0; b:0; a:0 );

function Color( r, g, b, a: Single ): TColor4; overload;
function Color4( r, g, b, a: Single ): TColor4; overload;
function Color( r, g, b: Single ): TColor4; overload;
function Color4( r, g, b: Single ): TColor4; overload;
function Color( const Q: TRGBQuad ): TColor4; overload;
function Color4( const Q: TRGBQuad ): TColor4; overload;
function Color( C: TColor ): TColor4; overload;
function Color4( C: TColor ): TColor4; overload;
function ColorEq( C1, C2: TColor4 ): Boolean;

function RGBA( r, g, b, a: Byte ): TRGBQuad; overload;
function RGBA( C: TColor4 ): TRGBQuad; overload;
function RGB( r, g, b: Byte ): TRGBQuad;

type TPoint4 = packed record
       x, y, z, w: Single;
     end; PPoint4 = ^TPoint4;

     TPoint4D = packed record
       X, Y, Z, W: Double;
     end; PPoint4D = ^TPoint4D;

     TPoint3 = packed record
       x, y, z: Single;
     end; PPoint3 = ^TPoint3;

     TPoint3D = packed record
       X, Y, Z: Double;
     end;

     TPoint3E = packed record
       X, Y, Z: Extended;
     end;

     TPoint2 = packed record
       x, y: Single;
     end; PPoint2 = ^TPoint2;

     TPoint2D = packed record
       X, Y: Double;
     end;

function Point4( x, y, z, w: Single ): TPoint4; overload;
function Point4( x, y, z: Single ): TPoint4; overload;
function Point4D( x, y, z, w: Double ): TPoint4D; overload;
function Point4D( x, y, z: Double ): TPoint4D; overload;
function Point4D( const p3: TPoint3D ): TPoint4D; overload;

function Point3( x, y, z: Single ): TPoint3; overload;
function Point3D( X, Y, Z: Double ): TPoint3D; overload;
function Point3( p4: TPoint4 ): TPoint3; overload;
function Point3( p3D: TPoint3D ): TPoint3; overload;
function Point3D( p4d: TPoint4D ): TPoint3D; overload;
function Point3D( p4: TPoint4 ): TPoint3D; overload;
function Point3D( p3: TPoint3): TPoint3D; overload;
function Point3D( p3: TVector3d): TPoint3D; overload;

procedure Normalize( var P: TPoint3D ); overload;
procedure Normalize( var P: TPoint3 ); overload;

function Point2( x, y: Single ): TPoint2; overload;
function Point2( P2D: TPoint2D ): TPoint2; overload;
function Point2D( X, Y: Double ): TPoint2D;

type TVector3 = TPoint3;

function Vector3( x, y, z: Single ): TVector3;
function Vector3d( x, y, z: Double ): TVector3d; overload;
function Vector3d( P: TPoint3D ): TVector3d; overload;
function Vector4d( x, y, z, w: Double ): TVector4d;
procedure RotateVector3( var T: TPoint3; angle: Single; axis: TPoint3 );

function Dist3D( const P1, P2: TPoint3D ): Double;
function Dist3sqrt( const P1, P2: TPoint3 ): Single;
function VectorMult3D( A, B: TPoint3D ): TPoint3D;
function VectorMult3( A, B: TPoint3 ): TPoint3;

procedure Transpose4D( var M: TMatrix4d );
function MultMatrix4D_by_Matrix4D( const A, B: TMatrix4d ): TMatrix4d;
function MultMatrix4D_by_Vector4D( const A: TMatrix4d; V: TVector4d ): TVector4d;
function MultMatrix4D_by_Vector3D( const A: TMatrix4d; V: TVector3d ): TVector3d;
function Determinant4D( const A: TMatrix4d ): Double;
function Inverse4D( const A: TMatrix4d ): TMatrix4d;

function Rotation4D( angle: Double; const axis: TVector3d ): TMatrix4d;

type
  TSide = ( sFront, sBack, sFrontAndBack );

  TBlendFunc = ( bfZero, bfOne, bfColor, bfOneMinusColor,
    bfSrcAlpha, bfOneMinusSrcAlpha, bfDstAlpha, bfOneMinusDstAlpha,
    bfSrcAlphaSaturate );

  TCullFace = ( cfNone, cfBack, cfFront );

  TFogMode = ( fmExp, fmExp2, fmLinear );

  TLogicOp = ( loDisabled, loClear, loSet, loCopy, loCopyInverted, loNoop,
    loInvert, loAnd, loNAnd, loOr, loNOr, loXOr, loEQuiv, loAndReverse,
    loAndInverted, loOrReverse, loOrInverted );
  TStencilOp = ( soKeep, soZero, soReplace, soInc, soDec, soInvert );
  TDepthFunc = ( dfLt, dfNo, dfEq, dfNE, dfLE, dfGT, dfGE, dfAlways );
  TAlphaFunc = ( afLt, afNo, afEq, afNE, afLE, afGT, afGE, afAlways );
  TStencilFunc = ( sfNever, sfLT, sfLE, sfGT, sfGE, sfEQ, sfNE, sfAlways );

  TPolygonMode = ( pmFill, pmLines, pmPoints );
  THintQuality = ( hqDontCare, hqFastest, hqNicest );

  TDrawMode = ( dmPoints, dmLineStrip, dmLineLoop, dmLines, dmTriangleStrip,
    dmTrianleFan, dmTriangles, dmQuadStrip, dmQuads, dmPolygon );

  TTextureFormat = ( tfRGBA, tfRGB );
  TTextureEnvironment = ( teModulate, teReplace, teDecal, teBlend );
  TTextureMinFilter = ( tminNearest, tminLinear, tminNearestMipmapNearest,
    tminLinearMipmapNearest, tminNearestMipmapLinear, tminLinearMipmapLinear );
  TTextureMagFilter = ( tmagNearest, tmagLinear );
  TTextureWrap = ( twRepeat, twClamp );

  TFlatFontStyles = ( ffsBold, ffsItalic, ffsAntialiased );
  TFlatFontStyle = Set of TFlatFontStyles;

  TElementType = ( etShort, etInt, etSingle, etDouble );
  TColorEltType = ( cetByte, cetUByte, cetShort, cetUShort, cetInt, cetUInt,
    cetSingle, cetDouble );
  TArrayType = ( atVertexes, atNormals, atColors, atTexCoords, atTex1Coords,
                 atTex2Coords, atTex3Coords, atTex4Coords, atEdgeFlags, atCombined );
  TCombinedElements = Set of TArrayType;

const
  TextureFirstLoadFromResourve: Boolean = TRUE;
  MaxFontCount: Integer = 10;

type
  PMaterial = ^TMaterial;
  PLightSource = ^TLightSource;
  PFog = ^TFog;
  PTexture = ^TTexture;
  PFontController = ^TFontController;
  PArray = ^TArray;
  PColorArray = ^TColorArray;
  PCombinedArray = ^TCombinedArray;
  PDisplayList = ^TDisplayList;

  //----------------------------------------------------------------------------
  // .    ,  (  )
  //  , .
  PContext = ^TContext;
  TContext = object( TObj )
  private
    FMaterial: PMaterial;
    FColor: TColor4;
    FAlphaTest: Boolean;
    FDepthTest: Boolean;
    FBlend: Boolean;
    FBlendFuncSrc: TBlendFunc;
    FBlendFuncDst: TBlendFunc;
    FAutoNormal: Boolean;
    FLighting: Boolean;
    FCullFace: TCullFace;
    FDither: Boolean;
    FFog: PFog;
    FLightModelAmbient: TColor4;
    FLightModelLocalViewer: Single;
    FLightModelTwoSide: Single;
    FLineSmooth: Boolean;
    FLineWidth: Single;
    FLineStipple: Boolean;
    FStipplePattern: Word;
    FStippleFactor: Integer;
    FLogicOp: TLogicOp;
    FNormalize: Boolean;
    FPointSmooth: Boolean;
    FPointSize: Single;
    FPolygonSmooth: Boolean;
    FPolygonMode: array[ TSide ] of TPolygonMode;
    FCameraCoords: TPoint3D;
    FCameraZDirection: TPoint3D;
    FCameraYDirection: TPoint3D;
    FTexture: PTexture;
    fTexPool: PList;
    FFontController: PFontController;
    FCaretPos: TPoint;
    fNextLineY: Integer;
    FEnableVertexArray: Boolean;
    FEnableNormalArray: Boolean;
    FEnableColorArray: Boolean;
    FEnableEdgeFlagArray: Boolean;
    FEnableTexCoordArray: array[0..4] of Boolean;
    FLineQuality: THintQuality;
    FPerspectiveQuality: THintQuality;
    FPointQuality: THintQuality;
    FPolygonQuality: THintQuality;
    FChgColor, FChgAlphaTest, FChgDepthTest, FChgBlend, FChgBlendFunc,
    FChgAutoNormal, FChgLighting, FChgCullFace, FChgDither,
    FChgLightModelAmbient, FChgLightModelLocalViewer, FChgLightModelTwoSide,
    FChgLineSmooth, FChgLineQuality, FChgLineWidth, FChgLineStipple, FChgStipplePattern,
    FChgStippleFactor, FChgLogicOp, FChgNormalize, FChgPointSmooth, FChgPointQuality,
    FChgPointSize, FChgPolygonSmooth, FChgPolygonMode,
    FChgPolygonQuality, FChgPolygonStipple, FChgCamera,
    FChgScissorTest
    : Boolean;
    FSmoothShade: Boolean;
    FFrontFaceCW: Boolean;
    FDepthWrite: Boolean;
    FDepthRange: TPoint2D;
    FDepthFunc: TDepthFunc;
    fAutoActivateTexture: Boolean;
    fShadowColor: TColor4;
    fShadowOffset: TPoint;
    fDoubleShadow: Boolean;
    FAlphaFunc: TAlphaFunc;
    FAlphaTestValue: Single;
    FStencilOpFail: TStencilOp;
    FStencilOpPassWODepth: TStencilOp;
    FStencilOpPass: TStencilOp;
    FStencilTest: Boolean;
    FStencilFunc: TStencilFunc;
    FStencilRef: Integer;
    FStencilMask: Integer;
    procedure SetColor(const Value: TColor4);
    procedure SetAlphaTest(const Value: Boolean);
    procedure SetDepthTest(const Value: Boolean);
    procedure SetBlend(const Value: Boolean);
    procedure SetBlendFuncDst(const Value: TBlendFunc);
    procedure SetBlendFuncSrc(const Value: TBlendFunc);
    procedure SetAutoNormal(const Value: Boolean);
    procedure SetLighting(const Value: Boolean);
    function GetLight(idx: Integer): PLightSource;
    procedure SetCullFace(const Value: TCullFace);
    procedure SetDither(const Value: Boolean);
    procedure SetLightModelAmbient(const Value: TColor4);
    procedure SetLightModelLocalViewer(const Value: Single);
    procedure SetLightModelTwoSide(const Value: Single);
    procedure SetLineSmooth(const Value: Boolean);
    procedure SetLineWidth(const Value: Single);
    procedure SetLineStipple(const Value: Boolean);
    procedure SetStipplePattern(const Value: Word);
    procedure SetStippleFactor(const Value: Integer);
    procedure SetLogicOp(const Value: TLogicOp);
    procedure SetNormalize(const Value: Boolean);
    procedure SetPointSmooth(const Value: Boolean);
    procedure SetPointSize(const Value: Single);
    procedure SetPolygonSmooth(const Value: Boolean);
    procedure SetPolygonMode(side: TSide; const Value: TPolygonMode);
    function GetPolygonMode(side: TSide): TPolygonMode;
    procedure SetCameraCoords(const Value: TPoint3D);
    procedure SetCameraZDirection(const Value: TPoint3D);
    procedure SetCameraYDirection(const Value: TPoint3D);
    function GetMaxTextureSize: Integer;
    procedure SetTexture(const Value: PTexture);
    function GetCaretX: Integer;
    function GetCaretY: Integer;
    procedure SetCaretPos(const Value: TPoint);
    procedure SetCaretX(const Value: Integer);
    procedure SetCaretY(const Value: Integer);
    procedure SetEnableVertexArray(const Value: Boolean);
    procedure SetEnableNormalArray(const Value: Boolean);
    procedure SetEnableColorArray(const Value: Boolean);
    procedure SetEnableEdgeFlagArray(const Value: Boolean);
    procedure SetEnableTexCoordArray(idx: Integer; const Value: Boolean);
    procedure SetViewPort(const Value: TRect);
    procedure SetMaterial(const Value: PMaterial);
    procedure SetFog(const Value: PFog);
    procedure SetLight(idx: Integer; const Value: PLightSource);
    procedure SetFont(const Value: PFontController);
    function GetClipPlane(idx: Integer): TPoint4D;
    function GetMaxClipPlanes: Integer;
    procedure SetClipPlane(idx: Integer; const Value: TPoint4D);
    function GetPointGranularity: Double;
    function GetPointMax: Double;
    function GetPointMin: Double;
    procedure SetLineQuality(const Value: THintQuality);
    procedure SetPerspectiveQuality(const Value: THintQuality);
    procedure SetPointQuality(const Value: THintQuality);
    procedure SetPolygonQuality(const Value: THintQuality);
    function GetViewportHeight: Integer;
    function GetViewportWidth: Integer;
    function GetMaxLights: Integer;
    procedure SetSmoothShade(const Value: Boolean);
    procedure SetFrontFaceCW(const Value: Boolean);
    procedure SetDepthRange(const Value: TPoint2D);
    procedure SetDepthWrite(const Value: Boolean);
    procedure SetDepthFunc(const Value: TDepthFunc);
    function GetEnableTexCoordArray(idx: Integer): Boolean;
    procedure SetAlphaFunc(const Value: TAlphaFunc);
    procedure SetAlphaTestValue(const Value: Single);
    procedure SetStencilOpFail(const Value: TStencilOp);
    procedure SetStencilOpPass(const Value: TStencilOp);
    procedure SetStencilOpPassWODepth(const Value: TStencilOp);
    procedure SetStencilTest(const Value: Boolean);
    procedure SetStencilFunc(const Value: TStencilFunc);
    procedure SetStencilRef(const Value: Integer);
    procedure SetStencilMask(const Value: Integer);
  protected
    FLights: array[ 0..7 ] of PLightSource;
    FPolygonStipple: array[ 0..31 ] of DWORD;
    FScissorTest: TRect;
    FCtxWnd: HWnd;
    FCtxBitmap: PBitmap;
    FCtxDC: HDC;
    FCtxRC: THandle;
    FViewRect: TRect;
    FFov, FzFar, FzNear, FzNearOrtho, FzFarOrtho: Double;
    FOrthoRect: TPoint4;
    FLastPerspective: Boolean;
    FMaxTexSize: Integer;
    fTextureEnvironment: TTextureEnvironment;
    fTextureColor: TColor4;
    FTextureBorderColor: TColor4;
    FTextureMinFilter: TTextureMinFilter;
    FTextureMagFilter: TTextureMagFilter;
    FTextureWrapS, FTextureWrapT: TTextureWrap;
    {$IFDEF GLCONTEXT_MULTITHREADS}
    fTextures: PList;
    {$ENDIF}
    fFontList: PList;
    FSelectedFont: PFontController;
    fIndices16: PWord; fI16count: Integer;
    fIndices32: PDWORD;  fI32count: Integer;
    fCurVrtxArray, fCurNormArray, fCurEdgeArray: PArray;
    fCurTexCArray: array[ 0..4 ] of PArray;
    fCurColrArray: PArray;
    fMaxClipPlanes: Integer;
    fClipPlanes: array of TPoint4D;
    fPointGran: Double;
    fPointSizes: array[ 0..1 ] of Double;
    fMaxLights: Integer;
    fColorMask: TRGBQuad;
    procedure INIT; virtual;
    function ProvideIndices( count: Integer ): Pointer;
    function SetBit( var Bit: Boolean ): Boolean;
  public
    destructor DESTROY; virtual;
    property RC: THandle read FCtxRC;
    property DC: HDC read FCtxDC;
    property Wnd: HWnd read FCtxWnd;
    procedure MakeCurrent;
    function IsCurrent: Boolean;
    procedure BeginScene;
    procedure EndScene;
    procedure Clear( clear_ColorBuffer, clear_DepthBuffer: Boolean;
                     clear_AccumBuffer: Boolean = FALSE;
                     clear_StencilBuffer: Boolean = FALSE );
    // 
    property AlphaFunc: TAlphaFunc read FAlphaFunc write SetAlphaFunc;
    property AlphaTest: Boolean read FAlphaTest write SetAlphaTest;
    property AlphaTestValue: Single read FAlphaTestValue write SetAlphaTestValue;
    property DepthFunc: TDepthFunc read FDepthFunc write SetDepthFunc;
    property DepthTest: Boolean read FDepthTest write SetDepthTest;
    property DepthWrite: Boolean read FDepthWrite write SetDepthWrite;
    property DepthRange: TPoint2D read FDepthRange write SetDepthRange;
    property StencilOpFail: TStencilOp read FStencilOpFail write SetStencilOpFail;
    property StencilOpPassWODepth: TStencilOp read FStencilOpPassWODepth write SetStencilOpPassWODepth;
    property StencilOpPass: TStencilOp read FStencilOpPass write SetStencilOpPass;
    property StencilTest: Boolean read FStencilTest write SetStencilTest;
    property StencilFunc: TStencilFunc read FStencilFunc write SetStencilFunc;
    property StencilRef: Integer read FStencilRef write SetStencilRef;
    property StencilMask: Integer read FStencilMask write SetStencilMask;
    property Blend: Boolean read FBlend write SetBlend;
    property AutoNormal: Boolean read FAutoNormal write SetAutoNormal;
    property Normalize: Boolean read FNormalize write SetNormalize;
    property Dither: Boolean read FDither write SetDither;
    property LogicOp: TLogicOp read FLogicOp write SetLogicOp;
    // 
    property FrontFaceCW: Boolean read FFrontFaceCW write SetFrontFaceCW;
    property CullFace: TCullFace read FCullFace write SetCullFace;
    procedure ScissorTest( Rect: PRect );
    property MaxClipPlanes: Integer read GetMaxClipPlanes;
    property ClipPlane[ idx: Integer ]: TPoint4D read GetClipPlane write SetClipPlane;
    // , 
    property Material: PMaterial read FMaterial write SetMaterial;
    property Color: TColor4 read FColor write SetColor;
    procedure ColorMask( mr, mg, mb, ma: Boolean );
    // 
    property PointSmooth: Boolean read FPointSmooth write SetPointSmooth;
    property PointQuality: THintQuality read FPointQuality write SetPointQuality;
    property PointSize: Single read FPointSize write SetPointSize;
    property PointMin: Double read GetPointMin;
    property PointMax: Double read GetPointMax;
    property PointGranularity: Double read GetPointGranularity;
    // 
    property LineSmooth: Boolean read FLineSmooth write SetLineSmooth;
    property LineQuality: THintQuality read FLineQuality write SetLineQuality;
    property LineWidth: Single read FLineWidth write SetLineWidth;
    property LineStipple: Boolean read FLineStipple write SetLineStipple;
    property StipplePattern: Word read FStipplePattern write SetStipplePattern;
    property StippleFactor: Integer read FStippleFactor write SetStippleFactor;
    // 
    property PolygonSmooth: Boolean read FPolygonSmooth write SetPolygonSmooth;
    property PolygonQuality: THintQuality read FPolygonQuality write SetPolygonQuality;
    property PolygonMode[ side: TSide ]: TPolygonMode read GetPolygonMode write SetPolygonMode;
    procedure PolygonStipple( Pattern: Pointer );
    // 
    property Fog: PFog read FFog write SetFog;
    // 
    property BlendFuncSrc: TBlendFunc read FBlendFuncSrc write SetBlendFuncSrc;
    property BlendFuncDst: TBlendFunc read FBlendFuncDst write SetBlendFuncDst;
    procedure BlendTransparent;
    procedure BlendOver;
    // 
    property Lighting: Boolean read FLighting write SetLighting;
    property LightModelAmbient: TColor4 read FLightModelAmbient write SetLightModelAmbient;
    property LightModelLocalViewer: Single read FLightModelLocalViewer write SetLightModelLocalViewer;
    property LightModelTwoSide: Single read FLightModelTwoSide write SetLightModelTwoSide;
    property Light[ idx: Integer ]: PLightSource read GetLight write SetLight;
    property MaxLights: Integer read GetMaxLights;
    property SmoothShade: Boolean read FSmoothShade write SetSmoothShade;
    //    
    property ViewPort: TRect read FViewRect write SetViewPort;
    property Height: Integer read GetViewportHeight;
    property Width: Integer read GetViewportWidth;
    procedure Perspective( newFOV, z_Near, z_Far: Double ); overload;
    procedure Perspective; overload;
    property FOV: Double read FFOV; // 
    property zNear: Double read FzNear;
    property zFar: Double read FzFar;
    procedure Ortho( left, right, bottom, top, z_near, z_far: Single ); overload;
    procedure Ortho( WithViewRect: Boolean = TRUE ); overload;
    property PerspectiveQuality: THintQuality read FPerspectiveQuality write SetPerspectiveQuality;
    property IsPerspective: Boolean read FLastPerspective;
    procedure PushProjection;
    procedure PopProjection;
    procedure PushModelview;
    procedure PopModelview;
    procedure PushTexture;
    procedure PopTexture;
    // 
    property CameraCoords: TPoint3D read FCameraCoords write SetCameraCoords;
    property CameraZDirection: TPoint3D read FCameraZDirection write SetCameraZDirection;
    property CameraYDirection: TPoint3D read FCameraYDirection write SetCameraYDirection;
    procedure LookFromCamera;
    // 
    function NewTexture( W, H, D: Integer; Fmt: TTextureFormat = tfRGBA ): PTexture;
    property MaxTextureSize: Integer read GetMaxTextureSize;
    property Texture: PTexture read FTexture write SetTexture;
    property AutoActivateTexture: Boolean read fAutoActivateTexture write fAutoActivateTexture;
             //   !!!
    //  
    function NewDisplayList: PDisplayList;
    //    
    property Font: PFontController read FFontController write SetFont;
    property CaretPos: TPoint read FCaretPos write SetCaretPos;
    property CaretX: Integer read GetCaretX write SetCaretX;
    property CaretY: Integer read GetCaretY write SetCaretY;
    property NextLineY: Integer read fNextLineY;
    procedure Print( const s: String );
    procedure CRLF;
    property ShadowColor: TColor4 read fShadowColor write fShadowColor;
    property ShadowOffset: TPoint read fShadowOffset write fShadowOffset;
    property DoubleShadow: Boolean read fDoubleShadow write fDoubleShadow;
    function ScaleFont( BaseFontHeight: Integer; BaseWindowHeight: Integer = 480 ):
      Integer; // :   BaseFontHeight    BaseWindowHeight,
               //        
    function ScaleD( BaseHeight: Double; BaseWindowHeight: Integer = 480 ): Double;
    //  , , 
    function NewVertexArray( etype: TElementType; ncoords: Integer; count: Integer;
      initptr: Pointer; shareptr: Boolean = FALSE ): PArray;
    function NewTexCoordArray( etype: TElementType; ncoords: Integer; count: Integer;
      initptr: Pointer; shareptr: Boolean = FALSE ): PArray;
    function NewNormalArray( eType: TElementType; count: Integer;
      initptr: Pointer; shareptr: Boolean = FALSE ): PArray;
    function NewColorArray( eType: TColorEltType; ncoords: Integer; count: Integer;
      initptr: Pointer; shareptr: Boolean = FALSE ): PColorArray;
    function NewEdgeFlagArray( count: Integer; initptr: Pointer;
      shareptr: Boolean = FALSE ): PArray;
    function NewCombinedArray( elements: TCombinedElements; ncoord_vertex,
      ncoord_textures, ncomp_color, count: Integer; initptr: Pointer; shareptr: Boolean = FALSE )
      : PCombinedArray;
    procedure DrawElements( DrawStyle: TDrawMode;
      VrtxArray, TexCArray, NormArray: PArray; ColrArray: PArray; EdgeArray: PArray;
      indices: Pointer = nil; indcount: Integer = 0; DWORD_indices: Boolean = FALSE );
      overload;
    procedure DrawElements( DrawStyle: TDrawMode; CombArray: PCombinedArray;
      indices: Pointer = nil; indcount: Integer = 0; DWORD_indices: Boolean = FALSE );
      overload;
    property EnableVertexArray: Boolean read FEnableVertexArray write SetEnableVertexArray;
    property EnableNormalArray: Boolean read FEnableNormalArray write SetEnableNormalArray;
    property EnableColorArray: Boolean read FEnableColorArray write SetEnableColorArray;
    property EnableEdgeFlagArray: Boolean read FEnableEdgeFlagArray write SetEnableEdgeFlagArray;
    property EnableTexCoordArray[ idx: Integer ]: Boolean read GetEnableTexCoordArray write SetEnableTexCoordArray;
    procedure DisableAllArrays;
  public //  
    function DebugState: String;
    procedure Dump( const filname: String );
  end;

  //----------------------------------------------------------------------------
  // 
  TColorMaterial = ( cmAmbientAndDiffuse, cmAmbient, cmDiffuse, cmSpecular, cmEmission );

  TMaterial = object( TObj )
  private
    FColorEnabled: Boolean;
    function GetAmbient(side: TSide): TColor4;
    function GetColorEffect(side: TSide): TColorMaterial;
    function GetDiffuse(side: TSide): TColor4;
    function GetEmission(side: TSide): TColor4;
    function GetShininess(side: TSide): Single;
    function GetSpecular(side: TSide): TColor4;
    procedure SetAmbient(side: TSide; const Value: TColor4);
    procedure SetDiffuse(side: TSide; const Value: TColor4);
    procedure SetAmbientAndDiffuse(side: TSide; const Value: TColor4);
    procedure SetColorEffect(side: TSide; const Value: TColorMaterial);
    procedure SetColorEnabled(const Value: Boolean);
    procedure SetEmission(side: TSide; const Value: TColor4);
    procedure SetShininess(side: TSide; const Value: Single);
    procedure SetSpecular(side: TSide; const Value: TColor4);
  protected
    fContext: PContext;
    FColorEffect: array[ TSide ] of TColorMaterial;
    FAmbient: array[ TSide ] of TColor4;
    FDiffuse: array[ TSide ] of TColor4;
    FSpecular: array[ TSide ] of TColor4;
    FEmission: array[ TSide ] of TColor4;
    FShininess: array[ TSide ] of Single;
    procedure INIT; virtual;
  public
    property ColorEffect[ side: TSide ]: TColorMaterial read GetColorEffect write SetColorEffect;
    property ColorEnabled: Boolean read FColorEnabled write SetColorEnabled;
    property Diffuse[ side: TSide ]: TColor4 read GetDiffuse write SetDiffuse;
    property Ambient[ side: TSide ]: TColor4 read GetAmbient write SetAmbient;
    property AmbientAndDiffuse[ side: TSide ]: TColor4 write SetAmbientAndDiffuse;
    property Specular[ side: TSide ]: TColor4 read GetSpecular write SetSpecular;
    property Emission[ side: TSide ]: TColor4 read GetEmission write SetEmission;
    property Shininess[ side: TSide ]: Single read GetShininess write SetShininess;
    procedure Assign( SrcMaterail: PMaterial );
    function DebugState( const prefx: String ): String;
  end;

  //----------------------------------------------------------------------------
  //  
  TLightSource = object( TObj )
  private
    FEnabled: Boolean;
    index: Integer;
    FSpotExponent: Integer;
    FSpotCutOff: Single;
    FConstantAttenuation: Single;
    FLinearAttenuation: Single;
    FQuadraticAttenuation: Single;
    FAmbient: TColor4;
    FDiffuse: TColor4;
    FSpecular: TColor4;
    FPosition: TPoint4;
    FSpotDirection: TPoint3D;
    procedure SetEnabled(const Value: Boolean);
    procedure SetSpotExponent(const Value: Integer);
    procedure SetSpotCutOff(const Value: Single);
    procedure SetConstantAttenuation(const Value: Single);
    procedure SetLinearAttenuation(const Value: Single);
    procedure SetQuadraticAttenuation(const Value: Single);
    procedure SetAmbient(const Value: TColor4);
    procedure SetDiffuse(const Value: TColor4);
    procedure SetSpecular(const Value: TColor4);
    procedure SetPosition(const Value: TPoint4);
    function GetDirectional: Boolean;
    procedure SetDirectional(const Value: Boolean);
    procedure SetSpotDirection(const Value: TPoint3D);
  protected
    fContext: PContext;
    procedure INIT; virtual;
  public
    property Enabled: Boolean read FEnabled write SetEnabled;
    property SpotExponent: Integer read FSpotExponent write SetSpotExponent;
    property SpotCutOff: Single read FSpotCutOff write SetSpotCutOff;
    property ConstantAttenuation: Single read FConstantAttenuation write SetConstantAttenuation;
    property LinearAttenuation: Single read FLinearAttenuation write SetLinearAttenuation;
    property QuadraticAttenuation: Single read FQuadraticAttenuation write SetQuadraticAttenuation;
    property Ambient: TColor4 read FAmbient write SetAmbient;
    property Diffuse: TColor4 read FDiffuse write SetDiffuse;
    property Specular: TColor4 read FSpecular write SetSpecular;
    property Position: TPoint4 read FPosition write SetPosition;
    property Directional: Boolean read GetDirectional write SetDirectional;
    property SpotDirection: TPoint3D read FSpotDirection write SetSpotDirection;
    procedure Assign( SrcLight: PLightSource );
    function DebugState( const prefx: String ): String;
  end;

  //----------------------------------------------------------------------------
  // 
  TFog = object( TObj )
  private
    F_End: Single;
    FIndex: Single;
    FDensity: Single;
    FStart: Single;
    FColor: TColor4;
    FMode: TFogMode;
    FEnabled: Boolean;
    FQuality: THintQuality;
    procedure Set_End(const Value: Single);
    procedure SetColor(const Value: TColor4);
    procedure SetDensity(const Value: Single);
    procedure SetIndex(const Value: Single);
    procedure SetMode(const Value: TFogMode);
    procedure SetStart(const Value: Single);
    procedure SetEnabled(const Value: Boolean);
    procedure SetQuality(const Value: THintQuality);
  protected
    fContext: PContext;
    procedure INIT; virtual;
  public
    property Enabled: Boolean read FEnabled write SetEnabled;
    property Mode: TFogMode read FMode write SetMode;
    property Density: Single read FDensity write SetDensity;
    property Start: Single read FStart write SetStart;
    property _End: Single read F_End write Set_End;
    property Index: Single read FIndex write SetIndex;
    property Color: TColor4 read FColor write SetColor;
    property Quality: THintQuality read FQuality write SetQuality;
    procedure Assign( SrcFog: PFog );
    function DebugState( const prefx: String ): String;
  end;

  //----------------------------------------------------------------------------
  // 
  TCalculateAlpha = ( caSetConst, caNotBlack, caAsIs );
  TOnBitmap = procedure( Sender: PObj; B: PBitmap ) of object;

  TTexture = object( TObj )
  private
    fWidth: Integer;
    fHeight: Integer;
    fDepth: Integer;
    FFormat: TTextureFormat;
    FEnvironment: TTextureEnvironment;
    FColor: TColor4;
    FMinFilter: TTextureMinFilter;
    FMagFilter: TTextureMagFilter;
    FWrapS: TTextureWrap;
    FWrapT: TTextureWrap;
    fAlpha: Single;
    fCalcAlpha: TCalculateAlpha;
    FBorderColor: TColor4;
    fOnBitmap: TOnBitmap;
    fCalcAvgColor: Boolean;
    fAvgColor: TColor4;
    fDimension: Integer;
    FMipMap: Boolean;
    function GetDimension: Integer;
    procedure SetEnvironment(const Value: TTextureEnvironment);
    procedure SetColor(const Value: TColor4);
    procedure SetBorderColor(const Value: TColor4);
    procedure SetMinFilter(const Value: TTextureMinFilter);
    procedure SetMagFilter(const Value: TTextureMagFilter);
    procedure SetWrapS(const Value: TTextureWrap);
    procedure SetWrapT(const Value: TTextureWrap);
    procedure SetMipMap(const Value: Boolean);
  protected
    {$IFDEF GLCONTEXT_MULTITHREADS}
    fImage: PBitmap;
    fContexts: PList;
    fTexIDs: PList;
    {$ELSE}
    fContext: PContext;
    fTexID: Integer;
    {$ENDIF}
    procedure INIT; virtual;
    procedure Bind;
    procedure ApplyParams;
    procedure UnBind;
  public
    destructor DESTROY; virtual;
    {$IFDEF GLCONTEXT_MULTITHREADS}
    {$ELSE}
    property Context: PContext read fContext;
    property TextureHandle: Integer read fTexID;
    {$ENDIF}
    procedure MoveToPool;
    property Width: Integer read fWidth;
    property Height: Integer read fHeight;
    property Depth: Integer read fDepth;
    property Dimension: Integer read GetDimension;
    property Format: TTextureFormat read FFormat;
    function Binded: Boolean; overload;
    function Binded( toContext: PContext ): Boolean; overload;
    //  
    property OnBitmap: TOnBitmap read fOnBitmap write fOnBitmap;
    property CalculateAlpha: TCalculateAlpha read fCalcAlpha write fCalcAlpha;
    property CalculateAverageColor: Boolean read fCalcAvgColor write fCalcAvgColor;
    property AvgColor: TColor4 read fAvgColor;
    property Alpha: Single read fAlpha write fAlpha;
    function Load( const rsrcname: String ): Boolean;
    function LoadFromStream( Strm: PStream; SwapRB: Boolean = TRUE ): Boolean;
    function LoadFromStreamMipMap( level: Integer; Strm: PStream; SwapRB: Boolean = TRUE ): Boolean;
    function LoadFromFile( const filename: String ): Boolean;
    function LoadFromFileMipMap( level: Integer; const filename: String ): Boolean;
    function LoadFromResourceName( const rsrcname: String ): Boolean;
    function LoadFromResourceNameMipMap( level: Integer; const rsrcname: String ): Boolean;
    function LoadFromResourceID( id: Integer ): Boolean;
    function LoadFromBitmap( B: PBitmap; SwapRB: Boolean = TRUE ): Boolean;
    function LoadFromBitmapPartially( B: PBitmap; y0, y1, z: Integer;
             SwapRB: Boolean = TRUE; level: Integer = 0 ): Boolean;
    function LoadFromBitmapMipMapLevel( level: Integer; B: PBitmap; SwapRB: Boolean = TRUE ): Boolean;
    function LoadMipMapLevel( level: Integer; const rsrcname: String ): Boolean;
    {$IFDEF SUPPORT_JPEG}
    function LoadFromJpeg( J: PJpeg; SwapRB: Boolean = TRUE ): Boolean;
    function LoadFromJpegMipMapLevel( level: Integer; J: PJpeg; SwapRB: Boolean = TRUE ): Boolean;
    {$ENDIF}
    {$IFDEF SUPPORT_PNG}
    function LoadFromPng( P: PPngObject; SwapRB: Boolean = TRUE ): Boolean;
    function LoadFromPngMipMapLevel( level: Integer; P: PPngObject; SwapRB: Boolean = TRUE ): Boolean;
    {$ENDIF}
    //  
    property Environment: TTextureEnvironment read FEnvironment write SetEnvironment;
    property Color: TColor4 read FColor write SetColor; //   teBlend !
    property BorderColor: TColor4 read FBorderColor write SetBorderColor;
    property MinFilter: TTextureMinFilter read FMinFilter write SetMinFilter;
    property MagFilter: TTextureMagFilter read FMagFilter write SetMagFilter;
    property WrapS: TTextureWrap read FWrapS write SetWrapS;
    property WrapT: TTextureWrap read FWrapT write SetWrapT;
    //todo: glTexGen...
    function DebugState( const prefx: String ): String;
    property MipMap: Boolean read FMipMap write SetMipMap;
  end;

  //----------------------------------------------------------------------------
  //   
  TFontController = object( TObj )
  private
    FFace: String;
    FHeight: Integer;
    FStyle: TFlatFontStyle;
    fWidth: Integer;
  protected
    fContext: PContext;
    fFontListID: Integer;
    procedure INIT; virtual;
    procedure SelectFont;
    function Match( F: PFontController ): Boolean;
    procedure NeedFont;
  protected
    {$IFDEF FAST_GLFONT}
    fColor: TColor4;
    fFontTexture: PTexture;
    procedure NeedFastFont( Font, Font2: HFont );
    {$ENDIF}
  public
    destructor DESTROY; virtual;
    property Face: String read FFace write FFace;
    property Height: Integer read FHeight write FHeight;
    property Width: Integer read fWidth write fWidth;
    property Style: TFlatFontStyle read FStyle write FStyle;
    procedure Assign( SrcFont: PFontController );
  end;

  //----------------------------------------------------------------------------
  //  , , ,  
  TArray = object( TObj )
  private
    fNCoords: Integer;
    FCapacity: Integer;
    function GetItems(idx: Integer): TPoint4D;
    procedure SetItems(idx: Integer; const Value: TPoint4D);
    function GetItemPtr(Idx: Integer): Pointer;
    procedure SetCapacity(const Value: Integer);
    procedure SetCount(const Value: Integer);
  protected
    fContext: PContext;
    fArrType: TArrayType;
    fData: Pointer;
    fCount: Integer;
    fShare: Boolean;
    fEltType: TElementType;
    fEltSize: Integer;
    fCombined: Boolean;
    procedure initEltSize;
    procedure SetCurrent( Ctx: PContext ); virtual;
  public
    destructor DESTROY; virtual;
    function DataSize: Integer; virtual;
    property Count: Integer read fCount write SetCount;
    property Capacity: Integer read FCapacity write SetCapacity;
    property ArrayType: TArrayType read fArrType;
    property EltType: TElementType read fEltType;
    property NCoords: Integer read fNCoords;
    property Items[ idx: Integer ]: TPoint4D read GetItems write SetItems;
    property ItemPtr[ Idx: Integer ]: Pointer read GetItemPtr;
    property Data: Pointer read fData;
    procedure Change;
  end;

  TColorArray = object( TArray )
  protected
    fColorEltType: TColorEltType;
  public
    function DataSize: Integer; virtual;
  end;

  TCombinedArray = object( TArray )
  private
    FActiveArrays: TCombinedElements;
    procedure SetActiveArrays(const Value: TCombinedElements);
  protected
    fArraysPresent: TCombinedElements;
    fVertexNCoords, fTexturesNCoords, fColorNcomponents: Integer;
    procedure SetCurrent( Ctx: PContext ); virtual;
  public
    function DataSize: Integer; virtual;
    property ActiveArrays: TCombinedElements read FActiveArrays write SetActiveArrays;
  end;

  //----------------------------------------------------------------------------
  //  
  TDisplayList = object( TObj )
  protected
    fContext: PContext;
    fDisplayList: Integer;
    procedure INIT; virtual;
  public
    destructor DESTROY; virtual;
    procedure BeginList;
    procedure EndList;
    procedure CallList;
  end;

function NewWindowContext( Wnd: HWnd; color_bits: Integer = 8;
         depth_bits: Integer = 16; accum_bits: Integer = 0;
         stensil_bits: Integer = 8 ): PContext;
function NewBitmapContext( Bmp: PBitmap ): PContext;

type
  //----------------------------------------------------------------------------
  // 
  TQuadricDrawStyle = ( qdsFill, qdsLine, qdsSilhouette, qdsPoint );
  TQuadricNormals = ( qnSmooth, qnFlat, qnNone );
  TQuadricOrientation = ( qoOutside, qoInside );

  PQuadric = ^TQuadric;
  TQuadric = object( TObj )
  private
    FTextured: Boolean;
    FDrawStyle: TQuadricDrawStyle;
    FNormalsType: TQuadricNormals;
    FOrientation: TQuadricOrientation;
    procedure SetTextured(const Value: Boolean);
    procedure SetDrawStyle(const Value: TQuadricDrawStyle);
    procedure SetNormalsType(const Value: TQuadricNormals);
    procedure SetOrientation(const Value: TQuadricOrientation);
  protected
    fQuadric: PGLUQuadric;
    procedure INIT; virtual;
  public
    destructor DESTROY; virtual;
    //  
    property Textured: Boolean read FTextured write SetTextured;
    property DrawStyle: TQuadricDrawStyle read FDrawStyle write SetDrawStyle;
    property NormalsType: TQuadricNormals read FNormalsType write SetNormalsType;
    property Orientation: TQuadricOrientation read FOrientation write SetOrientation;
    //  3D-
    procedure DrawSphere( radius: Double; slices, stacks: Integer );
    procedure DrawCylinder( baseradius, topradius, height: Double;
      slices, stacks: Integer );
    procedure DrawDisk( innerradius, outerradius: Double; slices, loops: Integer );
    procedure DrawPartialDisk( innerradius, outerradius: Double;
      slices, loops: Integer; startAngle, endAngle: Double );
  end;

function NewQuadric: PQuadric;

var Context: PContext;

{$IFDEF GLCONTEXT_MULTITHREADS}
function CurrentContext: PContext; overload;
function CurrentContext( thread: THandle ): PContext; overload;
{$ELSE}
var CurrentContext: PContext;
{$ENDIF}

function NewStandAloneMaterial: PMaterial;
function NewStandaloneFog: PFog;
function NewStandaloneLigthSource: PLightSource;
function NewStandaloneFont: PFontController;

function Float2Str( D: Extended; F: Integer ): String;
function Float2StrT( D: Extended; F: Integer ): String;
function NextPower2( i: Integer ): Integer;
function Point32Str( const P3: TPoint3D ): String;

implementation

function Float2Str( D: Extended; F: Integer ): String;
begin
  Str( D : 24 : F, Result );
  Result := Trim( Result );
end;

function Float2StrT( D: Extended; F: Integer ): String;
var i, j: Integer;
begin
  Result := Float2Str( D, F );
  j := pos( '.', Result );
  if j <= 0 then j := pos( ',', Result );
  if j > 0 then
  begin
    for i := Length( Result ) downto 1 do
    begin
      if Result[ i ] = '0' then Delete( Result, i, 1 )
      else
      begin
        if i = j then Delete( Result, i, 1 );
        break;
      end;
    end;
  end;
end;

function NextPower2( i: Integer ): Integer;
begin
  Result := 1;
  while Result < i do
    Result := Result shl 1;
end;

function Point32Str( const P3: TPoint3D ): String;
begin
  Result := '(' + Float2StrT( P3.x, 2 ) + ',' + Float2StrT( P3.y, 2 ) +',' +
    Float2StrT( P3.z, 2 ) + ')';
end;

function NewStandAloneMaterial: PMaterial;
begin
  new( Result, Create );
end;

function NewStandaloneFog: PFog;
begin
  new( Result, Create );
end;

function NewStandaloneLigthSource: PLightSource;
begin
  new( Result, Create );
end;

function NewStandaloneFont: PFontController;
begin
  new( Result, Create );
end;

function Color( r, g, b, a: Single ): TColor4; overload;
begin
  Result.r := r;
  Result.g := g;
  Result.b := b;
  Result.a := a;
end;

function Color4( r, g, b, a: Single ): TColor4; overload;
begin
  Result.r := r;
  Result.g := g;
  Result.b := b;
  Result.a := a;
end;

function Color( r, g, b: Single ): TColor4; overload;
begin
  Result.r := r;
  Result.g := g;
  Result.b := b;
  Result.a := 1;
end;

function Color4( r, g, b: Single ): TColor4; overload;
begin
  Result.r := r;
  Result.g := g;
  Result.b := b;
  Result.a := 1;
end;

function Color( const Q: TRGBQuad ): TColor4; overload;
begin
  Result.r := Q.rgbRed / 255;
  Result.g := Q.rgbGreen / 255;
  Result.b := Q.rgbBlue / 255;
  Result.a := Q.rgbReserved / 255;
end;

function Color4( const Q: TRGBQuad ): TColor4; overload;
begin
  Result.r := Q.rgbRed / 255;
  Result.g := Q.rgbGreen / 255;
  Result.b := Q.rgbBlue / 255;
  Result.a := Q.rgbReserved / 255;
end;

function Color( C: TColor ): TColor4; overload;
begin
  C := Color2RGB( C );
  Result.r := ((C       ) and 255) / 255;
  Result.g := ((C shr  8) and 255) / 255;
  Result.b := ((C shr 16) and 255) / 255;
  Result.a := 1;
end;

function Color4( C: TColor ): TColor4; overload;
begin
  C := Color2RGB( C );
  Result.r := ((C       ) and 255) / 255;
  Result.g := ((C shr  8) and 255) / 255;
  Result.b := ((C shr 16) and 255) / 255;
  Result.a := 1;
end;

function ColorEq( C1, C2: TColor4 ): Boolean;
begin
  Result := (C1.r = C2.r) and (C1.g = C2.g) and (C1.b = C2.b) and
            (C1.a = C2.a);
end;

function Color2Str( const c: TColor4 ): String;
begin
  Result := '(r' + Float2StrT( c.r, 2 ) + ',g' + Float2StrT( c.g, 2 ) + ',b' +
            Float2StrT( c.b, 2 ) + ',a' + Float2StrT( c.a, 2 ) + ')';
end;

function RGBA( r, g, b, a: Byte ): TRGBQuad;
begin
  Result.rgbRed := r;
  Result.rgbGreen := g;
  Result.rgbBlue := b;
  Result.rgbReserved := a;
end;

function RGBA( C: TColor4 ): TRGBQuad; overload;
begin
  Result.rgbRed := Round( C.r * 255 );
  Result.rgbGreen := Round( C.g * 255 );
  Result.rgbBlue := Round( C.b * 255 );
  Result.rgbReserved := Round( C.a * 255 );
end;

function RGB( r, g, b: Byte ): TRGBQuad;
begin
  Result.rgbRed := r;
  Result.rgbGreen := g;
  Result.rgbBlue := b;
  Result.rgbReserved := 255;
end;

function Point4( x, y, z, w: Single ): TPoint4; overload;
begin
  Result.x := x;
  Result.y := y;
  Result.z := z;
  Result.w := w;
end;

function Point4( x, y, z: Single ): TPoint4; overload;
begin
  Result.x := x;
  Result.y := y;
  Result.z := z;
  Result.w := 1;
end;

function Point4D( x, y, z, w: Double ): TPoint4D; overload;
begin
  Result.x := x;
  Result.y := y;
  Result.z := z;
  Result.w := w;
end;

function Point4D( x, y, z: Double ): TPoint4D; overload;
begin
  Result.x := x;
  Result.y := y;
  Result.z := z;
  Result.w := 1;
end;

function Point4D( const p3: TPoint3D ): TPoint4D; overload;
begin
  Result.X := p3.X;
  Result.Y := p3.Y;
  Result.Z := p3.Z;
  Result.W := 1;
end;

function Point3( x, y, z: Single ): TPoint3;
begin
  Result.x := x;
  Result.y := y;
  Result.z := z;
end;

function Point3D( X, Y, Z: Double ): TPoint3D;
begin
  Result.X := X;
  Result.Y := Y;
  Result.Z := Z;
end;

function Point3( p4: TPoint4 ): TPoint3; overload;
begin
  Result.x := p4.x;
  Result.y := p4.y;
  Result.z := p4.z;
end;

function Point3( p3D: TPoint3D ): TPoint3; overload;
begin
  Result.x := p3D.X;
  Result.y := p3D.Y;
  Result.z := p3D.Z;
end;

function Point3D( p4d: TPoint4D ): TPoint3D; overload;
begin
  Result.X := p4d.X;
  Result.Y := p4d.Y;
  Result.Z := p4d.Z;
end;

function Point3D( p4: TPoint4 ): TPoint3D; overload;
begin
  Result.X := p4.X;
  Result.Y := p4.Y;
  Result.Z := p4.Z;
end;

function Point3D( p3: TPoint3): TPoint3D; overload;
begin
  Result.X := p3.X;
  Result.Y := p3.Y;
  Result.Z := p3.Z;
end;

function Point3D( p3: TVector3d): TPoint3D; overload;
begin
  Result.X := p3[0];
  Result.Y := p3[1];
  Result.Z := p3[2];
end;

procedure Normalize( var P: TPoint3D ); overload;
var R: Double;
begin
    R := sqrt( sqr( P.X ) + sqr( P.Y ) + sqr( P.Z ) );
    P := Point3D( P.X/R, P.Y/R, P.Z/R );
end;

procedure Normalize( var P: TPoint3 ); overload;
var R: Single;
begin
    R := sqrt( sqr( P.X ) + sqr( P.Y ) + sqr( P.Z ) );
    P := Point3( P.X/R, P.Y/R, P.Z/R );
end;

function Point2( x, y: Single ): TPoint2;
begin
  Result.X := X;
  Result.Y := Y;
end;

function Point2( P2D: TPoint2D ): TPoint2; overload;
begin
  Result.x := P2D.X;
  Result.y := P2D.Y;
end;

function Point2D( X, Y: Double ): TPoint2D;
begin
  Result.X := X;
  Result.Y := Y;
end;

function Vector3( x, y, z: Single ): TVector3;
begin
  Result.x := x;
  Result.y := y;
  Result.z := z;
end;

function Vector3d( x, y, z: Double ): TVector3d;
begin
  Result[0] := x;
  Result[1] := y;
  Result[2] := z;
end;

function Vector3d( P: TPoint3D ): TVector3d; 
begin
  Result[0] := P.X;
  Result[1] := P.Y;
  Result[2] := P.Z;
end;

function Vector4d( x, y, z, w: Double ): TVector4d;
begin
  Result[0] := x;
  Result[1] := y;
  Result[2] := z;
  Result[3] := w;
end;

procedure RotateVector3( var T: TPoint3; angle: Single; axis: TPoint3 );
var M: TMatrix4d;
    R: TPoint3;
begin
    CurrentContext.PushModelview;
    //CurrentContext.PushProjection;

    glLoadIdentity;
    glRotatef( angle, axis.x, axis.y, axis.z );

    glGetDoublev( GL_MODELVIEW_MATRIX, @ M );

    R.x := M[0,0] * T.x + M[0,1] * T.y + M[0,2] * T.z;
    R.y := M[1,0] * T.x + M[1,1] * T.y + M[1,2] * T.z;
    R.z := M[2,0] * T.x + M[2,1] * T.y + M[2,2] * T.z;
    T := R;

    CurrentContext.PopModelview;
end;

function Dist3D( const P1, P2: TPoint3D ): Double;
begin
    Result := sqrt( sqr( P1.X - P2.X ) +
                    sqr( P1.Y - P2.Y ) +
                    sqr( P1.Z - P2.Z ) );
end;

function Dist3sqrt( const P1, P2: TPoint3 ): Single;
begin
    Result := sqr( P1.X - P2.X ) +
              sqr( P1.Y - P2.Y ) +
              sqr( P1.Z - P2.Z );
end;

function VectorMult3D( A, B: TPoint3D ): TPoint3D;
begin
    Result.X := A.Y * B.Z - A.Z * B.Y;
    Result.Y := A.Z * B.X - A.X * B.Z;
    Result.Z := A.X * B.Y - A.Y * B.X;
end;

function VectorMult3( A, B: TPoint3 ): TPoint3;
begin
    Result.X := A.Y * B.Z - A.Z * B.Y;
    Result.Y := A.Z * B.X - A.X * B.Z;
    Result.Z := A.X * B.Y - A.Y * B.X;
end;

procedure Transpose4D( var M: TMatrix4d );
var i, j: Integer;
    Tmp: Double;
begin
    for i := 0 to 3 do
    for j := 0 to i-1 do
    begin
        Tmp := M[i,j];
        M[i,j] := M[j,i];
        M[j,i] := Tmp;
    end;
end;

function MultMatrix4D_by_Matrix4D( const A, B: TMatrix4d ): TMatrix4d;
begin
        glPushMatrix;

	glLoadMatrixD(@B);
	glMultMatrixD(@A);
	glGetDoubleV(GL_MODELVIEW_MATRIX, @ Result);

	glPopMatrix;
end;
{var i, j, k: Integer;
begin
    for i := 0 to 3 do
    for j := 0 to 3 do
    begin
        Result[j,i] := 0;
        for k := 0 to 3 do
            Result[j,i] := Result[j,i] + A[k,i] * B[j,k];
    end;
end;}

function MultMatrix4D_by_Vector4D( const A: TMatrix4d; V: TVector4d ): TVector4d;
var i, j: Integer;
begin
    for i := 0 to 3 do
    begin
        Result[i] := 0;
        for j := 0 to 3 do
            Result[i] := Result[i] + A[j,i] * V[j];
    end;
end;

function MultMatrix4D_by_Vector3D( const A: TMatrix4d; V: TVector3d ): TVector3d;
var V4d: TVector4d;
begin
    V4d[0] := V[0];
    V4d[1] := V[1];
    V4d[2] := V[2];
    V4d[3] := 1;
    V4d := MultMatrix4D_by_Vector4D( A, V4d );
    Result[0] := V4d[0];
    Result[1] := V4d[1];
    Result[2] := V4d[2];
end;

function Det4D_( const A: TMatrix4d; row, exRow: Integer; exCols: Byte ): Double;
var col: Integer;
    neg: Integer;
    from: Integer;
begin
    if row = 3 then
    begin
        for col := 0 to 3 do
            if exCols and (1 shl col) = 0 then
            begin
                Result := A[row,col];
                Exit;
            end;
    end;
    neg := 1;
    Result := 0;
    for col := 0 to 3 do
    begin
        if exCols and (1 shl col) <> 0 then continue;
        begin
            exCols := exCols or (1 shl col);
            from := row+1;
            if from = exRow then inc( from );
            Result := Result + neg * A[row,col] * Det4D_( A, from, exRow, exCols );
            neg := -neg;
            exCols := exCols and not(1 shl col);
        end;
    end;
end;

function Determinant4D( const A: TMatrix4d ): Double;
begin
    Result := Det4D_( A, 0, -1, 0 );
end;

function Inverse4D( const A: TMatrix4d ): TMatrix4d;
var Det: Double;
    row, col, from: Integer;
    neg: Integer;
begin
    Det := Determinant4D( A );
    for row := 0 to 3 do
    for col := 0 to 3 do
    begin
        neg := 1;
        if (row+col)and 1 <> 0 then neg := -1;
        from := 0;
        if row = 1 then inc( from );
        Result[row, col] := neg * Det4D_( A, from, row, 1 shl col ) / Det;
    end;
end;

function Rotation4D( angle: Double; const axis: TVector3d ): TMatrix4d;
begin
    glPushMatrix;
    glLoadIdentity;
    glRotated( angle, axis[0], axis[1], axis[2] );
    glGetDoubleV( GL_MODELVIEW_MATRIX, @ Result );
    glPopMatrix;
end;

const ColorMaterialAsDWORD: array[ TColorMaterial ] of DWORD =
  ( {cmAmbientAndDiffuse} GL_AMBIENT_AND_DIFFUSE,
    {cmAmbient}           GL_AMBIENT,
    {cmDiffuse}           GL_DIFFUSE,
    {cmSpecular}          GL_SPECULAR,
    {cmEmission}          GL_EMISSION );

const SidesAsDWORD: array[ TSide ] of DWORD =
  ( {sFrontAndBack}       GL_FRONT_AND_BACK,
    {sFront}              GL_FRONT,
    {sBack}               GL_BACK );

const BlendSrcAsDWORD: array[ TBlendFunc ] of DWORD =
  ( {bfZero}              GL_ZERO,
    {bfOne}               GL_ONE,
    {bfColor}             GL_DST_COLOR,
    {bfOneMinusColor}     GL_ONE_MINUS_DST_COLOR,
    {bfSrcAlpha}          GL_SRC_ALPHA,
    {bfOneMinusSrcAlpha}  GL_ONE_MINUS_SRC_ALPHA,
    {bfDstAlpha}          GL_DST_ALPHA,
    {bfOneMinusDstAlpha}  GL_ONE_MINUS_DST_ALPHA,
    {bfSrcAlphaSaturate}  GL_SRC_ALPHA_SATURATE );

      BlendDstAsDWORD: array[ TBlendFunc ] of DWORD =
  ( {bfZero}              GL_ZERO,
    {bfOne}               GL_ONE,
    {bfColor}             GL_SRC_COLOR,
    {bfOneMinusColor}     GL_ONE_MINUS_SRC_COLOR,
    {bfSrcAlpha}          GL_SRC_ALPHA,
    {bfOneMinusSrcAlpha}  GL_ONE_MINUS_SRC_ALPHA,
    {bfDstAlpha}          GL_DST_ALPHA,
    {bfOneMinusDstAlpha}  GL_ONE_MINUS_DST_ALPHA,
    {bfSrcAlphaSaturate - !}  GL_SRC_ALPHA_SATURATE );

const HintQualityAsDWORD: array[ THintQuality ] of DWORD =
  ( {hqDontCare}          GL_DONT_CARE,
    {hqFastest}           GL_FASTEST,
    {hqNicest}            GL_NICEST );
  QualityAsStr: array[ THintQuality ] of String = ( 'hqDontCare', 'hqFastest',
    'hqNicest' );
  SideAsStr: array[ TSide ] of String = ( 'sFront', 'sBack', 'sFrontAndBack' );

const TextureDimensionAsDWORD: array[ 1..3 ] of DWORD =
  ( {1} GL_TEXTURE_1D,
    {2} GL_TEXTURE_2D,
    {3} GL_TEXTURE_3D );

      TextureWrapAsDWORD: array[ TTextureWrap ] of DWORD =
  ( {twRepeat} GL_REPEAT,
    {twClamp}  GL_CLAMP );

const PolygonModeAsDWORD: array[ TPolygonMode ] of DWORD =
  ( {pmFill}    GL_FILL,
    {pmLines}   GL_LINE,
    {pmPoints}  GL_POINT );

const ElementTypeAsDWORD: array[ TElementType ] of DWORD =
  ( {etShort}    GL_SHORT,
    {etInt}      GL_INT,
    {etSingle}   GL_FLOAT,
    {etDouble}   GL_DOUBLE );

      ColorEltTypeAsDWORD: array[ TColorEltType ] of DWORD =
  ( {cetByte}    GL_BYTE,
    {cetUByte}   GL_UNSIGNED_BYTE,
    {cetShort}   GL_SHORT,
    {cetUShort}  GL_UNSIGNED_SHORT,
    {cetInt}     GL_INT,
    {cetUInt}    GL_UNSIGNED_INT,
    {cetSingle}  GL_FLOAT,
    {cetDouble}  GL_DOUBLE );

const DrawModeAsDWORD: array[ TDrawMode ] of DWORD =
  ( {dmPoints}        GL_POINTS,
    {dmLineStrip}     GL_LINE_STRIP,
    {dmLineLoop}      GL_LINE_LOOP,
    {dmLines}         GL_LINES,
    {dmTriangleStrip} GL_TRIANGLE_STRIP,
    {dmTrianleFan}    GL_TRIANGLE_FAN,
    {dmTriangles}     GL_TRIANGLES,
    {dmQuadStrip}     GL_QUAD_STRIP,
    {dmQuads}         GL_QUADS,
    {dmPolygon}       GL_POLYGON );

const DepthFuncAsDWORD: array[ TDepthFunc ] of DWORD =
  ( {dfLt}            GL_LESS,
    {dfNo}            GL_NEVER,
    {dfEq}            GL_EQUAL,
    {dfNE}            GL_NOTEQUAL,
    {dfLE}            GL_LEQUAL,
    {dfGT}            GL_GREATER,
    {dfGE}            GL_GEQUAL,
    {dfAlways}        GL_ALWAYS
  );

const AlphaFuncAsDWORD: array[ TAlphaFunc ] of DWORD =
  ( {dfLt}            GL_LESS,
    {dfNo}            GL_NEVER,
    {dfEq}            GL_EQUAL,
    {dfNE}            GL_NOTEQUAL,
    {dfLE}            GL_LEQUAL,
    {dfGT}            GL_GREATER,
    {dfGE}            GL_GEQUAL,
    {dfAlways}        GL_ALWAYS
  );

const StencilOpAsDWORD: array[ TStencilOp ] of DWORD =
  ( {soKeep}          GL_KEEP,
    {soZero}          GL_ZERO,
    {soReplace}       GL_REPLACE,
    {soInc}           GL_INCR,
    {soDec}           GL_DECR,
    {soInvert}        GL_INVERT

  );

const StencilFuncAsDWORD: array[ TStencilFunc ] of DWORD =
  (  {sfNever}        GL_NEVER,
     {sfLT?}          GL_LESS,
     {sfLE}           GL_LEQUAL,
     {sfGT}           GL_GREATER,
     {sfGE}           GL_GEQUAL,
     {sgEQ}           GL_EQUAL,
     {sfNE}           GL_NOTEQUAL,
     {sfAlways}       GL_ALWAYS
  );

function bSetupPixelFormat( DC: HDC; ToWindow: Boolean;
         color_bits, depth_bits, accum_bits, stensil_bits: Integer;
         underlays, overlays: Integer ): Boolean;
var pfd: TPixelFormatDescriptor;
    PixelFormat: Integer;
begin
  FillChar( pfd, Sizeof( pfd ), 0 );
  pfd.nSize := sizeof( pfd );
  pfd.nVersion := 1;
  if ToWindow then
    pfd.dwFlags := PFD_DRAW_TO_WINDOW or PFD_SUPPORT_OPENGL or PFD_DOUBLEBUFFER
  else
    pfd.dwFlags := PFD_DRAW_TO_BITMAP or PFD_SUPPORT_OPENGL or PFD_SUPPORT_GDI;
  pfd.dwLayerMask := PFD_MAIN_PLANE;
  pfd.iPixelType := PFD_TYPE_RGBA; // PFD_TYPE_COLORINDEX;
  pfd.cColorBits := color_bits;
  pfd.cDepthBits := depth_bits;
  pfd.cAccumBits := accum_bits;
  pfd.cStencilBits := stensil_bits;

  pfd.bReserved := (underlays and 15) shl 4 or (overlays and 15);

  PixelFormat := ChoosePixelFormat( DC, @ pfd );
  if PixelFormat = 0 then
  begin
    ShowMessage( 'ChoosePixelFormat failed' );
    Halt;
  end;

  if not SetPixelFormat( DC, PixelFormat, @ pfd ) then
  begin
    ShowMessage( 'SetPixelFormat failed' );
    Halt;
  end;

  PixelFormat := GetPixelFormat( DC );
  DescribePixelFormat( DC, PixelFormat, Sizeof( pfd ), pfd );

  Result := TRUE;

end;

function NewWindowContext( Wnd: HWnd; color_bits: Integer = 8;
         depth_bits: Integer = 16; accum_bits: Integer = 0;
         stensil_bits: Integer = 8 ): PContext;
begin
  new( Result, Create );
  Result.FCtxWnd := Wnd;
  Result.FCtxDC := GetDC( Wnd );
  if Result.FCtxDC = 0 then
  begin
    ShowMessage( 'Can not obtain window DC' );
    Halt;
  end;
  if not bSetupPixelFormat( Result.FCtxDC, TRUE, color_bits, depth_bits,
     accum_bits, stensil_bits, 0, 0 ) then
  begin
    ShowMessage( 'Can not setup pixel format for OpenGL window' );
    Halt;
  end;
  Result.FCtxRC := wglCreateContext( Result.FCtxDC );
  GetClientRect( Wnd, Result.FViewRect );
  Result.MakeCurrent;
end;

function NewBitmapContext( Bmp: PBitmap ): PContext;
begin
  Bmp.RefInc;
  new( Result, Create );
  Result.FCtxDC := Bmp.Canvas.Handle;
  if Result.FCtxDC = 0 then
  begin
    ShowMessage( 'Can not obtain bitmap canvas DC' );
    Halt;
  end;
  if not bSetupPixelFormat( Result.FCtxDC, FALSE, 32, 16, 0, 0, 0, 0 ) then
  begin
    ShowMessage( 'Can not setup pixel format for OpenGL off-screen bitmap surface' );
    Halt;
  end;
  Result.FCtxRC := wglCreateContext( Result.FCtxDC );
  Result.FViewRect := Bmp.BoundsRect;
  Result.MakeCurrent;
end;

function NewQuadric: PQuadric;
begin
  new( Result, Create );
end;

//----------      -----------
{$IFDEF GLCONTEXT_MULTITHREADS}
var threadContexts: PList; //  -
    tc_sema: THandle;      //     threadContexts

function CurrentContext: PContext; overload;
begin
  Result := CurrentContext( GetCurrentThreadId );
end;

function CurrentContext( thread: THandle ): PContext; overload;
var i: Integer;
begin
  WaitForSingleObject( tc_sema, INFINITE );
  TRY
    for i := 0 to threadContexts.Count div 2 - 1 do
    begin
      if THandle( threadContexts.Items[ i * 2 ] ) = thread then
      begin
        Result := threadContexts.Items[ i * 2 + 1 ]; Exit;
      end;
    end;
    Result := nil;
  FINALLY
    ReleaseSemaphore( tc_sema, 1, nil );
  END;
end;

procedure SetCurrentContext( Ctx: PContext );
var t: THandle;
    i: Integer;
begin
  WaitForSingleObject( tc_sema, INFINITE );
  TRY
    t := GetCurrentThreadId;
    for i := 0 to threadContexts.Count div 2 - 1 do
    begin
      if THandle( threadContexts.Items[ i * 2 ] ) = t then
      begin
        threadContexts.Items[ i * 2 + 1 ] := Ctx; Exit;
      end;
    end;
    if Ctx = nil then Exit;
    threadContexts.Add( Pointer( t ) );
    threadContexts.Add( Ctx );
  FINALLY
    ReleaseSemaphore( tc_sema, 1, nil );
  END;
end;

procedure RemoveContext( Ctx: PContext );
var i: Integer;
begin
  WaitForSingleObject( tc_sema, INFINITE );
  TRY
    for i := 0 to threadContexts.Count div 2 - 1 do
    begin
      if threadContexts.Items[ i * 2 + 1 ] = Ctx then
      begin
        threadContexts.Delete( i * 2 );
        threadContexts.Delete( i * 2 );
        Exit;
      end;
    end;
  FINALLY
    ReleaseSemaphore( tc_sema, 1, nil );
  END;
end;
{$ENDIF GLCONTEXT_MULTITHREADS}

{ TContext }

destructor TContext.DESTROY;
var i: Integer;
    {$IFDEF GLCONTEXT_MULTITHREADS}
    tx: PTexture;
    j, t: Integer;
    was_cur: PContext;
    {$ENDIF}
begin
  Texture := nil;
  {$IFDEF GLCONTEXT_MULTITHREADS}
  was_cur := CurrentContext;
  for i := 0 to fTextures.Count-1 do
  begin
    tx := fTextures.Items[ i ];
    j := tx.fContexts.IndexOf( @ Self );
    if j >= 0 then
    begin
      if not IsCurrent then MakeCurrent;
      t := Integer( tx.fTexIDs.Items[ j ] );
      glDeleteTextures( 1, @ t );
      tx.fContexts.Delete( j );
      tx.fTexIDs.Delete( j );
    end;
  end;
  fTextures.Free;
  {$ENDIF}
  FFog.Free;
  FMaterial.Free;
  for i := 0 to High( FLights ) do
    FLights[ i ].Free;
  fFontList.ReleaseObjects;
  FFontController.Free;
  fTexPool.ReleaseObjects;
  if FCtxRC <> 0 then
  begin
    wglDeleteContext( FCtxRC ); FCtxRC := 0;
  end;
  if FCtxDC <> 0 then
  begin
    if FCtxWnd <> 0 then //ReleaseDC( FCtxWnd, FCtxDC );
                         DeleteDC( FCtxDC );
    FCtxDC := 0;
  end;
  if @ Self = CurrentContext then
  {$IFDEF GLCONTEXT_MULTITHREADS}
    RemoveContext( @ Self );
    //SetCurrentContext( nil );
  {$ELSE}
    CurrentContext := nil;
  {$ENDIF}
  if @ Self = Context then
  begin
    Context := nil;
  end;
  if FCtxBitmap <> nil then
    FCtxBitmap.RefDec;
  if fIndices16 <> nil then
    FreeMem( fIndices16 );
  if fIndices32 <> nil then
    FreeMem( fIndices32 );
  SetLength( fClipPlanes, 0 );
  inherited;
  {$IFDEF GLCONTEXT_MULTITHREADS}
  if (was_cur <> @ Self) then
  begin
    if was_cur <> nil then was_cur.MakeCurrent;
  end;
  {$ENDIF}
end;

function TContext.GetLight(idx: Integer): PLightSource;
begin
  Assert( (idx >= 0) and (idx <= 7), 'Only 8 light sources are available' );
  Result := FLights[ idx ];
end;

function TContext.GetPolygonMode(side: TSide): TPolygonMode;
begin
  Result := FPolygonMode[ side ];
end;

procedure TContext.INIT;
var i: Integer;
begin
  if Context = nil then
  begin
    InitOpenGL;
    Context := @ Self;
  end;
  new( FMaterial, Create ); FMaterial.fContext := @ Self;
  for i := 0 to High( FLights ) do
  begin
    new( FLights[ i ], Create );
    FLights[ i ].index := i;
    FLights[ i ].fContext := @ Self;
  end;
  FLights[ 0 ].FDiffuse.r := 1;
  FLights[ 0 ].FDiffuse.g := 1;
  FLights[ 0 ].FDiffuse.b := 1;
  FLights[ 0 ].FDiffuse.a := 1;
  FLights[ 0 ].FSpecular.r := 1;
  FLights[ 0 ].FSpecular.g := 1;
  FLights[ 0 ].FSpecular.b := 1;
  FLights[ 0 ].FSpecular.a := 1;
  new( FFog, Create ); FFog.fContext :=  @Self;
  FStippleFactor := 1;
  FStipplePattern := $FFFF;
  FFov := 45;
  FzNear := 1;
  FzFar := 10;
  FzNearOrtho := -1;
  FzFarOrtho := 1;
  FCameraZDirection := Point3D( 0, 0, 1 );
  FCameraYDirection := Point3D( 0, 1, 0 );
  new( FFontController, Create ); FFontController.fContext := @ Self;
  fFontList := NewList;
  fTexPool := NewList;
  fColorMask := RGBA( 1, 1, 1, 1 );
  {$IFDEF GLCONTEXT_MULTITHREADS}
  fTextures := NewList;
  {$ENDIF}
  fAutoActivateTexture := TRUE;
  fShadowOffset := MakePoint( 1, 1 );
end;

function TContext.IsCurrent: Boolean;
begin
  Result := CurrentContext = @ Self;
end;

procedure TContext.MakeCurrent;
begin
  if IsCurrent then Exit;
  //wglMakeCurrent( FCtxDC, FCtxRC );
  ActivateRenderingContext( FCtxDC, FCtxRC );
  {$IFDEF GLCONTEXT_MULTITHREADS}
  SetCurrentContext( @ Self );
  {$ELSE}
  CurrentContext := @ Self;
  {$ENDIF}
end;

procedure TContext.Ortho(left, right, bottom, top, z_near, z_far: Single);
begin
  FLastPerspective := FALSE;
  FzNearOrtho := z_near;
  FzFarOrtho := z_far;
  FOrthoRect := Point4( left, right, bottom, top );
  Ortho;
end;

procedure TContext.Ortho( WithViewRect: Boolean = TRUE );
begin
  FLastPerspective := FALSE;
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity;
  if (FOrthoRect.x = 0) and (FOrthoRect.y = 0) and
     (FOrthoRect.z = 0) and (FOrthoRect.w = 0) or WithViewRect then
  begin
    FOrthoRect.x := FViewRect.Left;
    FOrthoRect.y := FViewRect.Right;
    FOrthoRect.z := FViewRect.Top;
    FOrthoRect.w := FViewRect.Bottom;
  end;
  glOrtho( FOrthoRect.x, FOrthoRect.y, FOrthoRect.z, FOrthoRect.w,
           FzNearOrtho, FzFarOrtho );
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity;
  Assert( CheckErrors );
end;

procedure TContext.Perspective(newFOV, z_Near, z_Far: Double);
begin
  FFov := newFOV;
  FzNear := z_Near;
  FzFar := z_Far;
  Perspective;
end;

procedure TContext.Perspective;
begin
  FLastPerspective := TRUE;
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity;
  gluPerspective( FFOV,
    (FViewRect.Right - FViewRect.Left)/(FViewRect.Bottom - FViewRect.Top),
    FzNear, FzFar );

  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity;
  Assert( CheckErrors );
end;

procedure TContext.PolygonStipple(Pattern: Pointer);
begin
  if Pattern <> nil then
  begin
    Move( Pattern^, FPolygonStipple, Sizeof( FPolygonStipple ) );
    glPolygonStipple( Pattern );
    glEnable( GL_POLYGON_STIPPLE );
  end
  else glDisable( GL_POLYGON_STIPPLE );
  Assert( CheckErrors );
end;

procedure TContext.ScissorTest(Rect: PRect);
begin
  if Rect <> nil then
  begin
    FScissorTest := Rect^;
    glScissor( Rect.Left, Rect.Top, Rect.Right - Rect.Left, Rect.Bottom - Rect.Top );
    glEnable( GL_SCISSOR_TEST );
  end
    else
  begin
    FScissorTest := MakeRect( 0, 0, 0, 0 );
    glDisable( GL_SCISSOR_TEST );
  end;
  Assert( SetBit( FChgScissorTest ) );
end;

procedure TContext.SetAlphaTest(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FAlphaTest = Value then Exit;
  {$ENDIF}
  FAlphaTest := Value;
  if Value then glEnable( GL_ALPHA_TEST )
           else glDisable( GL_ALPHA_TEST );
  Assert( SetBit( FChgAlphaTest ) );
end;

procedure TContext.SetAutoNormal(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FAutoNormal = Value then Exit;
  {$ENDIF}
  FAutoNormal := Value;
  if Value then glEnable( GL_AUTO_NORMAL )
           else glDisable( GL_AUTO_NORMAL );
  Assert( SetBit( FChgAutoNormal ) );
end;

procedure TContext.SetBlend(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FBlend = Value then Exit;
  {$ENDIF}
  FBlend := Value;
  if Value then glEnable( GL_BLEND )
           else glDisable( GL_BLEND );
  Assert( SetBit( FChgBlend ) );
end;

procedure TContext.SetBlendFuncDst(const Value: TBlendFunc);
begin
  {$IFDEF CHECKING_STATE}
  if FBlendFuncDst = Value then Exit;
  {$ENDIF}
  FBlendFuncDst := Value;
  glBlendFunc( BlendSrcAsDWORD[ FBlendFuncSrc ],
               BlendDstAsDWORD[ FBlendFuncDst ] );
  Assert( CheckErrors );
  Assert( SetBit( FChgBlendFunc ) );
end;

procedure TContext.SetBlendFuncSrc(const Value: TBlendFunc);
begin
  {$IFDEF CHECKING_STATE}
  if Value = bfSrcAlphaSaturate then
  begin
    {$IFDEF ERROR_ON_INCORRECT_CALL}

    {$ENDIF}
    Exit;
  end;
  if FBlendFuncSrc = Value then Exit;
  {$ENDIF}
  FBlendFuncSrc := Value;
  glBlendFunc( BlendSrcAsDWORD[ FBlendFuncSrc ],
               BlendDstAsDWORD[ FBlendFuncDst ] );
  Assert( SetBit( FChgBlendFunc ) );
  Assert( CheckErrors );
end;

procedure TContext.SetColor(const Value: TColor4);
begin
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ FColor, @ Value, Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FColor := Value;
  glColor4fv( @ Value ); Assert( SetBit( FChgColor ) );
  //Assert( CheckErrors );
end;

procedure TContext.SetCullFace(const Value: TCullFace);
begin
  {$IFDEF CHECKING_STATE}
  if FCullFace = Value then Exit;
  {$ENDIF}
  FCullFace := Value;
  CASE Value OF
  cfNone: glDisable( GL_CULL_FACE );
  cfBack: begin
            glEnable( GL_CULL_FACE );
            glCullFace( GL_BACK );
          end;
  cfFront:begin
            glEnable( GL_CULL_FACE );
            glCullFace( GL_FRONT );
          end;
  END;
  Assert( SetBit( FChgCullFace ) );
  Assert( CheckErrors );
end;

procedure TContext.SetDepthTest(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FDepthTest = Value then Exit;
  {$ENDIF}
  FDepthTest := Value;
  if Value then glEnable( GL_DEPTH_TEST )
           else glDisable( GL_DEPTH_TEST );
  Assert( SetBit( FChgDepthTest ) );
end;

procedure TContext.SetDither(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FDither = Value then Exit;
  {$ENDIF}
  FDither := Value;
  if Value then glEnable( GL_DITHER )
           else glDisable( GL_DITHER );
  Assert( SetBit( FChgDither ) );
end;

procedure TContext.SetLighting(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FLighting = Value then Exit;
  {$ENDIF}
  FLighting := Value;
  if Value then glEnable( GL_LIGHTING )
           else glDisable( GL_LIGHTING );
  Assert( SetBit( FChgLighting ) );
end;

procedure TContext.SetLightModelAmbient(const Value: TColor4);
begin
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ FLightModelAmbient, @ Value, Sizeof( Value )) then Exit;
  {$ENDIF}
  FLightModelAmbient := Value;
  glLightModelfv( GL_LIGHT_MODEL_AMBIENT, @ Value );
  Assert( SetBit( FChgLightModelAmbient ) );
  Assert( CheckErrors );
end;

procedure TContext.SetLightModelLocalViewer(const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if FLightModelLocalViewer = Value then Exit;
  {$ENDIF}
  FLightModelLocalViewer := Value;
  glLightModelf( GL_LIGHT_MODEL_LOCAL_VIEWER, Value );
  Assert( SetBit( FChgLightModelLocalViewer ) );
  Assert( CheckErrors );
end;

procedure TContext.SetLightModelTwoSide(const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if FLightModelTwoSide = Value then Exit;
  {$ENDIF}
  FLightModelTwoSide := Value;
  glLightModelf( GL_LIGHT_MODEL_TWO_SIDE, Value );
  Assert( SetBit( FChgLightModelTwoSide ) );
  Assert( CheckErrors );
end;

procedure TContext.SetLineSmooth(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FLineSmooth = Value then Exit;
  {$ENDIF}
  FLineSmooth := Value;
  if Value then glEnable( GL_LINE_SMOOTH )
           else glDisable( GL_LINE_SMOOTH );
  Assert( SetBit( FChgLineSmooth ) );
end;

procedure TContext.SetLineStipple(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FLineStipple = Value then Exit;
  {$ENDIF}
  FLineStipple := Value;
  if Value then glEnable( GL_LINE_STIPPLE )
           else glDisable( GL_LINE_STIPPLE );
  Assert( SetBit( FChgLineStipple ) );
end;

procedure TContext.SetLineWidth(const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if FLineWidth = Value then Exit;
  {$ENDIF}
  FLineWidth := Value;
  glLineWidth( Value );
  Assert( SetBit( FChgLineWidth ) );
end;

procedure TContext.SetLogicOp(const Value: TLogicOp);
const OpCode: array[ TLogicOp ] of DWORD =
      ( {loDisabled}     0,
        {loClear}        GL_CLEAR,
        {loSet}          GL_SET,
        {loCopy}         GL_COPY,
        {loCopyInverted} GL_COPY_INVERTED,
        {loNoop}         GL_NOOP,
        {loInvert}       GL_INVERT,
        {loAnd}          GL_AND,
        {loNAnd}         GL_NAND,
        {loOr}           GL_OR,
        {loNOr}          GL_NOR,
        {loXOr}          GL_XOR,
        {loEQuiv}        GL_EQUIV,
        {loAndReverse}   GL_AND_REVERSE,
        {loAndInverted}  GL_AND_INVERTED,
        {loOrReverse}    GL_OR_REVERSE,
        {loOrInverted}   GL_OR_INVERTED );
begin
  {$IFDEF CHECKING_STATE}
  if FLogicOp = Value then Exit;
  {$ENDIF}
  FLogicOp := Value;
  if Value = loDisabled then glDisable( GL_LOGIC_OP )
  else begin
         glLogicOp( OpCode[ Value ] );
         glEnable( GL_LOGIC_OP );
       end;
  Assert( SetBit( FChgLogicOp ) );
  Assert( CheckErrors );
end;

procedure TContext.SetNormalize(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FNormalize = Value then Exit;
  {$ENDIF}
  FNormalize := Value;
  if Value then glEnable( GL_NORMALIZE )
           else glDisable( GL_NORMALIZE );
  Assert( SetBit( FChgNormalize ) );
end;

procedure TContext.SetPointSize(const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if FPointSize = Value then Exit;
  {$ENDIF}
  FPointSize := Value;
  glPointSize( Value );
  Assert( SetBit( FChgPointSize ) );
  Assert( CheckErrors );
end;

procedure TContext.SetPointSmooth(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FPointSmooth = Value then Exit;
  {$ENDIF}
  FPointSmooth := Value;
  if Value then glEnable( GL_POINT_SMOOTH )
           else glDisable( GL_POINT_SMOOTH );
  Assert( SetBit( FChgPointSmooth ) );
end;

procedure TContext.SetPolygonMode(side: TSide; const Value: TPolygonMode);
begin
  {$IFDEF CHECKING_STATE}
  if (side = sFrontAndBack) and
     (FPolygonMode[ sFront ] = Value) and (FPolygonMode[ sBack ] = Value) OR
     (side <> sFrontAndBack) and
     (FPolygonMode[ side ] = Value) then Exit;
  {$ENDIF}
  FPolygonMode[ side ] := Value;
  if side = sFrontAndBack then
  begin
    FPolygonMode[ sFront ] := Value;
    FPolygonMode[ sBack ] := Value;
  end;
  glPolygonMode( SidesAsDWORD[ side ], PolygonModeAsDWORD[ Value ] );
  Assert( SetBit( FChgPolygonMode ) );
  Assert( CheckErrors );
end;

procedure TContext.SetPolygonSmooth(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FPolygonSmooth = Value then Exit;
  {$ENDIF}
  FPolygonSmooth := Value;
  if Value then glEnable( GL_POLYGON_SMOOTH )
           else glDisable( GL_POLYGON_SMOOTH );
  Assert( SetBit( FChgPolygonSmooth ) );
end;

procedure TContext.SetStippleFactor(const Value: Integer);
begin
  {$IFDEF CHECKING_STATE}
  if FStippleFactor = Value then exit;
  {$ENDIF}
  FStippleFactor := Value;
  glLineStipple( FStippleFactor, FStipplePattern );
  Assert( SetBit( FChgStippleFactor ) );
  Assert( CheckErrors );
end;

procedure TContext.SetStipplePattern(const Value: Word);
begin
  {$IFDEF CHECKING_STATE}
  if FStipplePattern = Value then Exit;
  {$ENDIF}
  FStipplePattern := Value;
  glLineStipple( FStippleFactor, FStipplePattern );
  Assert( SetBit( FChgStipplePattern ) );
  Assert( CheckErrors );
end;

procedure TContext.SetCameraCoords(const Value: TPoint3D);
begin
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ FCameraCoords, @ Value, Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FCameraCoords := Value;
  LookFromCamera;
end;

procedure TContext.LookFromCamera;
begin
  gluLookAt( FCameraCoords.X, FCameraCoords.Y, FCameraCoords.Z,
             FCameraCoords.X + FCameraZDirection.X,
             FCameraCoords.Y + FCameraZDirection.Y,
             FCameraCoords.Z + FCameraZDirection.Z,
             FCameraYDirection.X, FCameraYDirection.Y, FCameraYDirection.Z );
  Assert( SetBit( FChgCamera ) );
  Assert( CheckErrors );
end;

procedure TContext.SetCameraZDirection(const Value: TPoint3D);
begin
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ FCameraZDirection, @ Value, Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FCameraZDirection := Value;
  LookFromCamera;
end;

procedure TContext.SetCameraYDirection(const Value: TPoint3D);
begin
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ FCameraYDirection, @ Value, Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FCameraYDirection := Value;
  LookFromCamera;
end;

procedure TContext.Clear(clear_ColorBuffer, clear_DepthBuffer,
  clear_AccumBuffer, clear_StencilBuffer: Boolean);
var F: DWORD;
begin
  F := 0;
  if clear_ColorBuffer then F := GL_COLOR_BUFFER_BIT;
  if clear_DepthBuffer then F := F or GL_DEPTH_BUFFER_BIT;
  if clear_AccumBuffer then F := F or GL_ACCUM_BUFFER_BIT;
  if clear_StencilBuffer then F := F or GL_STENCIL_BUFFER_BIT;
  if F = 0 then Exit;
  if clear_ColorBuffer then
    glClearColor( FColor.r, FColor.g, FColor.b, FColor.a );
  glClear( F );
end;

function TContext.NewTexture(W, H, D: Integer; Fmt: TTextureFormat=tfRGBA): PTexture;
var i: Integer;
begin
  for i := 0 to fTexPool.Count-1 do
  begin
    Result := fTexPool.Items[ i ];
    if (Result.Width = W) and (Result.Height = H) and
       (Result.Format = Fmt) then
    begin
      fTexPool.Delete( i );
      Exit;
    end;
  end;
  new( Result, Create );
  {$IFDEF GLCONTEXT_MULTITHREADS}
  Result.fContexts.Add( @ Self );
  Result.fTexIDs.Add( nil );
  FTextures.Add( Result );
  {$ELSE}
  Result.fContext := @ Self;
  {$ENDIF}
  Result.fDimension := 2;
  if h = 1 then Result.fDimension := 1
  else if d > 1 then Result.fDimension := 3;
  Result.fWidth := Min( W, MaxTextureSize );
  Result.fHeight := Min( H, MaxTextureSize );         
  Result.fDepth := Min( D, MaxTextureSize );
  Result.FFormat := Fmt;
end;

function TContext.GetMaxTextureSize: Integer;
begin
  if FMaxTexSize = 0 then
    glGetIntegerv( GL_MAX_TEXTURE_SIZE, @ fMaxTexSize );
  Result := fMaxTexSize;
end;

procedure TContext.SetTexture(const Value: PTexture);
begin
  //--!!--{$IFDEF CHECKING_STATE}
  //--!!--if FTexture = Value then Exit;
  //--!!--{$ENDIF}
  if (FTexture <> nil) and (Value = nil) then FTexture.UnBind;
  {$IFDEF GLCONTEXT_MULTITHREADS}
  Assert( IsCurrent, 'Context must be current to bind texture' );
  {$ELSE}
  Assert( (Value = nil) or (Value.fContext = @ Self), 'Texture to bind into ' +
    ' the context must be created for this context' );
  {$ENDIF}
  FTexture := Value;
  if Value <> nil then
  begin
    Value.Bind;
    Value.ApplyParams;
  end;
end;

function TContext.GetCaretX: Integer;
begin
  Result := FCaretPos.X;
end;

function TContext.GetCaretY: Integer;
begin
  Result := FCaretPos.Y;
end;

procedure TContext.SetCaretPos(const Value: TPoint);
begin
  FCaretPos := Value;
end;

procedure TContext.SetCaretX(const Value: Integer);
begin
  CaretPos := MakePoint( Value, CaretPos.Y );
end;

procedure TContext.SetCaretY(const Value: Integer);
begin
  CaretPos := MakePoint( CaretPos.X, Value );
end;

procedure TContext.Print(const s: String);
var SL: PStrList;
    x0, i, L: Integer;
    ss: String;
    NewPos: TPoint4D;
    was_Perspective: Boolean;
    {$IFDEF FAST_GLFONT}
    MD_matrix, PJ_matrix: TMatrix4D; VP_vector: TVector4I;
    {$ENDIF}
    SaveColor: TColor4;
    SavePos: TPoint;
    was_depth_test, was_depth_write: Boolean;
begin
  was_depth_test := DepthTest; DepthTest:=FALSE;
  was_depth_write := DepthWrite; DepthWrite:=FALSE;
  TRY

  if not ColorEq(ShadowColor, Color) and (ShadowColor.a <> 0) then
  begin
    SavePos := CaretPos;
    SaveColor := Color;
    Color := ShadowColor;
    FCaretPos.X := fCaretPos.X + ShadowOffset.X;
    FCaretPos.Y := fCaretPos.Y + ShadowOffset.Y;
    Print( s );
    if DoubleShadow then
    begin
      CaretPos := SavePos;
      FCaretPos.X := fCaretPos.X + 2*ShadowOffset.X;
      FCaretPos.Y := fCaretPos.Y + 2*ShadowOffset.Y;
      Print( s );
    end;
    CaretPos := SavePos;
    Color := SaveColor;
  end;
  Font.SelectFont;

  SL := NewStrList;
  TRY
    SL.Text := s;
    if SL.Count > 1 then //   ,    
    begin                //  ,      ()   X
      x0 := CaretPos.X;
      for i := 0 to SL.Count-2 do
      begin
        Print( SL.Items[ i ] ); CRLF;
        FCaretPos.X := x0;
      end;
      i := SL.Count-1; Print( SL.Items[ i ] );
      Exit;
    end;
  FINALLY
    SL.Free;
  END;

  //    
  //------    
  was_Perspective := FLastPerspective;
  glPushAttrib( GL_LIST_BIT or GL_COLOR_BUFFER_BIT or GL_CURRENT_BIT or
                GL_DEPTH_BUFFER_BIT or GL_ENABLE_BIT or GL_LIGHTING_BIT or
                GL_TEXTURE_BIT or GL_VIEWPORT_BIT );
  Lighting := FALSE;
  DepthTest := FALSE;
  Material.ColorEnabled := FALSE;
  PushModelview;

  if was_Perspective then
    Ortho( FViewRect.Left, FViewRect.Right, FViewRect.Top, FViewRect.Bottom,
           -1, 1 )
  else glLoadIdentity;
  //------   
  NewPos := Point4D( FCaretPos.X,
                     FViewRect.Bottom - FCaretPos.Y - Font.Height,
                     0, 0 );
  {$IFDEF FAST_GLFONT}
  glEnable( GL_BLEND );
  glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
  glDisable( GL_CULL_FACE );
  glTranslatef( NewPos.X, NewPos.Y, 0 );
  Texture := FSelectedFont.fFontTexture;
  Texture.MinFilter := tminLinear;
  Texture.MagFilter := tmagLinear;
  {$ELSE}
  glRasterPos3dv( @ NewPos );
  {$ENDIF}
  glListBase( FSelectedFont.fFontListID );
  ss := s;
  while ss <> '' do
  begin
    L := 1; //Min( 255, Length( ss ) );
    glCallLists( L, GL_UNSIGNED_BYTE, PChar( ss ) );
    ss := CopyEnd( ss, L+1 );
    {$IFDEF FAST_GLFONT}
    glGetDoublev( GL_MODELVIEW_MATRIX, @ MD_matrix );
    glGetDoublev( GL_PROJECTION_MATRIX, @ PJ_matrix );
    glGetIntegerv( GL_VIEWPORT, @ VP_vector );
    gluProject( 0, 0, 0, MD_matrix, PJ_matrix, VP_vector,
                @ NewPos.X, @ NewPos.Y, @ NewPos.Z );
    {$ELSE}
    glGetDoublev( GL_CURRENT_RASTER_POSITION, @ NewPos );
    {$ENDIF}
    if NewPos.X >= FViewRect.Right then
    begin
      fCaretPos.X := 0;
      fCaretPos.Y := fCaretPos.Y + Font.Height;
      {$IFDEF FAST_GLFONT}
      glTranslatef( fCaretPos.X - NewPos.X, FViewRect.Bottom - FCaretPos.Y - Font.Height - NewPos.Y, 0 );
      {$ELSE}
      glRasterPos3d( fCaretPos.X, fCaretPos.Y, 0 );
      {$ENDIF}
      NewPos := Point4D( FCaretPos.X,
                         FViewRect.Bottom - FCaretPos.Y - Font.Height,
                         0, 0 );
    end;
  end;
  FCaretPos.X := Trunc( NewPos.x );
  if FCaretPos.X < NewPos.X then FCaretPos.X := FCaretPos.X + 1;
  fNextLineY := fCaretPos.Y + Font.Height;
  //------    
  PopModelview;
  glPopAttrib;
  if was_Perspective then
  begin
    Perspective;
    LookFromCamera;
  end;
  FINALLY
    DepthTest := was_depth_test;
    DepthWrite := was_depth_write;
  END;
end;

procedure TContext.CRLF;
var y: Integer;
begin
  y := fNextLineY + fNextLineY - fCaretPos.Y;
  fCaretPos := MakePoint( 0, fNextLineY );
  if y > FViewRect.Bottom then y := 0;
  fNextLineY := y;
end;

procedure TContext.BeginScene;
var OldRect: TRect;
begin
  Assert( IsCurrent, 'Context must be current to begin scene' );
  if FCtxWnd <> 0 then
  begin
    OldRect := FViewRect;
    GetClientRect( FCtxWnd, FViewRect );
    if not CompareMem( @ FViewRect, @ OldRect, Sizeof( OldRect ) ) then
    begin
      glViewPort( FViewRect.Left, FViewRect.Top,
                  FViewRect.Right - FViewRect.Left,
                  FViewRect.Bottom - FViewRect.Top );
      if FLastPerspective then
      begin
        Perspective;
        LookFromCamera;
      end
      else Ortho;
    end;
    FCaretPos := MakePoint( 0, 0 );
  end;
end;

procedure TContext.EndScene;
begin
  Assert( IsCurrent, 'Context must be current to EndScene' );
  glFlush;
  if FCtxWnd <> 0 then
  begin
    SwapBuffers( FCtxDC );
    ValidateRect( FCtxWnd, nil);
  end;
  Assert( CheckErrors );
end;

function TContext.NewVertexArray(etype: TElementType; ncoords,
  count: Integer; initptr: Pointer; shareptr: Boolean = FALSE): PArray;
begin
  new( Result, Create );
  Result.fArrType := atVertexes;
  Result.fEltType := etype;
  Assert( (ncoords >= 2) and (ncoords <= 4), 'Only 2, 3 or 4 components of ' +
    'coordinates are allowed for a vertex array' );
  Result.fNCoords := ncoords;
  Result.fCount := count;
  if initptr = nil then
  begin
    GetMem( Result.fData, Result.DataSize );
  end
    else
  begin
    Result.fData := initptr;
    Result.fShare := shareptr;
  end;
  Result.initEltSize;
end;

function TContext.NewTexCoordArray(etype: TElementType; ncoords,
  count: Integer; initptr: Pointer; shareptr: Boolean): PArray;
begin
  new( Result, Create );
  Result.fArrType := atTexCoords;
  Result.fEltType := etype;
  Assert( (ncoords >= 1) and (ncoords <= 4), 'Only 1, 2, 3 or 4 components of ' +
    'coordinates are allowed for a texture coords array' );
  Result.fNCoords := ncoords;
  Result.fCount := count;
  if initptr = nil then
  begin
    GetMem( Result.fData, Result.DataSize );
  end
    else
  begin
    Result.fData := initptr;
    Result.fShare := shareptr;
  end;
  Result.initEltSize;
end;

function TContext.NewNormalArray(eType: TElementType; count: Integer;
  initptr: Pointer; shareptr: Boolean): PArray;
begin
  new( Result, Create );
  Result.fArrType := atNormals;
  Result.fEltType := etype;
  Result.fNCoords := 3;
  Result.fCount := count;
  if initptr = nil then
  begin
    GetMem( Result.fData, Result.DataSize );
  end
    else
  begin
    Result.fData := initptr;
    Result.fShare := shareptr;
  end;
  Result.initEltSize;
end;

function TContext.NewColorArray(eType: TColorEltType; ncoords, count: Integer;
  initptr: Pointer; shareptr: Boolean): PColorArray;
begin
  new( Result, Create );
  Result.fArrType := atColors;
  Result.fColorEltType := etype;
  Assert( (ncoords = 3) or (ncoords = 4), 'Only 3 or 4 color components are ' +
    'allowed' );
  Result.fNCoords := ncoords;
  Result.fCount := count;
  if initptr = nil then
  begin
    GetMem( Result.fData, Result.DataSize );
  end
    else
  begin
    Result.fData := initptr;
    Result.fShare := shareptr;
  end;
  Result.initEltSize;
end;

function TContext.NewEdgeFlagArray(count: Integer; initptr: Pointer;
  shareptr: Boolean): PArray;
begin
  new( Result, Create );
  Result.fArrType := atEdgeFlags;
  Result.fNCoords := 1;
  Result.fCount := count;
  if initptr = nil then
  begin
    GetMem( Result.fData, Result.DataSize );
  end
    else
  begin
    Result.fData := initptr;
    Result.fShare := shareptr;
  end;
  Result.initEltSize;
end;

function TContext.NewCombinedArray(elements: TCombinedElements;
  ncoord_vertex, ncoord_textures, ncomp_color, count: Integer; initptr: Pointer;
  shareptr: Boolean): PCombinedArray;
begin
  new( Result, Create );
  Result.fContext := @ Self;
  Result.fArrType := atCombined;
  Result.fArraysPresent := elements;
  Result.FActiveArrays := elements;
  Result.fNCoords := 0;
  Assert( not(atVertexes in elements) or (ncoord_vertex >= 2) and (ncoord_vertex <= 4),
    'ncoord_vertex must be 2, 3 or 4' );
  Assert( not(atTexCoords in elements) or (ncoord_vertex >= 1) and (ncoord_textures <= 4),
    'ncoord_textures must be 1, 2, 3 or 4' );
  Assert( not(atColors in elements) or (ncomp_color >= 3) and (ncomp_color <= 4),
    'ncomp_color must be 3 or 4' );
  Result.fVertexNCoords := ncoord_vertex;
  Result.fTexturesNCoords := ncoord_textures;
  Result.fColorNcomponents := ncomp_color;
  Result.fCount := count;
  if initptr = nil then
  begin
    GetMem( Result.fData, Result.DataSize );
  end
    else
  begin
    Result.fData := initptr;
    Result.fShare := shareptr;
  end;
  Result.initEltSize;
end;

procedure TContext.DrawElements(DrawStyle: TDrawMode; VrtxArray, TexCArray,
  NormArray, ColrArray, EdgeArray: PArray;
  indices: Pointer = nil; indcount: Integer = 0; DWORD_indices: Boolean = FALSE);
      function Min0( i, j: Integer ): Integer;
      begin
        if i = 0 then Result := j else Result := Min( i, j );
      end;
var count: Integer;
    P: Pointer;
    T: DWORD;
begin
  if (indcount > 0) and (indices <> nil) then
  begin
    P := indices;
    count := indcount;
    if count <= 65535 then T := GL_UNSIGNED_SHORT
                      else T := GL_UNSIGNED_INT;
    if DWORD_indices  then T := GL_UNSIGNED_INT;
  end
    else
  begin
    count := 0;
    if VrtxArray <> nil then count := VrtxArray.Count;
    if TexCArray <> nil then count := Min0( count, TexCArray.Count );
    if NormArray <> nil then count := Min0( count, NormArray.Count );
    if ColrArray <> nil then count := Min0( count, ColrArray.Count );
    if EdgeArray <> nil then count := Min0( count, EdgeArray.Count );
    if count = 0 then Exit;
    P := ProvideIndices( count );
    if count <= 65535 then T := GL_UNSIGNED_SHORT
                      else T := GL_UNSIGNED_INT;
  end;
  EnableVertexArray   := VrtxArray <> nil;
  EnableTexCoordArray[ 0 ] := TexCArray <> nil;
  EnableNormalArray   := NormArray <> nil;
  EnableColorArray    := ColrArray <> nil;
  EnableEdgeFlagArray := EdgeArray <> nil;
  Assert( CheckErrors );
  VrtxArray.SetCurrent( @ Self );
  TexCArray.SetCurrent( @ Self );
  NormArray.SetCurrent( @ Self );
  ColrArray.SetCurrent( @ Self );
  EdgeArray.SetCurrent( @ Self );
  Assert( CheckErrors );
  glDrawElements( DrawModeAsDWORD[ DrawStyle ], count, T, P );
  Assert( CheckErrors );
end;

procedure TContext.DrawElements(DrawStyle: TDrawMode;
  CombArray: PCombinedArray;
  indices: Pointer = nil; indcount: Integer = 0; DWORD_indices: Boolean = FALSE);
var P: Pointer;
    T: DWORD;
    count: Integer;
begin
  TRY
    if (indcount > 0) and (indices <> nil) then
    begin
      P := indices;
      count := indcount;
      if count <= 65535 then T := GL_UNSIGNED_SHORT
                        else T := GL_UNSIGNED_INT;
      if DWORD_indices  then T := GL_UNSIGNED_INT;
    end
      else
    begin
      count := CombArray.Count;
      P := ProvideIndices( CombArray.Count );
      if CombArray.Count <= 65535 then T := GL_UNSIGNED_SHORT
                                  else T := GL_UNSIGNED_INT;
    end;
    EnableVertexArray   := atVertexes  in CombArray.ActiveArrays;
    EnableTexCoordArray[ 0 ] := atTexCoords in CombArray.ActiveArrays;
    {$IFDEF MULTITEXTURING}
    if AutoActivateTexture then
    begin
      EnableTexCoordArray[ 1 ] := atTex1Coords in CombArray.ActiveArrays;
      EnableTexCoordArray[ 2 ] := atTex2Coords in CombArray.ActiveArrays;
      EnableTexCoordArray[ 3 ] := atTex3Coords in CombArray.ActiveArrays;
      EnableTexCoordArray[ 4 ] := atTex4Coords in CombArray.ActiveArrays;
    end
      else
    begin
      EnableTexCoordArray[ 0 ] := ( [ atTexCoords, atTex1Coords, atTex2Coords,
        atTex3Coords, atTex4Coords ] * CombArray.ActiveArrays ) <> [ ];
    end;
    {$ENDIF}
    EnableNormalArray   := atNormals   in CombArray.ActiveArrays;
    EnableColorArray    := atColors    in CombArray.ActiveArrays;
    EnableEdgeFlagArray := atEdgeFlags in CombArray.ActiveArrays;
    //CombArray.fContext := @ Self;
    CombArray.SetCurrent( @ Self );
    glDrawElements( DrawModeAsDWORD[ DrawStyle ], count, T, P );
    Assert( CheckErrors );
  EXCEPT
    asm
      nop
    end;
  END;
end;

function TContext.ProvideIndices(count: Integer): Pointer;
var PW: PWord;
    PD: PDWORD;
    i: Integer;
begin
  if count <= 65535 then
  begin
    if fIndices16 = nil then
    begin
      fI16count := 65535;
      GetMem( fIndices16, 65536 * Sizeof( Word ) );
      PW := fIndices16;
      for i := 0 to 65534 do
      begin
        PW^ := i; inc( PW );
      end;
    end;
    Result := fIndices16;
  end
    else
  begin
    if fI32count < count then
      if fIndices32 <> nil then
      begin
        FreeMem( fIndices32 ); fIndices32 := nil;
      end;
    if fIndices32 = nil then
    begin
      fI32count := count;
      GetMem( fIndices32, fI32count * Sizeof( DWORD ) );
      PD := fIndices32;
      for i := 0 to fI32count-1 do
      begin
        PD^ := i; inc( PD );
      end;
    end;
    Result := fIndices32;
  end;
end;

procedure TContext.SetEnableVertexArray(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FEnableVertexArray = Value then Exit;
  {$ENDIF}
  FEnableVertexArray := Value;
  if Value then glEnableClientState( GL_VERTEX_ARRAY )
           else glDisableClientState( GL_VERTEX_ARRAY );
end;

procedure TContext.SetEnableNormalArray(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FEnableNormalArray = Value then Exit;
  {$ENDIF}
  FEnableNormalArray := Value;
  if Value then glEnableClientState( GL_NORMAL_ARRAY )
           else glDisableClientState( GL_NORMAL_ARRAY )
end;

procedure TContext.SetEnableColorArray(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FEnableColorArray = Value then Exit;
  {$ENDIF}
  FEnableColorArray := Value;
  if Value then glEnableClientState( GL_COLOR_ARRAY )
           else glDisableClientState( GL_COLOR_ARRAY );
end;

procedure TContext.SetEnableEdgeFlagArray(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FEnableEdgeFlagArray = Value then Exit;
  {$ENDIF}
  FEnableEdgeFlagArray := Value;
  if Value then glEnableClientState( GL_EDGE_FLAG_ARRAY )
           else glDisableClientState( GL_EDGE_FLAG_ARRAY );
end;

procedure TContext.SetEnableTexCoordArray(idx: Integer; const Value: Boolean);
begin
  //{$IFDEF CHECKING_STATE}
  if FEnableTexCoordArray[ idx ] = Value then Exit;
  //{$ENDIF}
  {$IFDEF MULTITEXTURING}
  if idx <> 0 then
  begin
    if AutoActivateTexture then
      glClientActiveTextureARB( GL_TEXTURE0_ARB + idx );
    CheckErrors;
  end;
  {$ENDIF}
  FEnableTexCoordArray[ idx ] := Value;
  if Value then glEnableClientState( GL_TEXTURE_COORD_ARRAY )
           else glDisableClientState( GL_TEXTURE_COORD_ARRAY );
  {$IFDEF MULTITEXTURING}
  if (idx <> 0) and AutoActivateTexture then
    glClientActiveTextureARB( GL_TEXTURE0_ARB );
  {$ENDIF}
end;

procedure TContext.SetViewPort(const Value: TRect);
begin
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ FViewRect, @ Value, Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FViewRect := Value;
  glViewPort( FViewRect.Left, FViewRect.Top,
              FViewRect.Right - FViewRect.Left,
              FViewRect.Bottom - FViewRect.Top );
  Assert( CheckErrors );
end;

function TContext.ScaleFont(BaseFontHeight: Integer;
  BaseWindowHeight: Integer = 480): Integer;
begin
  Result := Round( Abs( FViewRect.Bottom - FViewRect.Top ) *
            BaseFontHeight / BaseWindowHeight );
end;

function TContext.ScaleD(BaseHeight: Double;
  BaseWindowHeight: Integer): Double;
begin
  Result := Abs( FViewRect.Bottom - FViewRect.Top ) *
            BaseHeight / BaseWindowHeight;
end;

function TContext.NewDisplayList: PDisplayList;
begin
  new( Result, Create );
  Result.fContext := @ Self;
end;

procedure TContext.SetMaterial(const Value: PMaterial);
begin
  Assert( Value <> nil, 'Assigning NIL material not allowed' );
  FMaterial.Assign( Value );
end;

procedure TContext.SetFog(const Value: PFog);
begin
  if Value = nil then Fog.Enabled := FALSE
  else FFog.Assign( Value );
end;

procedure TContext.SetLight(idx: Integer; const Value: PLightSource);
begin
  if Value = nil then Light[ idx ].Enabled := FALSE
  else Light[ idx ].Assign( Value );
end;

procedure TContext.SetFont(const Value: PFontController);
begin
  Assert( Value <> nil, 'Assigning NIL font is not allowed' );
  FFontController.Assign( Value );
end;

function TContext.GetClipPlane(idx: Integer): TPoint4D;
begin
  if idx > High( fClipPlanes ) then Result := Point4D( 0, 0, 0, 0 )
  else Result := fClipPlanes[ idx ];
end;

function TContext.GetMaxClipPlanes: Integer;
begin
  if fMaxClipPlanes = 0 then
    glGetIntegerv( GL_MAX_CLIP_PLANES, @ fMaxClipPlanes );
  Result := fMaxClipPlanes;
end;

procedure TContext.SetClipPlane(idx: Integer; const Value: TPoint4D);
begin
  if Length( fClipPlanes ) = 0 then
    SetLength( fClipPlanes, MaxClipPlanes );
  if idx >= MaxClipPlanes then Exit;
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ fClipPlanes[ idx ], @ Value, Sizeof( Value ) ) then Exit;
  {$ENDIF}
  fClipPlanes[ idx ] := Value;
  if (Value.X = 0) and (Value.Y = 0) and (Value.Z = 0) and (Value.W = 0) then
    glDisable( GL_CLIP_PLANE0 + idx )
  else
  begin
    glClipPlane( GL_CLIP_PLANE0 + idx, @ Value );
    glEnable( GL_CLIP_PLANE0 + idx );
  end;
  Assert( CheckErrors );
end;

function TContext.GetPointGranularity: Double;
begin
  if fPointGran = 0 then
    glGetDoublev( GL_POINT_SIZE_GRANULARITY, @ fPointGran );
  Result := fPointGran;
end;

function TContext.GetPointMax: Double;
begin
  if fPointSizes[ 0 ] = 0 then
    glGetDoublev( GL_POINT_SIZE_RANGE, @ fPointSizes );
  Result := fPointSizes[ 0 ];
end;

function TContext.GetPointMin: Double;
begin
  if fPointSizes[ 0 ] = 0 then
    glGetDoublev( GL_POINT_SIZE_RANGE, @ fPointSizes );
  Result := fPointSizes[ 1 ];
end;

procedure TContext.SetLineQuality(const Value: THintQuality);
begin
  {$IFDEF CHECKING_STATE}
  if (FLineQuality = Value) and (FLineQuality <> hqDontCare) then Exit;
  {$ENDIF}
  FLineQuality := Value;
  glHint( GL_LINE_SMOOTH_HINT, HintQualityAsDWORD[ FLineQuality ] );
  Assert( SetBit( FChgLineQuality ) );
end;

procedure TContext.SetPerspectiveQuality(const Value: THintQuality);
begin
  {$IFDEF CHECKING_STATE}
  if (FPerspectiveQuality = Value) and (FPerspectiveQuality <> hqDontCare) then
    Exit;
  {$ENDIF}
  FPerspectiveQuality := Value;
  glHint( GL_PERSPECTIVE_CORRECTION_HINT,
          HintQualityAsDWORD[ FPerspectiveQuality ] );
end;

procedure TContext.SetPointQuality(const Value: THintQuality);
begin
  {$IFDEF CHECKING_STATE}
  if (FPointQuality = Value) and (FPointQuality <> hqDontCare) then Exit;
  {$ENDIF}
  FPointQuality := Value;
  glHint( GL_POINT_SMOOTH_HINT, HintQualityAsDWORD[ FPointQuality ] );
  Assert( SetBit( FChgPointQuality ) );
end;

procedure TContext.SetPolygonQuality(const Value: THintQuality);
begin
  {$IFDEF CHECKING_STATE}
  if (FPolygonQuality = Value) and (FPolygonQuality <> hqDontCare) then Exit;
  {$ENDIF}
  FPolygonQuality := Value;
  glHint( GL_POLYGON_SMOOTH_HINT, HintQualityAsDWORD[ FPolygonQuality ] );
  Assert( SetBit( FChgPolygonQuality ) );
end;

function TContext.GetViewportHeight: Integer;
begin
  Result := Abs( FViewRect.Bottom - FViewRect.Top );
end;

function TContext.GetViewportWidth: Integer;
begin
  Result := Abs( FViewRect.Right - FViewRect.Left );
end;

function TContext.GetMaxLights: Integer;
begin
  if fMaxLights = 0 then
    glGetIntegerv( GL_MAX_LIGHTS, @ fMaxLights );
  Result := fMaxLights;
end;

procedure TContext.ColorMask(mr, mg, mb, ma: Boolean);
var Q: TRGBQuad;
begin
  Q.rgbRed := Byte( mb );
  Q.rgbGreen := Byte( mg );
  Q.rgbBlue := Byte( mb );
  Q.rgbReserved := Byte( ma );
  {$IFDEF CHECKING_STATE}
  if DWORD( Q ) = DWORD( fColorMask ) then Exit;
  {$ENDIF}
  fColorMask := Q;
  glColorMask( Q.rgbRed, Q.rgbGreen, Q.rgbRed, Q.rgbReserved );
end;

procedure TContext.Dump(const filname: String);
var SL: PStrList;
begin
  SL := NewStrList;
  SL.Text := DebugState;
  SL.SaveToFile( filname );
  SL.Free;
end;

function TContext.DebugState: String;
var M: TMatrix4D;
  function Int2Bin( i: DWORD; n: Integer ): String;
  var j: Integer;
  begin
    SetLength( Result, n );
    FillChar( Result[ 1 ], n, '0' );
    for j := n downto 1 do
    begin
      Result[ j ] := Int2Str( i and 1 )[ 1 ];
      i := i shr 1;
    end;
  end;
  function Bits2Bin( P: PByte; n: Integer ): String;
  begin
    Result := '';
    while n > 0 do
    begin
      Result := Result + Int2Bin( P^, Min( 8, n ) );
      dec( n, 8 ); inc( P );
    end;
  end;
  function Rect2Str( const R: TRect ): String;
  begin
    Result := '(' + Int2Str( R.Left ) + ',' + Int2Str( R.Top ) + ',' +
      Int2Str( R.Right ) + ',' + Int2Str( R.Bottom ) + ')';
  end;
  function Marker( a: Boolean ): String;
  begin
    Result := '';
    if a then Result := ' *';
  end;
  function Point42Str( P: PPoint4D ): String;
  begin
    Result := '[' + Float2StrT( P.X, 4 ) + ','#9 +
                    Float2StrT( P.Y, 4 ) + ','#9 +
                    Float2StrT( P.Z, 4 ) + ','#9 +
                    Float2StrT( P.W, 4 ) + ']';
  end;
  function Matrix2Str: String;
  begin
    Result := Point42Str( @ M[0] ) + #13 +
              Point42Str( @ M[1] ) + #13 +
              Point42Str( @ M[2] ) + #13 +
              Point42Str( @ M[3] );

  end;
const
  LogicOpAsStr: array[ TLogicOp ] of String = (
    'loDisabled', 'loClear', 'loSet', 'loCopy', 'loCopyInverted', 'loNoop',
    'loInvert', 'loAnd', 'loNAnd', 'loOr', 'loNOr', 'loXOr', 'loEQuiv',
    'loAndReverse', 'loAndInverted', 'loOrReverse', 'loOrInverted' );
  CullFaceAsStr: array[ TCullFace ] of String = (
    'cfNone', 'cfBack', 'cfFront' );
  PolygonModeAsStr: array[ TPolygonMode ] of String = ( 'pmFill', 'pmLines',
    'pmPoints' );
  BlendFuncAsStr: array[ TBlendFunc ] of String = ( 'bfZero', 'bfOne', 'bfColor',
    'bfOneMinusColor', 'bfSrcAlpha', 'bfOneMinusSrcAlpha', 'bfDstAlpha',
    'bfOneMinusDstAlpha', 'bfSrcAlphaSaturate' );
var SL: PStrList;
    i: Integer;
begin
  SL := NewStrList;
  TRY
    SL.Add( 'Context debug state: ' + Int2Hex( DWORD( @ Self ), 6 ) );
    SL.Add( 'RC: ' + Int2Str( RC ) );
    SL.Add( 'Wnd: ' + Int2Str( Wnd ) );
    SL.Add( 'IsCurrent: ' + Int2Str( Integer( IsCurrent ) ) );
    // 
    SL.Add( 'AlphaTest: ' + Int2Str( Integer( AlphaTest ) ) + Marker( FChgAlphaTest ) );
    SL.Add( 'DepthTest: ' + Int2Str( Integer( DepthTest ) ) + Marker( FChgDepthTest ) );
    SL.Add( 'Blend: ' + Int2Str( Integer( Blend ) ) + Marker( FChgBlend ) );
    SL.Add( 'AutoNormal: ' + Int2Str( Integer( AutoNormal ) ) + Marker( FChgAutoNormal ) );
    SL.Add( 'Normalize: ' + Int2Str( Integer( Normalize ) ) + Marker( FChgNormalize ) );
    SL.Add( 'Dither: ' + Int2Str( Integer( Dither ) ) + Marker( FChgDither ) );
    SL.Add( 'LogicOp: ' + LogicOpAsStr[ LogicOp ] + Marker( FChgLogicOp ) );
    // 
    SL.Add( 'CullFace: ' + CullFaceAsStr[ CullFace ] + Marker( FChgCullFace ) );
    SL.Add( 'ScissorTest: (' + Int2Str( FScissorTest.Left ) + ',' +
      Int2Str( FScissorTest.Top ) + ',' + Int2Str( FScissorTest.Right ) + ',' +
      Int2Str( FScissorTest.Bottom ) + ')' + Marker( FChgScissorTest ) );
    SL.Add( 'MaxClipPlanes: ' + Int2Str( fMaxClipPlanes ) );
    for i := 0 to High( fClipPlanes ) do
      SL.Add( 'ClipPlane[' + Int2Str( i ) + ']: ' +
      Float2StrT( fClipPlanes[ i ].X, 3 ) + ' X + ' +
      Float2StrT( fClipPlanes[ i ].Y, 3 ) + ' Y + ' +
      Float2StrT( fClipPlanes[ i ].Z, 3 ) + ' Z = ' +
      Float2StrT( fClipPlanes[ i ].W, 3 ) );
    // , 
    SL.Add( 'Color: ' + Color2Str( Color ) + Marker( FChgColor ) );
    SL.Add( 'ColorMask: ' + Color2Str( OpenGLContext.Color( fColorMask ) ) );
    SL.Add( Material.DebugState( 'Material.' ) );
    // 
    SL.Add( 'PointSmooth: ' + Int2Str( Integer( PointSmooth ) ) + Marker( FChgPointSmooth ) );
    SL.Add( 'PointQuality: ' + QualityAsStr[ PointQuality ] + Marker( FChgPointQuality ) );
    SL.Add( 'PointSize: ' + Float2StrT( PointSize, 2 ) + Marker( FChgPointSize ) );
    SL.Add( 'PointMin: ' + Float2StrT( fPointSizes[ 0 ], 2 ) );
    SL.Add( 'PointMax: ' + Float2StrT( fPointSizes[ 1 ], 2 ) );
    SL.Add( 'PointGranularity: ' + Float2StrT( fPointGran, 2 ) );
    // 
    SL.Add( 'LineSmooth: ' + Int2Str( Integer( LineSmooth ) ) + Marker( FChgLineSmooth ) );
    SL.Add( 'LineQuality: ' + QualityAsStr[ LineQuality ] + Marker( FChgLineQuality ) );
    SL.Add( 'LineWidth: ' + Float2StrT( LineWidth, 2 ) + Marker( FChgLineWidth ) );
    SL.Add( 'LineStipple: ' + Int2Str( Integer( LineStipple ) ) + Marker( FChgLineStipple ) );
    SL.Add( 'StipplePattern: ' + Int2Bin( StipplePattern, 16 ) + Marker( FChgStipplePattern ) );
    SL.Add( 'StippleFactor: ' + Int2Str( StippleFactor ) + Marker( FChgStippleFactor ) );
    // 
    SL.Add( 'PolygonSmooth: ' + Int2Str( Integer( PolygonSmooth ) )
      + Marker( FChgPolygonSmooth ));
    SL.Add( 'PolygonQuality: ' + QualityAsStr[ PolygonQuality ] +
      Marker( FChgPolygonQuality ) );
    SL.Add( 'PolygonMode[ sFront ]: ' + PolygonModeAsStr[ PolygonMode[ sFront ] ] +
      Marker( FChgPolygonMode ) );
    SL.Add( 'PolygonMode[ sBack ]: ' + PolygonModeAsStr[ PolygonMode[ sBack ] ] +
      Marker( FChgPolygonMode ) );
    SL.Add( 'PolygonStipple: ' + Bits2Bin( @ FPolygonStipple[ 0 ], 1024 ) +
      Marker( FChgPolygonStipple ) );
    // 
    SL.Add( Fog.DebugState( 'Fog.' ) );
    // 
    SL.Add( 'BlendFuncSrc: ' + BlendFuncAsStr[ BlendFuncSrc ] + Marker( FChgBlendFunc ) );
    SL.Add( 'BlendFuncDst: ' + BlendFuncAsStr[ BlendFuncDst ] + Marker( FChgBlendFunc ) );
    // 
    SL.Add( 'Lighting: ' + Int2Str( Integer( Lighting ) ) );
    SL.Add( 'LightModelAmbient: ' + Color2Str( LightModelAmbient ) );
    SL.Add( 'LightModelLocalViewer: ' + Float2StrT( LightModelLocalViewer, 2 ) );
    SL.Add( 'LightModelTwoSide: ' + Float2StrT( LightModelTwoSide, 2 ) );
    for i := 0 to 7 do
      SL.Add( Light[ i ].DebugState( 'Light[' + Int2Str( i ) + ']' ) );
    SL.Add( 'MaxLights: ' + Int2Str( fMaxLights ) );
    //    
    SL.Add( 'ViewPort: ' + Rect2Str( ViewPort ) );
    SL.Add( 'Height: ' + Int2Str( Height ) );
    SL.Add( 'Width: ' + Int2Str( Width ) );
    SL.Add( 'FOV: ' + Float2StrT( FFOV, 2 ) );
    SL.Add( 'zNear: ' + Float2StrT( FzNear, 4 ) );
    SL.Add( 'zFar: ' + Float2StrT( FzFar, 4 ) );
    SL.Add( 'Ortho-zNear: ' + Float2StrT( FzNearOrtho, 4 ) );
    SL.Add( 'Ortho-zFar: ' + Float2StrT( FzFarOrtho, 4 ) );
    SL.Add( 'IsPerspecive: ' + Int2Str( Integer( IsPerspective ) ) );
    SL.Add( 'PerspectiveQuality: ' + QualityAsStr[ PerspectiveQuality ] );
    // 
    SL.Add( 'CameraCoords: ' + Point32Str( CameraCoords ) );
    SL.Add( 'CameraZDirection: ' + Point32Str( CameraZDirection ) );
    SL.Add( 'CameraYDirection: ' + Point32Str( CameraYDirection ) );
    //     
    glGetDoublev( GL_PROJECTION_MATRIX, @ M );
    SL.Add( 'Projection matrix:'#13 + Matrix2Str );
    glGetDoublev( GL_MODELVIEW_MATRIX, @ M );
    SL.Add( 'ModelView matrix:'#13 + Matrix2Str );
    // 
    SL.Add( 'Texture: ' + Int2Hex( DWORD( Texture) , 6 ) );
    if Assigned( Texture ) then
      SL.Add( Texture.DebugState( 'Texture.' ) );
    //    
    {SL.Add( )
    property Font: PFontController read FFontController write SetFont;
    property CaretPos: TPoint read FCaretPos write SetCaretPos;
    property CaretX: Integer read GetCaretX write SetCaretX;
    property CaretY: Integer read GetCaretY write SetCaretY;
    property NextLineY: Integer read fNextLineY;
    procedure Print( const s: String );
    procedure CRLF;
    function ScaleFont( BaseFontHeight: Integer; BaseWindowHeight: Integer = 480 ):
      Integer; // :   BaseFontHeight    BaseWindowHeight,
               //        
    //  , , 
    function NewVertexArray( etype: TElementType; ncoords: Integer; count: Integer;
      initptr: Pointer; shareptr: Boolean = FALSE ): PArray;
    function NewTexCoordArray( etype: TElementType; ncoords: Integer; count: Integer;
      initptr: Pointer; shareptr: Boolean = FALSE ): PArray;
    function NewNormalArray( eType: TElementType; count: Integer;
      initptr: Pointer; shareptr: Boolean = FALSE ): PArray;
    function NewColorPointer( eType: TColorEltType; ncoords: Integer; count: Integer;
      initptr: Pointer; shareptr: Boolean = FALSE ): PColorArray;
    function NewEdgeFlagArray( count: Integer; initptr: Pointer;
      shareptr: Boolean = FALSE ): PArray;
    function NewCombinedArray( elements: TCombinedElements; ncoord_vertex,
      ncoord_textures, ncomp_color, count: Integer; initptr: Pointer; shareptr: Boolean = FALSE )
      : PCombinedArray;
    procedure DrawElements( DrawStyle: TDrawMode;
      VrtxArray, TexCArray, NormArray: PArray; ColrArray: PArray; EdgeArray: PArray;
      indices: Pointer = nil; indcount: Integer = 0; DWORD_indices: Boolean = FALSE );
      overload;
    procedure DrawElements( DrawStyle: TDrawMode; CombArray: PCombinedArray;
      indices: Pointer = nil; indcount: Integer = 0; DWORD_indices: Boolean = FALSE );
      overload;
    property EnableVertexArray: Boolean read FEnableVertexArray write SetEnableVertexArray;
    property EnableNormalArray: Boolean read FEnableNormalArray write SetEnableNormalArray;
    property EnableColorArray: Boolean read FEnableColorArray write SetEnableColorArray;
    property EnableEdgeFlagArray: Boolean read FEnableEdgeFlagArray write SetEnableEdgeFlagArray;
    property EnableTexCoordArray: Boolean read FEnableTexCoordArray write SetEnableTexCoordArray;}
  FINALLY
    Result := SL.Text;
    SL.Free;
  END;
end;

function TContext.SetBit(var Bit: Boolean): Boolean;
begin
  Bit := TRUE;
  Result := TRUE;
end;

procedure TContext.PopModelview;
begin
  glPopMatrix;
  Assert( CheckErrors );
end;

procedure TContext.PopProjection;
begin
  glMatrixMode( GL_PROJECTION );
  glPopMatrix;
  glMatrixMode( GL_MODELVIEW );
  Assert( CheckErrors );
end;

procedure TContext.PushModelview;
begin
  glPushMatrix;
  Assert( CheckErrors );
end;

procedure TContext.PushProjection;
begin
  glMatrixMode( GL_PROJECTION );
  glPushMatrix;
  glMatrixMode( GL_MODELVIEW );
  Assert( CheckErrors );
end;

procedure TContext.SetSmoothShade(const Value: Boolean);
var SM: DWORD;
begin
  {$IFDEF CHECKING_STATE}
  if FSmoothShade = Value then Exit;
  {$ENDIF}
  FSmoothShade := Value;
  if Value then SM := GL_SMOOTH
           else SM := GL_FLAT;
  glShadeModel( SM );
  Assert( CheckErrors );
end;

procedure TContext.BlendOver;
begin
  Blend := TRUE;
  BlendFuncSrc := bfSrcAlpha;
  BlendFuncDst := bfOneMinusSrcAlpha;
end;

procedure TContext.BlendTransparent;
begin
  Blend := TRUE;
  BlendFuncSrc := bfSrcAlpha;
  BlendFuncDst := bfOne;
end;

procedure TContext.SetFrontFaceCW(const Value: Boolean);
var FF: DWORD;
begin
  {$IFDEF CHECKING_STATE}
  if FFrontFaceCW = Value then Exit;
  {$ENDIF}
  if Value then FF := GL_CW
           else FF := GL_CCW;
  glFrontFace( FF );
  Assert( CheckErrors );
end;

procedure TContext.DisableAllArrays;
var i: Integer;
begin
  EnableVertexArray := FALSE;
  EnableNormalArray := FALSE;
  EnableColorArray := FALSE;
  {$IFDEF MULTITEXTURING}
  for i := 0 to 4 do
    EnableTexCoordArray[ i ] := FALSE;
  {$ELSE}
  EnableTexCoordArray[ 0 ] := FALSE;
  {$ENDIF}
  EnableEdgeFlagArray := FALSE;
end;

procedure TContext.PopTexture;
begin
  glMatrixMode( GL_TEXTURE );
  glPopMatrix;
  glMatrixMode( GL_MODELVIEW );
  Assert( CheckErrors );
end;

procedure TContext.PushTexture;
begin
  glMatrixMode( GL_TEXTURE );
  glPushMatrix;
  glMatrixMode( GL_MODELVIEW );
  Assert( CheckErrors );
end;

procedure TContext.SetDepthRange(const Value: TPoint2D);
begin
  {$IFDEF CHECKING_STATE}
  if (FDepthRange.X = Value.X) and (FDepthRange.Y = Value.Y) then Exit;
  {$ENDIF}
  FDepthRange := Value;
  glDepthRange( Value.X, Value.Y );
end;

procedure TContext.SetDepthWrite(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FDepthWrite = Value then Exit;
  {$ENDIF}
  FDepthWrite := Value;
  glDepthMask( Byte( Value ) );
end;

procedure TContext.SetDepthFunc(const Value: TDepthFunc);
begin
  {$IFDEF CHECKING_STATE}
  if FDepthFunc = Value then Exit;
  {$ENDIF}
  FDepthFunc := Value;
  glDepthFunc( DepthFuncAsDWORD[ Value ] );
end;

function TContext.GetEnableTexCoordArray(idx: Integer): Boolean;
begin
  Result := FEnableTexCoordArray[ idx ];
end;

procedure TContext.SetAlphaFunc(const Value: TAlphaFunc);
begin
  {$IFDEF CHECKING_STATE}
  if FAlphaFunc = Value then Exit;
  {$ENDIF}
  FAlphaFunc := Value;
  glAlphaFunc( AlphaFuncAsDWORD[ Value ], AlphaTestValue );
end;

procedure TContext.SetAlphaTestValue(const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if FAlphaTestValue = Value then Exit;
  {$ENDIF}
  FAlphaTestValue := Value;
  glAlphaFunc( AlphaFuncAsDWORD[ FAlphaFunc ], AlphaTestValue );
end;

procedure TContext.SetStencilOpFail(const Value: TStencilOp);
begin
  {$IFDEF CHECKING_STATE}
  if FStencilOpFail = Value then Exit;
  {$ENDIF}
  FStencilOpFail := Value;
  glStencilOp( StencilOpAsDWORD[ FStencilOpFail ],
               StencilOpAsDWORD[ FStencilOpPassWODepth ],
               StencilOpAsDWORD[ FStencilOpPass ] );
end;

procedure TContext.SetStencilOpPass(const Value: TStencilOp);
begin
  {$IFDEF CHECKING_STATE}
  if FStencilOpPass = Value then Exit;
  {$ENDIF}
  FStencilOpPass := Value;
  glStencilOp( StencilOpAsDWORD[ FStencilOpFail ],
               StencilOpAsDWORD[ FStencilOpPassWODepth ],
               StencilOpAsDWORD[ FStencilOpPass ] );
end;

procedure TContext.SetStencilOpPassWODepth(const Value: TStencilOp);
begin
  {$IFDEF CHECKING_STATE}
  if FStencilOpPassWODepth = Value then Exit;
  {$ENDIF}
  FStencilOpPassWODepth := Value;
  glStencilOp( StencilOpAsDWORD[ FStencilOpFail ],
               StencilOpAsDWORD[ FStencilOpPassWODepth ],
               StencilOpAsDWORD[ FStencilOpPass ] );
end;

procedure TContext.SetStencilTest(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FStencilTest = Value then Exit;
  {$ENDIF}
  FStencilTest := Value;
  if Value then glEnable( GL_STENCIL_TEST )
           else glDisable( GL_STENCIL_TEST );
end;

procedure TContext.SetStencilFunc(const Value: TStencilFunc);
begin
  {$IFDEF CHECKING_STATE}
  if FStencilFunc = Value then Exit;
  {$ENDIF}
  FStencilFunc := Value;
  glStencilFunc( StencilFuncAsDWORD[ Value ], FStencilRef, fStencilMask );
end;

procedure TContext.SetStencilRef(const Value: Integer);
begin
  {$IFDEF CHECKING_STATE}
  if FStencilRef = Value then Exit;
  {$ENDIF}
  FStencilRef := Value;
  glStencilFunc( StencilFuncAsDWORD[ FStencilFunc ], FStencilRef, fStencilMask );
end;

procedure TContext.SetStencilMask(const Value: Integer);
begin
  {$IFDEF CHECKING_STATE}
  if FStencilMask = Value then Exit;
  {$ENDIF}
  FStencilMask := Value;
  glStencilFunc( StencilFuncAsDWORD[ FStencilFunc ], FStencilRef, fStencilMask );
end;

{ TMaterial }

procedure TMaterial.Assign(SrcMaterail: PMaterial);
var side: TSide;
begin
  ColorEnabled := SrcMaterail.ColorEnabled;
  for side := sFront to sBack do
  begin
    ColorEffect[ side ] := SrcMaterail.ColorEffect[ side ];
    Diffuse[ side ]     := SrcMaterail.Diffuse[ side ];
    Ambient[ side ]     := SrcMaterail.Ambient[ side ];
    Specular[ side ]    := SrcMaterail.Specular[ side ];
    Emission[ side ]    := SrcMaterail.Emission[ side ];
    Shininess[ side ]   := SrcMaterail.Shininess[ side ];
  end;
end;

function TMaterial.DebugState(const prefx: String): String;
const ColorMaterialAsStr: array[ TColorMaterial ] of String = (
        'cmAmbientAndDiffuse', 'cmAmbient', 'cmDiffuse', 'cmSpecular',
        'cmEmission' );
var SL: PStrList;
    side: TSide;
begin
  SL := NewStrList;
  TRY
    SL.Add( prefx + 'ColorEnabled: ' + Int2Str( Integer( ColorEnabled ) ) );
    for side := sFront to sBack do
    begin
      SL.Add( prefx + 'ColorEffect[' + SideAsStr[ side ] + ']: ' +
        ColorMaterialAsStr[ ColorEffect[ side ] ] );
      SL.Add( prefx + 'Diffuse[' + SideAsStr[ side ] + ']: ' + Color2Str( Diffuse[ side ] ) );
      SL.Add( prefx + 'Ambient[' + SideAsStr[ side ] + ']: ' + Color2Str( Ambient[ side ] ) );
      SL.Add( prefx + 'Specular[' + SideAsStr[ side ] + ']: ' + Color2Str( Specular[ side ] ) );
      SL.Add( prefx + 'Emission[' + SideAsStr[ side ] + ']: ' + Color2Str( Emission[ side ] ) );
      SL.Add( prefx + 'Shininess[' + SideAsStr[ side ] + ']: ' + Float2StrT( Shininess[ side ], 2 ) );
    end;
    Result := SL.Text;
  FINALLY
    SL.Free;
  END;
end;

function TMaterial.GetAmbient(side: TSide): TColor4;
begin
  Result := FAmbient[ side ];
end;

function TMaterial.GetColorEffect(side: TSide): TColorMaterial;
begin
  Result := FColorEffect[ side ];
end;

function TMaterial.GetDiffuse(side: TSide): TColor4;
begin
  Result := FDiffuse[ side ];
end;

function TMaterial.GetEmission(side: TSide): TColor4;
begin
  Result := FEmission[ side ];
end;

function TMaterial.GetShininess(side: TSide): Single;
begin
  Result := FShininess[ side ];
end;

function TMaterial.GetSpecular(side: TSide): TColor4;
begin
  Result := FSpecular[ side ];
end;

procedure TMaterial.INIT;
var side: TSide;
begin
  for side := Low( side ) to High( side ) do
  begin
    FAmbient[ side ] := Color4( 0.2, 0.2, 0.2, 1 );
    FDiffuse[ side ] := Color4( 0.8, 0.8, 0.8, 1 );
    FSpecular[ side ] := Color4( 0, 0, 0, 1 );
  end;
end;

procedure TMaterial.SetAmbient(side: TSide; const Value: TColor4);
begin
  {$IFDEF CHECKING_STATE}
  if (side = sFrontAndBack) and
     CompareMem( @ Value, @ FAmbient[ sFront ], Sizeof( Value ) ) and
     CompareMem( @ Value, @ FAmbient[ sBack ], Sizeof( Value )) OR
     (side <> sFrontAndBack) and
     CompareMem( @ Value, @ FAmbient[ side ], Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FAmbient[ side ] := Value;
  if side = sFrontAndBack then
  begin
    FAmbient[ sFront ] := Value;
    FAmbient[ sBack ] := Value;
  end;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TMaterial.SetAmbient: Context must be current' );
    glMaterialfv( SidesAsDWORD[ side ], GL_AMBIENT, @ Value );
    Assert( CheckErrors );
  end;
end;

procedure TMaterial.SetAmbientAndDiffuse(side: TSide; const Value: TColor4);
begin
  {$IFDEF CHECKING_STATE}
  if (side = sFrontAndBack) and
     CompareMem( @ Value, @ FAmbient[ sFront ], Sizeof( Value ) ) and
     CompareMem( @ Value, @ FAmbient[ sBack ], Sizeof( Value )) OR
     (side <> sFrontAndBack) and
     CompareMem( @ Value, @ FAmbient[ side ], Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FAmbient[ side ] := Value;
  FDiffuse[ side ] := Value;
  if side = sFrontAndBack then
  begin
    FAmbient[ sFront ] := Value;
    FAmbient[ sBack ] := Value;
    FDiffuse[ sFront ] := Value;
    FDiffuse[ sBack ] := Value;
  end;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TMaterial.SetAmbientAndDiffuse: Context must be current' );
    glMaterialfv( SidesAsDWORD[ side ], GL_AMBIENT_AND_DIFFUSE, @ Value );
    Assert( CheckErrors );
  end;
end;

procedure TMaterial.SetColorEffect(side: TSide;
  const Value: TColorMaterial);
begin
  //{$IFDEF CHECKING_STATE}
  if (side = sFrontAndBack) and
     (FColorEffect[ sFront ] = Value) and (FColorEffect[ sBack ] = Value) OR
     (side <> sFrontAndBack) and
     (FColorEffect[ side ] = Value) then Exit;
  //{$ENDIF}
  FColorEffect[ side ] := Value;
  if side = sFrontAndBack then
  begin
    FColorEffect[ sFront ] := Value;
    FColorEffect[ sBack ] := Value;
  end;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TMaterial.SetColorEffect: Context must be current' );
    glColorMaterial( SidesAsDWORD[ side ], ColorMaterialAsDWORD[ Value ] );
    Assert( CheckErrors );
  end;
end;

procedure TMaterial.SetColorEnabled(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FColorEnabled = Value then Exit;
  {$ENDIF}
  FColorEnabled := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TMaterial.SetColorEnabled: Context must be current' );
    if Value then glEnable( GL_COLOR_MATERIAL )
             else glDisable( GL_COLOR_MATERIAL );
    Assert( CheckErrors );
  end;
end;

{$UNDEF CHECKING_STATE}
procedure TMaterial.SetDiffuse(side: TSide; const Value: TColor4);
begin
  {$IFDEF CHECKING_STATE}
  if (side = sFrontAndBack) and
     CompareMem( @ Value, @ FDiffuse[ sFront ], Sizeof( Value ) ) and
     CompareMem( @ Value, @ FDiffuse[ sBack ], Sizeof( Value )) OR
     (side <> sFrontAndBack) and
     CompareMem( @ Value, @ FDiffuse[ side ], Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FDiffuse[ side ] := Value;
  if side = sFrontAndBack then
  begin
    FDiffuse[ sFront ] := Value;
    FDiffuse[ sBack ] := Value;
  end;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TMaterial.SetDiffuse: Context must be current' );
    glMaterialfv( SidesAsDWORD[ side ], GL_DIFFUSE, @ Value );
    Assert( CheckErrors );
  end;
end;

procedure TMaterial.SetEmission(side: TSide; const Value: TColor4);
begin
  {$IFDEF CHECKING_STATE}
  if (side = sFrontAndBack) and
     CompareMem( @ Value, @ FEmission[ sFront ], Sizeof( Value ) ) and
     CompareMem( @ Value, @ FEmission[ sBack ], Sizeof( Value )) OR
     (side <> sFrontAndBack) and
     CompareMem( @ Value, @ FEmission[ side ], Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FEmission[ side ] := Value;
  if side = sFrontAndBack then
  begin
    FEmission[ sFront ] := Value;
    FEmission[ sBack ] := Value;
  end;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TMaterial.SetEmission: Context must be current' );
    glMaterialfv( SidesAsDWORD[ side ], GL_EMISSION, @ Value );
    Assert( CheckErrors );
  end;
end;

procedure TMaterial.SetShininess(side: TSide; const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if (side = sFrontAndBack) and
     (FShininess[ sFront ] = Value) and (FShininess[ sBack ] = Value) OR
     (side <> sFrontAndBack) and
     (FShininess[ side ] = Value) then Exit;
  {$ENDIF}
  FShininess[ side ] := Value;
  if side = sFrontAndBack then
  begin
    FShininess[ sFront ] := Value;
    FShininess[ sBack ] := Value;
  end;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TMaterial.SetShininess: Context must be current' );
    glMaterialf( SidesAsDWORD[ side ], GL_SHININESS, Value );
    Assert( CheckErrors );
  end;
end;

procedure TMaterial.SetSpecular(side: TSide; const Value: TColor4);
begin
  {$IFDEF CHECKING_STATE}
  if (side = sFrontAndBack) and
     CompareMem( @ Value, @ FSpecular[ sFront ], Sizeof( Value ) ) and
     CompareMem( @ Value, @ FSpecular[ sBack ], Sizeof( Value )) OR
     (side <> sFrontAndBack) and
     CompareMem( @ Value, @ FSpecular[ side ], Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FSpecular[ side ] := Value;
  if side = sFrontAndBack then
  begin
    FSpecular[ sFront ] := Value;
    FSpecular[ sBack ] := Value;
  end;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TMaterial.SetSpecular: Context must be current' );
    glMaterialfv( SidesAsDWORD[ side ], GL_SPECULAR, @ Value );
    Assert( CheckErrors );
  end;
end;

{ TLightSource }

procedure TLightSource.Assign(SrcLight: PLightSource);
begin
  Enabled              := SrcLight.Enabled;
  SpotExponent         := SrcLight.SpotExponent;
  SpotCutOff           := SrcLight.SpotCutOff;
  ConstantAttenuation  := SrcLight.ConstantAttenuation;
  LinearAttenuation    := SrcLight.LinearAttenuation;
  QuadraticAttenuation := SrcLight.QuadraticAttenuation;
  Ambient              := SrcLight.Ambient;
  Diffuse              := SrcLight.Diffuse;
  Specular             := SrcLight.Specular;
  Position             := SrcLight.Position;
  SpotDirection        := SrcLight.SpotDirection;
end;

function TLightSource.DebugState(const prefx: String): String;
var SL: PStrList;
begin
  SL := NewStrList;
  TRY
    SL.Add( prefx + 'Enabled: ' + Int2Str( Integer( Enabled ) ) );
    if Enabled then
    begin
      SL.Add( prefx + 'SpotExponent: ' + Int2Str( SpotExponent ) );
      SL.Add( prefx + 'SpotCutOff: ' + Float2StrT( SpotCutOff, 2 ) );
      SL.Add( prefx + 'ConstantAttenuation: ' + Float2StrT( ConstantAttenuation, 2 ) );
      SL.Add( prefx + 'LinearAttenuation: ' + Float2StrT( LinearAttenuation, 2 ) );
      SL.Add( prefx + 'QuadraticAttenuation: ' + Float2StrT( QuadraticAttenuation, 2 ) );
      SL.Add( prefx + 'Ambient: ' + Color2Str( Ambient ) );
      SL.Add( prefx + 'Diffuse: ' + Color2Str( Diffuse ) );
      SL.Add( prefx + 'Specular: ' + Color2Str( Specular ) );
      SL.Add( prefx + 'Position: ' + Point32Str( Point3D( Position ) ) );
      SL.Add( prefx + 'Directional: ' + Int2Str( Integer( Directional ) ) );
      SL.Add( prefx + 'SpotDirection: ' + Point32Str( SpotDirection ) );
    end;
    Result := SL.Text;
  FINALLY
    SL.Free;
  END;
end;

function TLightSource.GetDirectional: Boolean;
begin
  Result := FPosition.w = 0;
end;

procedure TLightSource.INIT;
begin
  FConstantAttenuation := 1;
  FAmbient := Color( 0, 0, 0, 1 );
  FDiffuse := Color( 0, 0, 0, 1 );
  FSpecular := Color( 0, 0, 0, 1 );
  FSpotDirection := Point3D( 0, 0, -1 );
  FSpotCutOff := 180;
end;

procedure TLightSource.SetAmbient(const Value: TColor4);
begin
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ FAmbient, @ Value, Sizeof( Value ) ) then exit;
  {$ENDIF}
  FAmbient := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TLightSource.SetAmbient: Context must be current' );
    glLightfv( GL_LIGHT0 + index, GL_AMBIENT, @ Value );
    Assert( CheckErrors );
  end;
end;

procedure TLightSource.SetConstantAttenuation(const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if FConstantAttenuation = Value then exit;
  {$ENDIF}
  FConstantAttenuation := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TLightSource.SetConstantAttenuation: Context must be current' );
    glLightf( GL_LIGHT0 + index, GL_CONSTANT_ATTENUATION, Value );
    Assert( CheckErrors );
  end;
end;

procedure TLightSource.SetDiffuse(const Value: TColor4);
begin
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ FDiffuse, @ Value, Sizeof( Value ) ) then exit;
  {$ENDIF}
  FDiffuse := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TLightSource.SetDiffuse: Context must be current' );
    glLightfv( GL_LIGHT0 + index, GL_DIFFUSE, @ Value );
    Assert( CheckErrors );
  end;
end;

procedure TLightSource.SetDirectional(const Value: Boolean);
begin
  if Value then Position := Point4( Position.x, Position.y, Position.z, 0 )
           else Position := Point4( Position.x, Position.y, Position.z, 1 );
end;

procedure TLightSource.SetEnabled(const Value: Boolean);
begin
  if Value and Assigned( fContext ) then
    fContext.Lighting := TRUE;
  {$IFDEF CHECKING_STATE}
  if FEnabled = Value then Exit;
  {$ENDIF}
  FEnabled := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TLightSource.SetEnabled: Context must be current' );
    if Value then glEnable( GL_LIGHT0 + index )
             else glDisable( GL_LIGHT0 + index );
  end;
end;

procedure TLightSource.SetLinearAttenuation(const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if FLinearAttenuation = Value then Exit;
  {$ENDIF}
  FLinearAttenuation := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TLightSource.SetLinearAttenuation: Context must be current' );
    glLightf( GL_LIGHT0 + index, GL_LINEAR_ATTENUATION, Value );
    Assert( CheckErrors );
  end;
end;

procedure TLightSource.SetPosition(const Value: TPoint4);
begin
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ Value, @ FPosition, Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FPosition := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TLightSource.SetPosition: Context must be current' );
    glLightfv( GL_LIGHT0 + index, GL_POSITION, @ Value );
    Assert( CheckErrors );
  end;
end;

procedure TLightSource.SetQuadraticAttenuation(const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if FQuadraticAttenuation = Value then Exit;
  {$ENDIF}
  FQuadraticAttenuation := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TLightSource.SetQuadraticAttenuation: Context must be current' );
    glLightf( GL_LIGHT0 + index, GL_QUADRATIC_ATTENUATION, Value );
    Assert( CheckErrors );
  end;
end;

procedure TLightSource.SetSpecular(const Value: TColor4);
begin
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ FSpecular, @ Value, Sizeof( Value ) ) then exit;
  {$ENDIF}
  FSpecular := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TLightSource.SetSpecular: Context must be current' );
    glLightfv( GL_LIGHT0 + index, GL_SPECULAR, @ Value );
    Assert( CheckErrors );
  end;
end;

procedure TLightSource.SetSpotCutOff(const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if FSpotCutOff = Value then Exit;
  {$ENDIF}
  FSpotCutOff := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TLightSource.SetSpotCutOff: Context must be current' );
    glLightf( GL_LIGHT0 + index, GL_SPOT_CUTOFF, Value );
    Assert( CheckErrors );
  end;
end;

procedure TLightSource.SetSpotDirection(const Value: TPoint3D);
begin
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ Value, @ FSpotDirection, Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FSpotDirection := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TLightSource.SetSpotDirection: Context must be current' );
    glLightfv( GL_LIGHT0 + index, GL_SPOT_DIRECTION, @ Value );
    Assert( CheckErrors );
  end;
end;

procedure TLightSource.SetSpotExponent(const Value: Integer);
begin
  {$IFDEF CHECKING_STATE}
  Assert( (Value >= 0) and (Value <= 128), 'Spot exponent ' + Int2Str( Value ) +
    ' out of range' );
  if FSpotExponent = Value then Exit;
  {$ENDIF}
  FSpotExponent := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TLightSource.SetSpotExponent: Context must be current' );
    glLighti( GL_LIGHT0 + index, GL_SPOT_EXPONENT, Value );
    Assert( CheckErrors );
  end;
end;

{ TFog }

procedure TFog.Set_End(const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if F_End = Value then Exit;
  {$ENDIF}
  F_End := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TFog.Set_End: Context must be current' );
    glFogf( GL_FOG_END, Value );
    Assert( CheckErrors );
  end;
  Enabled := TRUE;
end;

procedure TFog.SetColor(const Value: TColor4);
begin
  {$IFDEF CHECKING_STATE}
  if CompareMem( @ FColor, @ Value, Sizeof( Value ) ) then Exit;
  {$ENDIF}
  FColor := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TFog.SetColor: Context must be current' );
    glFogfv( GL_FOG_COLOR, @ Value );
    Assert( CheckErrors );
  end;
  Enabled := TRUE;
end;

procedure TFog.SetDensity(const Value: Single);
begin
  Assert( Value >= 0, 'Fog density must not be negative' );
  {$IFDEF CHECKING_STATE}
  if FDensity = Value then Exit;
  {$ENDIF}
  FDensity := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TFog.SetDensity: Context must be current' );
    glFogf( GL_FOG_DENSITY, Value );
    Assert( CheckErrors );
  end;
  Enabled := TRUE;
end;

procedure TFog.SetIndex(const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if FIndex = Value then Exit;
  {$ENDIF}
  FIndex := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TFog.SetIndex: Context must be current' );
    glFogf( GL_FOG_INDEX, Value );
    Assert( CheckErrors );
  end;
  Enabled := TRUE;
end;

procedure TFog.SetMode(const Value: TFogMode);
const FogModeAsDWORD: array[ TFogMode ] of DWORD =
      ( {fmExp}  GL_EXP,
        {fmExp2} GL_EXP2,
        {fmLinear} GL_LINEAR );
begin
  {$IFDEF CHECKING_STATE}
  if fMode = Value then Exit;
  {$ENDIF}
  FMode := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TFog.SetMode: Context must be current' );
    glFogi( GL_FOG_MODE, FogModeAsDWORD[ Value ] );
    Assert( CheckErrors );
  end;
  Enabled := TRUE;
end;

procedure TFog.SetStart(const Value: Single);
begin
  {$IFDEF CHECKING_STATE}
  if fStart = Value then Exit;
  {$ENDIF}
  FStart := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TFog.SetStart: Context must be current' );
    glFogf( GL_FOG_START, Value );
    Assert( CheckErrors );
  end;
  Enabled := TRUE;
end;

procedure TFog.SetEnabled(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FEnabled = Value then Exit;
  {$ENDIF}
  FEnabled := Value;
  if Assigned( fContext ) then
  begin
    Assert( fContext.IsCurrent, 'TFog.SetEnabled: Context must be current' );
    if Value then glEnable( GL_FOG )
             else glDisable( GL_FOG );
  end;
end;

procedure TFog.Assign(SrcFog: PFog);
begin
  Mode       := SrcFog.Mode;
  Density    := SrcFog.Density;
  Start      := SrcFog.Start;
  _End       := SrcFog._End;
  Index      := SrcFog.Index;
  Color      := SrcFog.Color;
  Enabled    := SrcFog.Enabled;
  Quality    := SrcFog.Quality;
end;

procedure TFog.INIT;
begin
  FDensity := 1;
  F_End := 1;
end;

procedure TFog.SetQuality(const Value: THintQuality);
begin
  {$IFDEF CHECKING_STATE}
  if (FQuality = Value) and (FQuality <> hqDontCare) then Exit;
  {$ENDIF}
  FQuality := Value;
  if Assigned( fContext ) then
    glHint( GL_FOG_HINT, HintQualityAsDWORD[ FQuality ] );
end;

function TFog.DebugState(const prefx: String): String;
const FogModeAsStr: array[ TFogMode ] of String = (
      'fmExp', 'fmExp2', 'fmLinear' );
var SL: PStrList;
begin
  SL := NewStrList;
  TRY
    SL.Add( prefx + 'Enabled: ' + Int2Str( Integer( Enabled ) ) );
    SL.Add( prefx + 'Mode: ' + FogModeAsStr[ Mode ] );
    SL.Add( prefx + 'Density: ' + Float2StrT( Density, 2 ) );
    SL.Add( prefx + 'Start: ' + Float2StrT( Start, 4 ) );
    SL.Add( prefx + '_End: ' + Float2StrT( _End, 4 ) );
    SL.Add( prefx + 'Index: ' + Float2StrT( Index, 4 ) );
    SL.Add( prefx + 'Color: ' + Color2Str( Color ) );
    SL.Add( prefx + 'Quality: ' + QualityAsStr[ Quality ] );
    Result := SL.Text;
  FINALLY
    SL.Free;
  END;
end;

{ TTexture }

procedure TTexture.ApplyParams;
begin
  if not Binded then Exit;
  Environment := Environment;
  Color := Color;
end;

procedure TTexture.Bind;
{$IFDEF GLCONTEXT_MULTITHREADS}
var i: Integer;
    Ctx: PContext;
{$ENDIF}
begin
  if CurrentContext = nil then
  asm
    nop
  end;
  Assert( CurrentContext <> nil, 'No current context to bind texture to' );
  {$IFDEF GLCONTEXT_MULTITHREADS}
  Ctx := CurrentContext;
  i := Ctx.FTextures.IndexOf( @ Self );
  if i < 0 then Ctx.FTextures.Add( @ Self );
  i := fContexts.Indexof( Ctx );
  if i < 0 then
  begin
    i := fContexts.Count;
    fContexts.Add( Ctx );
    fTexIDs.Add( nil );
  end;
  if fTexIDs.Items[ i ] = nil then //      
  begin
    if fImage = nil then
    begin //    ,  
      Exit;
    end;
    CalculateAlpha := caAsIs;
    LoadFromBitmap( fImage, FALSE );
    Assert( fTexIDs.Items[ i ] <> nil, 'Load texture failed' );
  end;
  glEnable( TextureDimensionAsDWORD[ Dimension ] );
  glBindTexture( TextureDimensionAsDWORD[ Dimension ],
                 Integer( fTexIDs.Items[ i ] ) );
  Ctx.fTexture := @ Self;
  Assert( CheckErrors );
  {$ELSE}
  Assert( fContext.IsCurrent, 'Context must be the current to bind a texture' );
  if fTexID = 0 then Exit;
  glEnable( TextureDimensionAsDWORD[ Dimension ] );
  glBindTexture( TextureDimensionAsDWORD[ Dimension ], fTexID );
  fContext.fTexture := @ Self;
  Assert( CheckErrors );
  {$ENDIF}
end;

function TTexture.Binded: Boolean;
{$IFDEF GLCONTEXT_MULTITHREADS}
var i: Integer;
    Ctx: PContext;
{$ENDIF}
begin
  {$IFDEF GLCONTEXT_MULTITHREADS}
  if fContexts = nil then
  asm
    nop
  end;
  i := fContexts.Indexof( CurrentContext );
  if i >= 0 then
  begin
    Ctx := fContexts.Items[ i ];
    Result := Ctx.Texture = @Self;
  end
    else
    Result := FALSE;
  {$ELSE}
  Result := fContext.Texture = @ Self;
  {$ENDIF}
end;

function TTexture.Binded(toContext: PContext): Boolean;
begin
  Result := toContext.Texture = @ Self;
end;

function TTexture.DebugState(const prefx: String): String;
const TextureFormatAsStr: array[ TTextureFormat ] of String = ( 'tfRGBA', 'tfRGB' );
      TextureEnvironmentasStr: array[ TTextureEnvironment ] of String = (
        'teModulate', 'teReplace', 'teDecal', 'teBlend' );
      TexMinFilterAsStr: array[ TTextureMinFilter ] of String = ( 'tminNearest',
        'tminLinear', 'tminNearestMipmapNearest', 'tminLinearMipmapNearest',
        'tminNearestMipmapLinear', 'tminLinearMipmapLinear' );
      TexMagFilterAsStr: array[ TTextureMagFilter ] of String = ( 'tmagNearest',
        'tmagLinear' );
      TexWrapAsStr: array[ TTextureWrap ] of String = ( 'twRepeat', 'twClamp' );
var SL: PStrList;
begin
  SL := NewStrList;
  TRY
    SL.Add( 'Context: ' + Int2Hex( DWORD( Context ), 6 ) );
    SL.Add( 'Width: ' + Int2Str( Width ) );
    SL.Add( 'Height: ' + Int2Str( Height ) );
    SL.Add( 'Dimension: ' + Int2Str( Dimension ) );
    SL.Add( 'Format: ' + TextureFormatAsStr[ Format ] );
    //  
    SL.Add( 'Environment: ' + TextureEnvironmentasStr[ Environment ] );
    SL.Add( 'Color: ' + Color2Str( Color ) );
    SL.Add( 'MinFilter: ' + TexMinFilterAsStr[ MinFilter ] );
    SL.Add( 'MagFilter: ' + TexMagFilterAsStr[ MagFilter ] );
    SL.Add( 'WrapS: ' + TexWrapAsStr[ WrapS ] );
    SL.Add( 'WrapT: ' + TexWrapAsStr[ WrapT ] );
    //  :
    SL.Add( 'CalculateAlpha: ' + Int2Str( Integer( CalculateAlpha ) ) );
    //SL.Add( 'Alpha: ' + Float2StrT( Alpha, 2 ) );
    Result := SL.Text;
  FINALLY
    SL.Free;
  END;
end;

destructor TTexture.DESTROY;
{$IFDEF GLCONTEXT_MULTITHREADS}
var i, t: Integer;
    Ctx: PContext;
{$ENDIF}
begin
  {$IFDEF GLCONTEXT_MULTITHREADS}
  fImage.Free;
  for i := fContexts.Count-1 downto 0 do
  begin
    Ctx := fContexts.Items[ i ];
    if Binded( Ctx ) or (fTexIds.Items[ i ] <> nil) then
    begin
      Ctx.MakeCurrent;
      UnBind;
      t := Integer( fTexIDs.Items[ i ] );
      glDeleteTextures( 1, @ t );
    end;
    Ctx.FTextures.Remove( @ Self );
  end;
  {$ELSE}
  if Binded then UnBind;
  if fTexID <> 0 then
    glDeleteTextures( 1, @ fTexID );
  if fContext.Texture = @ Self then
    fContext.Texture := nil;
  {$ENDIF}
  inherited;
end;

function TTexture.GetDimension: Integer;
begin
  Result := fDimension;
  //if fHeight > 1 then Result := 2;
end;

procedure TTexture.INIT;
begin
  fAlpha := 1;
  {$IFDEF GLCONTEXT_MULTITHREADS}
  fContexts := NewList;
  fTexIDs := NewList;
  {$ENDIF}
  fDimension := 2;
  fMipMap := TRUE;
end;

function TTexture.Load(const rsrcname: String): Boolean;
begin
  if TextureFirstLoadFromResourve then
  begin
    Result := LoadFromResourceName( rsrcname );
    if Result then Exit;
    Result := LoadFromFile( rsrcname );
  end
    else
  begin
    Result := LoadFromFile( rsrcname );
    if Result then Exit;
    Result := LoadFromResourceName( rsrcname );
  end;
end;

function TTexture.LoadFromBitmap(B: PBitmap;
  SwapRB: Boolean = TRUE): Boolean;
//var i: Integer;
begin
  if B <> nil then
    if Dimension < 3 then
         Result := LoadFromBitmapPartially( B, 0, B.Height-1, 0, SwapRB )
    else
    begin
        Result := LoadFromBitmapPartially( B, 0, B.Height-1, 0, SwapRB );
        {for i := 0 to Depth-1 do
        begin
        Result := Result and
               LoadFromBitmapPartially( B, (B.Height div Depth) * i,
                   (B.Height div Depth) * (i+1) - 1, i, SwapRB );
        end;}
    end
  else
    Result := LoadFromBitmapPartially( B, 0, 0, 0, SwapRB );
end;

function TTexture.LoadFromBitmapMipMapLevel(level: Integer; B: PBitmap; SwapRB: Boolean): Boolean;
var B1, B2: PBitmap;
begin
  Result := LoadFromBitmapPartially( B, 0, 0, 0, SwapRB, level );
  if Result then
  begin
    B1 := B;
    TRY
      while (B1.Width > 1) and (B1.Height > 1) and Result do
      begin
        inc( level );
        B2 := NewDibBitmap( Max( 1, B1.Width div 2 ), Max( 1, B1.Height div 2 ), pf32bit );
        TRY
          B1.StretchDraw( B2.Canvas.Handle, B2.BoundsRect );
          Result := Result and LoadFromBitmapPartially( B2, 0, -1, 0, FALSE, level );
        FINALLY
          if B1 <> B then B1.Free;
          B1 := B2;
        END;
      end;
    FINALLY
      if B1 <> B then
        B1.Free;
    END;
  end;
end;

var BmpN: Integer;

function TTexture.LoadFromBitmapPartially(B: PBitmap; y0, y1, z: Integer;
  SwapRB: Boolean = TRUE; level: Integer = 0): Boolean;
var B2, B3: PBitmap;
    PF: TPixelFormat;
    rebuild: Boolean;
    y, w, n, f: Integer;
    Dst: PDWORD;
    c: Byte;
    {$IFDEF GLCONTEXT_MULTITHREADS}
    t: Integer;
    {$ENDIF}
    i, j: Integer;
    Ctx: PContext;
    SumColors: array[ 0..3 ] of Integer;
    TotalPixels: Integer;
begin
  {$IFDEF GLCONTEXT_MULTITHREADS}
  Ctx := CurrentContext;
  {$ELSE}
  Ctx := fContext;
  {$ENDIF}
  c := Min( 255, Round( 255 * fAlpha ) );
  if FFormat = tfRGBA then PF := pf32bit
                      else PF := pf24bit;
  Result := FALSE;
  while not Result do
  begin
    B2 := B;
    if (fWidth = 0) {or (NextPower2( B.Width ) <> B.Width)} then
       fWidth := NextPower2( B.Width );
    if (fHeight = 0) {or (NextPower2( B.Height ) <> B.Height)} then
       fHeight := NextPower2( B.Height );
    if (B.PixelFormat <> PF) or
       (fDimension < 3) and ((B.Width <> fWidth div (1 shl (level and 1023))) or
       (B.Height <> fHeight div (1 shl (level and 1023)))) or
       (B.Width <> fWidth div (1 shl (level and 1023))) then
    begin
      if fDimension < 3 then
      begin
          B2 := NewDibBitmap( fWidth, fHeight, PF );
          SetStretchBltMode( B2.Canvas.Handle, halftone );
          B.StretchDraw( B2.Canvas.Handle, B2.BoundsRect );
          if y1 >= B2.Height then y1 := B2.Height-1;
          if y0 >= B2.Height then y0 := B2.Height-1;
      end
        else
      begin
          B2 := NewDibBitmap( fWidth, B.Height, PF );
          SetStretchBltMode( B2.Canvas.Handle, halftone );
          B.StretchDraw( B2.Canvas.Handle, B2.BoundsRect );
          if y1 >= fHeight then y1 := fHeight-1;
          if y0 >= fHeight then y0 := fHeight-1;
      end;
    end;
    //if (y0 = y1) and (y1 = 0) then y1 := B2.Height-1;
    TRY
      {$IFDEF GLCONTEXT_MULTITHREADS}
      i := fContexts.IndexOf( CurrentContext );
      if i < 0 then
      begin
        i := fContexts.Count;
        fContexts.Add( CurrentContext );
        fTexIDs.Add( nil );
      end;
      rebuild := fTexIDs.Items[ i ] = nil;
      {$ELSE}
      rebuild := fTexID = 0;
      {$ENDIF}

      if Assigned( OnBitmap ) then
        OnBitmap( @ Self, B2 );

      if SwapRB then
      begin
        for y := y0 to y1 do
        begin
          Dst := B2.ScanLine[ y ];
          w := B2.Width;
          if B2.PixelFormat = pf24bit then
          asm
            PUSH ESI
            PUSH EDI
            MOV  ESI, [Dst]
            MOV  EDI, ESI
            MOV  ECX, [w]
          @@loop:
            LODSD
            DEC ESI
            BSWAP EAX
            SHR   EAX, 8
            STOSW
            SHR   EAX, 8
            STOSB
            LOOP @@loop
            POP  EDI
            POP  ESI
          end
            else // pf32bit
          if CalculateAlpha = caSetConst then
          asm
            PUSH ESI
            PUSH EDI
            MOV  ESI, [Dst]
            MOV  EDI, ESI
            MOV  ECX, [w]
            MOV  DL,  [c]
            SHL  EDX, 24
          @@loop:
            LODSD
            BSWAP EAX
            SHR   EAX, 8
            OR    EAX, EDX
            STOSD
            LOOP @@loop
            POP  EDI
            POP  ESI
          end
            else
          if CalculateAlpha = caAsIs then
          asm
            PUSH ESI
            PUSH EDI
            MOV  ESI, [Dst]
            MOV  EDI, ESI
            MOV  ECX, [w]
          @@loop:
            LODSD
            MOV  EDX, EAX
            AND  EDX, $FF000000
            BSWAP EAX
            SHR   EAX, 8
            OR    EAX, EDX
            STOSD
            LOOP @@loop
            POP  EDI
            POP  ESI
          end
            else
          if CalculateAlpha = caNotBlack then
          asm
            PUSH ESI
            PUSH EDI
            MOV  ESI, [Dst]
            MOV  EDI, ESI
            MOV  ECX, [w]
          @@loop:
            LODSD
            BSWAP EAX
            SHR   EAX, 8
            JZ    @@1
            OR    EAX, $FF000000
          @@1:
            STOSD
            LOOP @@loop
            POP  EDI
            POP  ESI
          end;
        end;
      end
        else
      begin
        if CalculateAlpha <> caAsIs then
        begin
          for y := 0 to y1 do
          begin
            Dst := B2.ScanLine[ y ];
            w := B2.Width;
            if CalculateAlpha = caSetConst then
            asm
              PUSH ESI
              PUSH EDI
              MOV  ESI, [Dst]
              MOV  EDI, ESI
              MOV  ECX, [w]
              MOV  DL,  [c]
              SHL  EDX, 24
            @@loop:
              LODSD
              AND   EAX, $FFFFFF
              OR    EAX, EDX
              STOSD
              LOOP @@loop
              POP  EDI
              POP  ESI
            end
              else
            if CalculateAlpha = caNotBlack then
            asm
              PUSH ESI
              PUSH EDI
              MOV  ESI, [Dst]
              MOV  EDI, ESI
              MOV  ECX, [w]
            @@loop:
              LODSD
              AND   EAX, $FFFFFF
              JZ    @@1
              OR    EAX, $FF000000
            @@1:
              STOSD
              LOOP @@loop
              POP  EDI
              POP  ESI
            end;
          end;
        end;
      end;

      if CalculateAverageColor then
      begin
        for j := 0 to 3 do SumColors[ j ] := 0;
        w := B2.Width;
        for y := y0 to y1 do
        begin
          Dst := B2.ScanLine[ y ];
          if B2.PixelFormat = pf24bit then
          asm
            PUSH ESI
            PUSH EDI
            PUSH EBX
            MOV  ESI, [Dst]
            MOV  EDI, ESI
            MOV  ECX, [w]
            LEA  EBX, [SumColors]
          @@loop:
            LODSD
            DEC ESI
            MOVZX EDX, AL
            ADD [EBX], EDX
            MOVZX EDX, AH
            ADD [EBX+4], EDX
            SHR EAX, 16
            MOVZX EDX, AL
            ADD [EBX+8], EDX
            ADD [EBX+12], 255
            LOOP @@loop
            POP  EBX
            POP  EDI
            POP  ESI
          end
            else
          asm
            PUSH ESI
            PUSH EDI
            PUSH EBX
            MOV  ESI, [Dst]
            MOV  EDI, ESI
            MOV  ECX, [w]
            LEA  EBX, [SumColors]
          @@loop:
            LODSD
            MOVZX EDX, AL
            ADD [EBX], EDX
            MOVZX EDX, AH
            ADD [EBX+4], EDX
            SHR EAX, 16
            MOVZX EDX, AL
            ADD [EBX+8], EDX
            MOVZX EDX, AH
            ADD [EBX+12], EDX
            LOOP @@loop
            POP  EBX
            POP  EDI
            POP  ESI
          end
        end;
        TotalPixels := (y1 - y0 + 1)*w;
        fAvgColor := Color4( SumColors[ 0 ]/TotalPixels,
                             SumColors[ 1 ]/TotalPixels,
                             SumColors[ 2 ]/TotalPixels,
                             SumColors[ 3 ]/TotalPixels );
      end;

      {$IFDEF GLCONTEXT_MULTITHREADS}
      fImage := NewBitmap( 0, 0 );
      fImage.Assign( B2 );
      {$ENDIF}

      if rebuild then
      begin
      {$IFDEF GLCONTEXT_MULTITHREADS}
        glGenTextures( 1, @ t );
        fTexIDs.Items[ i ] := Pointer( t );
      {$ELSE}
        glGenTextures( 1, @ fTexID );
      {$ENDIF}
      end;

      Bind;
      if PF = pf32bit then
      begin
        n := 4;
        f := GL_RGBA;
        //if  Compressed then
        //    f := GL_COMPRESSED_RGBA_ARB;
      end
      else // pf24bit
      begin
        n := 3;
        f := GL_RGB;
        //if  Compressed then
        //    f := GL_COMPRESSED_RGB_ARB;
      end;
        Result := TRUE;
        TRY
          if rebuild and ((level and 1023) = 0)
          or (level and 1024 <> 0) and
          (y0 = 0) and (y1 = fHeight div (1 shl (level and 1023)) -1) then
          begin
            CASE Dimension OF
            1: gluBuild1DMipmaps(GL_TEXTURE_1D, {components} n, fWidth,
                 {format} f, GL_UNSIGNED_BYTE, B2.ScanLine[ 0 ] );
            2:
               if  MipMap AND
               ((fWidth div 2) and 1 = 0) and (level and 1024 = 0) then
                   gluBuild2DMipmaps(GL_TEXTURE_2D, {components} n,
                       fWidth, fHeight,
                       {format} f, GL_UNSIGNED_BYTE,
                       B2.ScanLine[ B2.Height-1 ] )
               else
               begin
                   (*gluBuild2DMipmaps(GL_TEXTURE_2D, {components} n,
                       fWidth, fHeight,
                       {format} f, GL_UNSIGNED_BYTE,
                       B2.ScanLine[ B2.Height-1 ] )*)
                   i := 0; //max( fWidth-2, fHeight-2 ) div 128;
                   j := 0;
                   if  level and 1024 <> 0 then
                   begin
                       i := level and 1023;
                       j := i;
                   end;
                   for i := j to i do
                   begin
                       if  (i > 0) and (level and 1024 = 0) then
                       begin
                           B3 := NewDibBitmap( (B2.Width-2) div 2 + 2,
                                               (B2.Height-2)div 2 + 2,
                                               B2.PixelFormat  );
                           SetStretchBltMode( B3.Canvas.Handle, halftone );
                           B2.StretchDraw( B3.Canvas.Handle, B3.BoundsRect );
                           B2.Free;
                           B2 := B3;
                       end;
                       //B2.SaveToFile( GetStartDir + 'test' + Int2Str(BmpN) + '.bmp' );
                       inc( BmpN );
                       glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );

                       glTexImage2D( GL_TEXTURE_2D, i, n, B2.Width, B2.Height,
                           (fWidth div 2) and 1, f, GL_UNSIGNED_BYTE,
                           B2.ScanLine[B2.Height-1]);
                       //CheckErrors( 2 );
                       if  (B2.Width < 128) or (B2.Height < 128) or not MipMap then break;
                   end;
               end;
            3: //gluBuild3DMipmaps(GL_TEXTURE_3D, {components} n, fWidth, fHeight, fDepth,
               //  {format} f, GL_UNSIGNED_BYTE, B2.ScanLine[ B2.Height-1 ] );
               glTexImage3D( GL_TEXTURE_3D, level and 1023, GL_RGBA, fWidth, fHeight, fDepth,
                                 0, f, GL_UNSIGNED_BYTE, B2.ScanLine[B2.Height-1] );
            END;
          end
          else
          begin
            CASE Dimension OF
            1:  //if (y0 > 0) or (y1 < B2.Height-1) then
                   glTexSubImage1D( GL_TEXTURE_1D, level and 1023,
                                    0, fWidth, {format} f, GL_UNSIGNED_BYTE,
                                    B2.ScanLine[ 0 ] );
                //else
                //   glTexImage1D( GL_TEXTURE_1D, level, GL_RGBA, fWidth, 0, f,
                //                 GL_UNSIGNED_BYTE, B2.ScanLine[ 0 ] );
            2:  //if (y0 > 0) or (y1 < B2.Height-1) then
                   glTexSubImage2D( GL_TEXTURE_2D, level and 1023,
                                    0, fHeight-y1-1, fWidth, y1-y0+1,
                                    {format} f, GL_UNSIGNED_BYTE,
                                    B2.ScanLine[ y1 ] )
                //else
                //   glTexImage2D( GL_TEXTURE_2D, level, GL_RGBA, fWidth, fHeight,
                //                 fWidth and 1, f, GL_UNSIGNED_BYTE, B2.ScanLine[fHeight-1] )
                ;
            3: glTexSubImage3D( GL_TEXTURE_2D, level and 1023,
                                    0, fHeight-y1-1, z,
                                    fWidth, y1-y0+1, 1,
                                    {format} f, GL_UNSIGNED_BYTE,
                                    B2.ScanLine[ y1 ] );
            END;
          end;
        EXCEPT
          Result := FALSE;
        END;

        if not CheckErrors( 2 ) then
           Result := FALSE;

        if not Result and
           ( (fWidth >= Ctx.MaxTextureSize) or
             (fHeight >= Ctx.MaxTextureSize) ) and
           (Ctx.MaxTextureSize > 256) then
        begin
          if (fWidth > Ctx.MaxTextureSize) or
             (fHeight > Ctx.MaxTextureSize) then
          begin
            fWidth := Min( fWidth, Ctx.MaxTextureSize );
            fHeight := Min( fHeight, Ctx.MaxTextureSize );
          end
            else
          if Ctx.MaxTextureSize > 256 then
          begin
            Ctx.FMaxTexSize := Ctx.FMaxTexSize div 2;
            fWidth := Min( fWidth, Ctx.MaxTextureSize );
            fHeight := Min( fHeight, Ctx.MaxTextureSize );
          end
            else break;
        end;
    FINALLY
      if B2 <> B then B2.Free;
    END;
  end;

end;

function TTexture.LoadFromFile(const filename: String): Boolean;
var Strm: PStream;
    filname: String;
begin
  Result := FALSE;
  filname := filename;
  if pos( ':', filname ) <= 0 then
  begin
    filname := GetStartDir + filname;
    if not FileExists( filname ) and
       {$IFDEF SUPPORT_JPEG}
       not FileExists( filname + '.jpg' ) and
       {$ENDIF}
       {$IFDEF SUPPORT_PNG}
       not FileExists( filname + '.png' ) and
       {$ENDIF}
       not FileExists( filname + '.bmp' ) then
       filname := GetStartDir + filename + 'Textures\';
  end;
  if not FileExists( filname ) then
  begin
    {$IFDEF SUPPORT_JPEG}
    if FileExists( filname + '.jpg' ) then filname := filname + '.jpg'
    else
    {$ENDIF}
    {$IFDEF SUPPORT_PNG}
    if FileExists( filname + '.png' ) then filname := filname + '.png'
    else
    {$ENDIF}
    if FileExists( filname + '.bmp' ) then filname := filname + '.bmp'
    else Exit;
  end;
  Strm := NewReadFileStream( filname );
  TRY
    Result := LoadFromStream( Strm );
  FINALLY
    Strm.Free;
  END;
end;

function TTexture.LoadFromFileMipMap(level: Integer;
  const filename: String): Boolean;
var Strm: PStream;
    filname: String;
begin
  Result := FALSE;
  filname := filename;
  if pos( ':', filname ) <= 0 then
  begin
    filname := GetStartDir + filname;
    if not FileExists( filname ) and
       {$IFDEF SUPPORT_JPEG}
       not FileExists( filname + '.jpg' ) and
       {$ENDIF}
       {$IFDEF SUPPORT_PNG}
       not FileExists( filname + '.png' ) and
       {$ENDIF}
       not FileExists( filname + '.bmp' ) then
       filname := GetStartDir + filename + 'Textures\';
  end;
  if not FileExists( filname ) then
  begin
    {$IFDEF SUPPORT_JPEG}
    if FileExists( filname + '.jpg' ) then filname := filname + '.jpg'
    else
    {$ENDIF}
    {$IFDEF SUPPORT_PNG}
    if FileExists( filname + '.png' ) then filname := filname + '.png'
    else
    {$ENDIF}
    if FileExists( filname + '.bmp' ) then filname := filname + '.bmp'
    else Exit;
  end;
  Strm := NewReadFileStream( filname );
  TRY
    Result := LoadFromStreamMipMap( level, Strm );
  FINALLY
    Strm.Free;
  END;
end;

{$IFDEF SUPPORT_JPEG}
function TTexture.LoadFromJpeg(J: PJpeg; SwapRB: Boolean = TRUE): Boolean;
begin
  Result := LoadFromBitmap( J.Bitmap, SwapRB );
end;

function TTexture.LoadFromJpegMipMapLevel(level: Integer; J: PJpeg;
  SwapRB: Boolean): Boolean;
begin
  Result := LoadFromBitmapMipMapLevel( level, J.Bitmap, SwapRB );
end;
{$ENDIF SUPPORT_JPEG}

{$IFDEF SUPPORT_PNG}
function TTexture.LoadFromPng(P: PPngObject; SwapRB: Boolean): Boolean;
var B: PBitmap;
begin
  B := NewDibBitmap( P.Width, P.Height, pf32bit );
  P.Draw( B.Canvas.Handle, 0, 0 );
  Result := LoadFromBitmap( B, SwapRB );
  B.Free;
end;

function TTexture.LoadFromPngMipMapLevel(level: Integer; P: PPngObject;
  SwapRB: Boolean): Boolean;
var B: PBitmap;
begin
  B := NewDibBitmap( P.Width, P.Height, pf32bit );
  P.Draw( B.Canvas.Handle, 0, 0 );
  Result := LoadFromBitmapMipMapLevel( level, B, SwapRB );
  B.Free;
end;
{$ENDIF SUPPORT_PNG}

function TTexture.LoadFromResourceID(id: Integer): Boolean;
var Strm: PStream;
begin
  Result := FALSE;
  Strm := NewMemoryStream;
  TRY
    if FindResource( hInstance, PChar( id ), RT_BITMAP ) <> 0 then
      Resource2Stream( Strm, hInstance, PChar( id ), RT_BITMAP )
    else
    if FindResource( hInstance, PChar( id ), RT_RCDATA ) <> 0 then
      Resource2Stream( Strm, hInstance, PChar( id ), RT_RCDATA )
    else Exit;
    Strm.Position := 0;
    Result := LoadFromStream( Strm );
  FINALLY
    Strm.Free;
  END;
end;

function TTexture.LoadFromResourceName(const rsrcname: String): Boolean;
var Strm: PStream;
begin
  Result := FALSE;
  Strm := NewMemoryStream;
  TRY
    if FindResource( hInstance, PChar( rsrcname ), RT_BITMAP ) <> 0 then
      Resource2Stream( Strm, hInstance, PChar( rsrcname ), RT_BITMAP )
    else
    if FindResource( hInstance, PChar( rsrcname ), RT_RCDATA ) <> 0 then
      Resource2Stream( Strm, hInstance, PChar( rsrcname ), RT_RCDATA )
    else Exit;
    Strm.Position := 0;
    Result := LoadFromStream( Strm );
  FINALLY
    Strm.Free;
  END;
end;

function TTexture.LoadFromResourceNameMipMap(level: Integer;
  const rsrcname: String): Boolean;
var Strm: PStream;
begin
  Result := FALSE;
  Strm := NewMemoryStream;
  TRY
    if FindResource( hInstance, PChar( rsrcname ), RT_BITMAP ) <> 0 then
      Resource2Stream( Strm, hInstance, PChar( rsrcname ), RT_BITMAP )
    else
    if FindResource( hInstance, PChar( rsrcname ), RT_RCDATA ) <> 0 then
      Resource2Stream( Strm, hInstance, PChar( rsrcname ), RT_RCDATA )
    else Exit;
    Strm.Position := 0;
    Result := LoadFromStreamMipMap( level, Strm );
  FINALLY
    Strm.Free;
  END;
end;

function TTexture.LoadFromStream(Strm: PStream; SwapRB: Boolean = TRUE): Boolean;
var Pos: DWORD;
    Ch: Char;
    B: PBitmap;
    {$IFDEF SUPPORT_JPEG}
    J: PJpeg;
    {$ENDIF}
    {$IFDEF SUPPORT_PNG}
    P: PPngObject;
    {$ENDIF}
begin
  Pos := Strm.Position;
  Ch := ' ';
  Strm.Read( Ch, 1 );
  Strm.Position := Pos;
  {$IFDEF SUPPORT_PNG}
  if Ch = #$89 then
  begin
    P := NewPngObject;
    TRY
      P.LoadFromStream( Strm );
      B := NewDibBitmap( P.Width, P.Height, pf32bit );
      TRY
        P.Draw( B.Canvas.Handle, 0, 0 );
        Result := LoadFromBitmap( B, SwapRB );
      FINALLY
        B.Free;
      END;
    FINALLY
      P.Free;
    END;
  end
    else
  {$ENDIF}
  {$IFDEF SUPPORT_JPEG}
  if Ch = 'B' then
  {$ENDIF}
  begin
    B := NewBitmap( 0, 0 );
    TRY
      B.LoadFromStream( Strm );
      Result := LoadFromBitmap( B, SwapRB );
    FINALLY
      B.Free;
    END;
  end
  {$IFDEF SUPPORT_JPEG}
    else
  begin
    J := NewJpeg;
    TRY
      J.LoadFromStream( Strm );
      Result := LoadFromJpeg( J );
    FINALLY
      J.Free;
    END;
  end
  {$ENDIF}
  ;
end;

function TTexture.LoadFromStreamMipMap(level: Integer; Strm: PStream;
  SwapRB: Boolean): Boolean;
var Pos: DWORD;
    Ch: Char;
    B: PBitmap;
    {$IFDEF SUPPORT_JPEG}
    J: PJpeg;
    {$ENDIF}
    {$IFDEF SUPPORT_PNG}
    P: PPngObject;
    {$ENDIF}
begin
  Pos := Strm.Position;
  Ch := ' ';
  Strm.Read( Ch, 1 );
  Strm.Position := Pos;
  {$IFDEF SUPPORT_PNG}
  if Ch = #$89 then
  begin
    P := NewPngObject;
    TRY
      P.LoadFromStream( Strm );
      B := NewDibBitmap( P.Width, P.Height, pf32bit );
      TRY
        P.Draw( B.Canvas.Handle, 0, 0 );
        Result := LoadFromBitmapMipMapLevel( level, B, SwapRB );
      FINALLY
        B.Free;
      END;
    FINALLY
      P.Free;
    END;
  end
    else
  {$ENDIF}
  {$IFDEF SUPPORT_JPEG}
  if Ch = 'B' then
  {$ENDIF}
  begin
    B := NewBitmap( 0, 0 );
    TRY
      B.LoadFromStream( Strm );
      Result := LoadFromBitmapMipMapLevel( level, B, SwapRB );
    FINALLY
      B.Free;
    END;
  end
  {$IFDEF SUPPORT_JPEG}
    else
  begin
    J := NewJpeg;
    TRY
      J.LoadFromStream( Strm );
      Result := LoadFromJpegMipMapLevel( level, J );
    FINALLY
      J.Free;
    END;
  end
  {$ENDIF}
  ;
end;

function TTexture.LoadMipMapLevel(level: Integer;
  const rsrcname: String): Boolean;
begin
  if TextureFirstLoadFromResourve then
  begin
    Result := LoadFromResourceNameMipMap( level, rsrcname );
    if Result then Exit;
    Result := LoadFromFileMipMap( level, rsrcname );
  end
    else
  begin
    Result := LoadFromFileMipMap( level, rsrcname );
    if Result then Exit;
    Result := LoadFromResourceNameMipMap( level, rsrcname );
  end;
end;

procedure TTexture.MoveToPool;
begin
  if Binded then UnBind;
  CurrentContext.fTexPool.Add( @ Self );
end;

procedure TTexture.SetBorderColor(const Value: TColor4);
begin
  FBorderColor := Value;
  if not Binded then Exit;
  {$IFDEF GLCONTEXT_MULTITHREADS}
  if CompareMem( @ CurrentContext.FTextureBorderColor,
                 @ fBorderColor, Sizeof( fBorderColor ) ) then Exit;
  {$ELSE}
  if CompareMem( @ fContext.FTextureBorderColor,
                 @ fBorderColor, Sizeof( fBorderColor ) ) then Exit;
  {$ENDIF}
  glTexEnvfv( GL_TEXTURE_ENV, GL_TEXTURE_BORDER_COLOR, @ fBorderColor );
end;

procedure TTexture.SetColor(const Value: TColor4);
begin
  FColor := Value;
  if not Binded then Exit;
  {$IFDEF GLCONTEXT_MULTITHREADS}
  if CompareMem( @ CurrentContext.FTextureColor, @ fColor, Sizeof( fColor ) ) then
    Exit;
  {$ELSE}
  if CompareMem( @ fContext.FTextureColor, @ fColor, Sizeof( fColor ) ) then
    Exit;
  {$ENDIF}
  glTexEnvfv( GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, @ fColor );
end;

procedure TTexture.SetEnvironment(const Value: TTextureEnvironment);
const TextureEnvironmentAsDWORD: array[ TTextureEnvironment ] of DWORD =
      ( {teModulate} GL_MODULATE,
        {teReplace}  GL_REPLACE,
        {teDecal}    GL_DECAL,
        {teBlend}    GL_BLEND );
begin
  FEnvironment := Value;
  if not Binded then Exit;
  {$IFDEF GLCONTEXT_MULTITHREADS}
  if CurrentContext.FTextureEnvironment = Value then Exit;
  CurrentContext.fTextureEnvironment := Value;
  {$ELSE}
  if fContext.FTextureEnvironment = Value then Exit;
  fContext.fTextureEnvironment := Value;
  {$ENDIF}
  glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE,
             TextureEnvironmentAsDWORD[ Value ] );
  Assert( CheckErrors );
end;

procedure TTexture.SetMagFilter(const Value: TTextureMagFilter);
const TextureMagFilterAsDWORD: array[ TTextureMagFilter ] of DWORD =
      ( {tmagNearest}  GL_NEAREST,
        {tmagLinear}   GL_LINEAR );
begin
  FMagFilter := Value;
  if not Binded then Exit;
  {$IFDEF GLCONTEXT_MULTITHREADS}
  if CurrentContext.FTextureMagFilter = Value then Exit;
  {$ELSE}
  if fContext.FTextureMagFilter = Value then Exit;
  {$ENDIF}
  glTexParameteri( TextureDimensionAsDWORD[ Dimension ], GL_TEXTURE_MAG_FILTER,
    TextureMagFilterAsDWORD[ Value ] );
end;

procedure TTexture.SetMinFilter(const Value: TTextureMinFilter);
const TextureMinFilterAsDWORD: array[ TTextureMinFilter ] of DWORD =
      ( {tminNearest}                 GL_NEAREST,
        {tminLinear}                  GL_LINEAR,
        {tminNearestMipmapNearest}    GL_NEAREST_MIPMAP_NEAREST,
        {tminLinearMipmapNearest}     GL_LINEAR_MIPMAP_NEAREST,
        {tminNearestMipmapLinear}     GL_NEAREST_MIPMAP_LINEAR,
        {tminLinearMipmapLinear}      GL_LINEAR_MIPMAP_LINEAR );
begin
  FMinFilter := Value;
  if not Binded then Exit;
  {$IFDEF GLCONTEXT_MULTITHREADS}
  if CurrentContext.FTextureMinFilter = Value then Exit;
  {$ELSE}
  if fContext.FTextureMinFilter = Value then Exit;
  {$ENDIF}
  glTexParameteri( TextureDimensionAsDWORD[ Dimension ], GL_TEXTURE_MIN_FILTER,
    TextureMinFilterAsDWORD[ Value ] );
end;

procedure TTexture.SetMipMap(const Value: Boolean);
begin
    FMipMap := Value;
end;

procedure TTexture.SetWrapS(const Value: TTextureWrap);
begin
  FWrapS := Value;
  if not Binded then Exit;
  {$IFDEF GLCONTEXT_MULTITHREADS}
  if CurrentContext.FTextureWrapS = Value then Exit;
  {$ELSE}
  if fContext.FTextureWrapS = Value then Exit;
  {$ENDIF}
  glTexParameteri( TextureDimensionAsDWORD[ Dimension ], GL_TEXTURE_WRAP_S,
    TextureWrapAsDWORD[ Value ] );
end;

procedure TTexture.SetWrapT(const Value: TTextureWrap);
begin
  FWrapT := Value;
  if not Binded then Exit;
  {$IFDEF GLCONTEXT_MULTITHREADS}
  if CurrentContext.FTextureWrapT = Value then Exit;
  {$ELSE}
  if fContext.FTextureWrapT = Value then Exit;
  {$ENDIF}
  glTexParameteri( TextureDimensionAsDWORD[ Dimension ], GL_TEXTURE_WRAP_T,
    TextureWrapAsDWORD[ Value ] );
end;

procedure TTexture.UnBind;
begin
  if not Binded then Exit;
  {$IFDEF GLCONTEXT_MULTITHREADS}
  if CurrentContext.FTexture = @ Self then
    CurrentContext.FTexture := nil;
  {$ELSE}
  fContext.MakeCurrent;
  if fContext.FTexture = @ Self then
    fContext.FTexture := nil;
  {$ENDIF}
  glBindTexture( TextureDimensionAsDWORD[ Dimension ], 0 );
  glDisable( TextureDimensionAsDWORD[ Dimension ] );
  Assert( CheckErrors );
end;

{ TFontController }

procedure TFontController.Assign(SrcFont: PFontController);
begin
  Face    := SrcFont.Face;
  Height  := SrcFont.Height;
  Width   := SrcFont.Width;
  Style   := SrcFont.Style;
end;

destructor TFontController.DESTROY;
begin
  FFace := '';
  {$IFDEF FAST_GLFONT}
  fFontTexture.Free;
  {$ENDIF}
  inherited;
end;

procedure TFontController.INIT;
begin
  FFace := 'Arial';
  FHeight := 16;
  FStyle := [ ffsAntialiased ];
end;

function TFontController.Match(F: PFontController): Boolean;
begin
  Result := (Face = F.face) and (Height = F.Height) and
       (Width = F.Width) and (Style = F.Style)
       {$IFDEF FAST_GLFONT}
       and CompareMem( @fContext.fColor, @ fColor, Sizeof( TColor4 ))
       {$ENDIF};
end;

{$IFDEF FAST_GLFONT}
procedure TFontController.NeedFastFont(Font, Font2: HFont);
var B: PBitmap;
    s: String;
    i, L: Integer;
    R: TRect;
    OldFont: HFont;
    Sz, Sz1: TSize;
    fCharRects: array[ ' '..#255 ] of TRect;
    R_Left, R_Top, R_Right, R_Bottom: Double;
    dX0: Double;

    procedure PaintColoredAlphed;
    var y, x: Integer;
        Dst: PRGBQuad;
        C: TRGBQuad;
    begin
      C.rgbBlue := Round( fColor.b * 255 );
      C.rgbGreen := Round( fColor.g * 255 );
      C.rgbRed := Round( fColor.r * 255 );
      C.rgbReserved := 255;
      for y := 0 to B.Height-1 do
      begin
        Dst := B.ScanLine[ y ];
        for x := 0 to B.Width-1 do
        begin
          if Dst.rgbBlue <> 0 then
            Dst^ := C;
          inc( Dst );
        end;
      end;
    end;
begin
  //----------     
  SetLength( s, 256-32 );
  for i := 32 to 255 do
    s[ i-31 ] := Char( i );
  B := NewBitmap( 1, 1 );
  OldFont := Windows.SelectObject( B.Canvas.Handle, Font2 );
  TRY
    R := MakeRect( 0, 0, 0, fHeight );
    Sz.cx := 0;
    Sz.cy := 0;
    for L := 1 to Length( s ) do //  
    begin
      Windows.GetTextExtentPoint32( B.Canvas.Handle, PChar( @ s[ L ] ), 1, Sz1 );
      if (Sz.cx + Sz1.cx > 512) and (Sz.cy > 0) then
      begin
        Sz.cx := 0;
        inc( R.Top, Sz.cy ); 
        Sz.cy := 0;
      end;
      fCharRects[ s[ L ] ] := MakeRect( Sz.cx, R.Top,
        Sz.cx + Sz1.cx, R.Top + Sz1.cy );
      if Sz1.cy > Sz.cy then Sz.cy := Sz1.cy;
      R.Bottom := R.Top + Sz.cy;
      Sz.cx := Sz.cx + Sz1.cx;
      if R.Right < Sz.cx then
        R.Right := min( fContext.MaxTextureSize div 2, NextPower2( Sz.cx ) );
    end;
  FINALLY
    Windows.SelectObject( B.Canvas.Handle, OldFont );
    B.Free;
  END;
  Sz.cx := R.Right;
  Sz.cy := NextPower2( R.Bottom );
  B := NewDibBitmap( Sz.cx, Sz.cy, pf32bit );
  OldFont := Windows.SelectObject( B.Canvas.Handle, Font2 );
  Windows.SetBkColor( B.Canvas.Handle, 0 );
  Windows.SetTextColor( B.Canvas.Handle, $FFFFFF );
  TRY
    R := MakeRect( 0, 0, 512, fHeight );
    for L := 1 to Length( s ) do //  
    begin
      R := fCharRects[ s[ L ] ];
      Windows.ExtTextOut( B.Canvas.Handle, R.Left, R.Top, ETO_CLIPPED or
        ETO_OPAQUE, @ R, PChar( @ s[ L ] ), 1, nil );
    end;
    Windows.SelectObject( B.Canvas.Handle, OldFont );
    //---   ---
    fFontTexture := fContext.NewTexture( B.Width, B.Height );
    PaintColoredAlphed;
    fFontTexture.CalculateAlpha := caAsIs;
    fFontTexture.LoadFromBitmap( B );
  FINALLY
    //B.SaveToFile( GetStartDir + 'font.bmp' );
    B.Free;
  END;
  //------------       ' '..#255

  dX0 := 0;
  {$IFDEF ITALIC_SMOOTH}
  if ffsItalic in Style then
    dX0 := fHeight * 0.3;
  {$ENDIF}
  B := NewBitmap( 1, 1 );
  OldFont := Windows.SelectObject( B.Canvas.Handle, Font );
  for L := 1 to Length( S ) do
  begin
    R := fCharRects[ S[ L ] ];
    R_Left := R.Left / (Sz.cx);
    R_Right := R.Right / (Sz.cx);
    R_Top := R.Bottom / (-Sz.cy);
    R_Bottom := R.Top / (-Sz.cy);
    glNewList( fFontListID + 31 + L, GL_COMPILE );
      glBegin( GL_QUADS );

        Windows.GetTextExtentPoint32( B.Canvas.Handle, PChar( @ s[ L ] ), 1, Sz1 );
        R := MakeRect( 0,0,Sz1.cx,Sz1.cy );

        glTexCoord2f( R_Left, R_Top );
        glVertex2f( 0, 0 );

        glTexCoord2f( R_Right, R_Top );
        glVertex2f( R.Right - R.Left, 0 );

        glTexCoord2f( R_Right, R_Bottom );
        glVertex2f( dX0+R.Right - R.Left, (R.Bottom - R.Top) );

        glTexCoord2f( R_Left, R_Bottom );
        glVertex2f( dX0, (R.Bottom - R.Top) );
      glEnd;
      glTranslatef( R.Right - R.Left, 0, 0 );
    glEndList;
  end;
  Windows.SelectObject( B.Canvas.Handle, OldFont );
  B.Free;
end;
{$ENDIF}

procedure TFontController.NeedFont;
var TmpFont, TmpFont2 {$IFNDEF FAST_GLFONT}, OldFont{$ENDIF}: HFont;
    LogFont: TLogFont;
    {$IFDEF FAST_GLFONT}
    LogFont2: TLogFont;
    {$ENDIF}
begin
  if fFontListID <> 0 then Exit;
  fFontListID := glGenLists( 256 );
  FillChar( LogFont, Sizeof( LogFont ), 0 );
  LogFont.lfHeight := fHeight;
  LogFont.lfWidth := fWidth;
  if ffsBold in Style then LogFont.lfWeight := 700
                      else LogFont.lfWeight := 400;
  {$IFDEF ITALIC_SMOOTH and FAST_GLFONT}
  LogFont.lfItalic := 0;
  {$ELSE}
  LogFont.lfItalic := Integer( ffsItalic in Style );
  {$ENDIF}
  LogFont.lfOutPrecision := OUT_TT_ONLY_PRECIS;
  if ffsAntialiased in Style then
       LogFont.lfQuality := ANTIALIASED_QUALITY
  else LogFont.lfQuality := DRAFT_QUALITY;
  StrLCopy( LogFont.lfFaceName, PChar( Face ), LF_FACESIZE );
  TmpFont := CreateFontIndirect( LogFont );
  {$IFDEF FAST_GLFONT}
  LogFont2 := LogFont;
    {$IFDEF ITALIC_SMOOTH}
    if ffsItalic in Style then
    begin
      LogFont2.lfHeight := LogFont2.lfHeight * 3;
      LogFont2.lfWidth := LogFont2.lfWidth * 3;
    end;
    {$ENDIF}
    if LogFont2.lfHeight < MinFontHeight then
    begin
      if LogFont2.lfWidth <> 0 then
        LogFont2.lfWidth := LogFont2.lfWidth * MinFontHeight div LogFont2.lfHeight;
      LogFont2.lfHeight := MinFontHeight;
    end
      else
    if LogFont2.lfHeight > MaxFontHeight then
    begin
      if LogFont2.lfWidth <> 0 then
        LogFont2.lfWidth := LogFont2.lfWidth * MaxFontHeight div LogFont2.lfHeight;
      LogFont2.lfHeight := MaxFontHeight;
    end;
  TmpFont2 := CreateFontIndirect( LogFont2 );
  {$ELSE}
  TmpFont2 := TmpFont;
  {$ENDIF}
  {$IFDEF FAST_GLFONT}
  NeedFastFont( TmpFont, TmpFont2 );
  DeleteObject( TmpFont2 );
  {$ELSE}
  OldFont := SelectObject( fContext.FCtxDC, TmpFont );
  wglUseFontBitmapsA( fContext.FCtxDC, 0, 256, fFontListID );
  SelectObject( fContext.FCtxDC, OldFont );
  if TmpFont2 <> TmpFont then;
    DeleteObject( TmpFont2 );
  {$ENDIF}
  DeleteObject( TmpFont );
  Assert( CheckErrors );
end;

procedure TFontController.SelectFont;
var i: Integer;
    F: PFontController;
begin
  for i := 0 to fContext.fFontList.Count-1 do
  begin
    F := fContext.fFontList.Items[ i ];
    if F.Match( @ Self ) then
    begin
      fContext.fFontList.MoveItem( i, 0 );
      fContext.FSelectedFont := F;
      Exit;
    end;
  end;
  //    , :
  new( F, Create );
  F.fContext := fContext;
  {$IFDEF FAST_GLFONT}
  F.fColor := fContext.fColor;
  {$ENDIF}
  fContext.fFontList.Insert( 0, F );
  F.Face := Face;
  F.Height := Height;
  F.Width := Width;
  F.Style := Style;
  F.NeedFont;
  fContext.FSelectedFont := F;
  //   ,    :
  if fContext.fFontList.Count > MaxFontCount then
  begin
    F := fContext.fFontList.Last;
    fContext.fFontList.Delete( fContext.fFontList.Count-1 );
    F.Free;
  end;
end;

{ TDisplayList }

procedure TDisplayList.BeginList;
begin
  Assert( fContext = CurrentContext, 'Display list owned by another context' );
  //todo:  ShareLists  
  glNewList( fDisplayList, GL_COMPILE );
end;

procedure TDisplayList.CallList;
begin
  glCallList( fDisplayList );
  Assert( CheckErrors );
end;

destructor TDisplayList.DESTROY;
begin
  if fDisplayList <> 0 then
    glDeleteLists( fDisplayList, 1 );
  Assert( CheckErrors );
  inherited;
end;

procedure TDisplayList.EndList;
begin
  glEndList;
  Assert( CheckErrors );
end;

procedure TDisplayList.INIT;
begin
  fDisplayList := glGenLists( 1 );
end;

{ TArray }

procedure TArray.Change;
begin
  CASE fArrType OF
  atVertexes:   fContext.fCurVrtxArray := nil;
  atTexCoords:  fContext.fCurTexCArray[ 0 ] := nil;
  atNormals:    fContext.fCurNormArray := nil;
  atColors:     fContext.fCurColrArray := nil;
  atEdgeFlags:  fContext.fCurEdgeArray := nil;
  END;
end;

function TArray.DataSize: Integer;
begin
  if fArrType = atEdgeFlags then
                Result := 1
  else
  CASE fEltType OF
  etShort:      Result := 2;
  etInt:        Result := 4;
  etSingle:     Result := 4;
  etDouble:     Result := 8;
  else          Result := 8;
  END;
  Result := Result * NCoords * Count;
end;

destructor TArray.DESTROY;
begin
  if not fShare then
    FreeMem( fData );
  inherited;
end;

function TArray.GetItemPtr(Idx: Integer): Pointer;
begin
  Result := Pointer( Integer( fData ) + Idx * fEltSize );
end;

function TArray.GetItems(idx: Integer): TPoint4D;
type PShort = ^Short;
var P: Pointer;
    P_Short: PShort;
    P_Int: PInt;
    P_Single: PSingle;
    P_Double: PDouble;
begin
  Assert( not fCombined, 'Combined array items must not be accesses via Items[] ' +
    'property' );
  P := Pointer( Integer( fData ) + fEltSize * idx );
  CASE fEltType OF
  etShort: begin
             P_Short := P;
             Result.X := P_Short^; inc( P_Short );
             Result.Y := P_Short^;
             if NCoords > 2 then
             begin
               inc( P_Short );
               Result.Z := P_Short^;
               if NCoords > 3 then
               begin
                 inc( P_Short );
                 Result.W := P_Short^;
               end;
             end;
           end;
  etInt:   begin
             P_Int := P;
             Result.X := P_Int^; inc( P_Int );
             Result.Y := P_Int^;
             if NCoords > 2 then
             begin
               inc( P_Int );
               Result.Z := P_Int^;
               if NCoords > 3 then
               begin
                 inc( P_Int );
                 Result.W := P_Int^;
               end;
             end;
           end;
  etSingle:begin
             P_Single := P;
             Result.X := P_Single^; inc( P_Single );
             Result.Y := P_Single^;
             if NCoords > 2 then
             begin
               inc( P_Single );
               Result.Z := P_Single^;
               if NCoords > 3 then
               begin
                 inc( P_Single );
                 Result.W := P_Single^;
               end;
             end;
           end;
  etDouble:begin
             P_Double := P;
             Result.X := P_Double^; inc( P_Double );
             Result.Y := P_Double^;
             if NCoords > 2 then
             begin
               inc( P_Double );
               Result.Z := P_Double^;
               if NCoords > 3 then
               begin
                 inc( P_Double );
                 Result.W := P_Double^;
               end;
             end;
           end;
  END;
end;

procedure TArray.initEltSize;
begin
  if Count > 0 then
    fEltSize := DataSize div Count;
end;

procedure TArray.SetCapacity(const Value: Integer);
var NewData: Pointer;
begin
  if Value > FCapacity then
  begin
    GetMem( NewData, Value * fEltSize );
    if fData <> nil then
    begin
      Move( fData^, NewData^, fCount * fEltSize );
      FreeMem( fData );
    end;
    fData := NewData;
    FCapacity := Value;
  end;
end;

procedure TArray.SetCount(const Value: Integer);
begin
  if fCount > FCapacity then
    FCapacity := fCount;
  if Value > Capacity then
    Capacity := Value;
  fCount := Value;
  if fEltSize = 0 then
    initEltSize;
end;

procedure TArray.SetCurrent( Ctx: PContext );
begin
  if @ Self = nil then Exit;
  Assert( Ctx.IsCurrent, 'Context must be current to set current array' );
  CASE fArrType OF
  atVertexes:   begin
                  if Ctx.fCurVrtxArray = @ Self then Exit;
                  Ctx.fCurVrtxArray := @ Self;
                  glVertexPointer( NCoords, ElementTypeAsDWORD[ fEltType ],
                    0, fData );
                end;
  atTexCoords:  begin
                  if Ctx.fCurTexCArray[ 0 ] = @ Self then Exit;
                  Ctx.fCurTexCArray[ 0 ] := @ Self;
                  glTexCoordPointer( NCoords, ElementTypeAsDWORD[ fEltType ],
                    0, fData );
                end;
  atNormals:    begin
                  if Ctx.fCurNormArray = @ Self then Exit;
                  Ctx.fCurNormArray := @ Self;
                  glNormalPointer( ElementTypeAsDWORD[ fEltType ], 0, fData );
                end;
  atColors:     begin
                  if Ctx.fCurColrArray = @ Self then Exit;
                  Ctx.fCurColrArray := @ Self;
                  glColorPointer( NCoords, ColorEltTypeAsDWORD[
                    PColorArray( @Self ).fColorEltType ], 0, fData );
                end;
  atEdgeFlags:  begin
                  if Ctx.fCurEdgeArray = @ Self then Exit;
                  Ctx.fCurEdgeArray := @ Self;
                  glEdgeFlagPointer( 0, fData );
                end;
  END;
end;

procedure TArray.SetItems(idx: Integer; const Value: TPoint4D);
type PShort = ^Short;
var P: Pointer;
    P_Short: PShort;
    P_Int: PInt;
    P_Single: PSingle;
    P_Double: PDouble;
begin
  Assert( not fCombined, 'Combined array items must not be changed via Items[] ' +
    'property' );
  P := Pointer( Integer( fData ) + fEltSize * idx );
  CASE fEltType OF
  etShort: begin
             P_Short := P;
             P_Short^ := Round( Value.X );
             if NCoords > 1 then
             begin
               inc( P_Short );
               P_Short^ := Round( Value.Y );
               if NCoords > 2 then
               begin
                 inc( P_Short );
                 P_Short^ := Round( Value.Z );
                 if NCoords > 3 then
                 begin
                   inc( P_Short );
                   P_Short^ := Round( Value.W );
                 end;
               end;
             end;
           end;
  etInt:   begin
             P_Int := P;
             P_Int^ := Round( Value.X );
             if NCoords > 1 then
             begin
               inc( P_Int );
               P_Int^ := Round( Value.Y );
               if NCoords > 2 then
               begin
                 inc( P_Int );
                 P_Int^ := Round( Value.Z );
                 if NCoords > 3 then
                 begin
                   inc( P_Int );
                   P_Int^ := Round( Value.W );
                 end;
               end;
             end;
           end;
  etSingle:begin
             P_Single := P;
             P_Single^ := Value.X;
             if NCoords > 1 then
             begin
               inc( P_Single );
               P_Single^ := Value.Y;
               if NCoords > 2 then
               begin
                 inc( P_Single );
                 P_Single^ := Value.Z;
                 if NCoords > 3 then
                 begin
                   inc( P_Single );
                   P_Single^ := Value.W;
                 end;
               end;
             end;
           end;
  etDouble:begin
             P_Double := P;
             P_Double^ := Value.X;
             if NCoords > 1 then
             begin
               inc( P_Double );
               P_Double^ := Value.Y;
               if NCoords > 2 then
               begin
                 inc( P_Double );
                 P_Double^ := Value.Z;
                 if NCoords > 3 then
                 begin
                   inc( P_Double );
                   P_Double^ := Value.W;
                 end;
               end;
             end;
           end;
  END;
  Change;
end;

{ TColorArray }

function TColorArray.DataSize: Integer;
begin
  CASE fColorEltType OF
  cetByte, cetUByte:          Result := 1;
  cetShort, cetUShort:        Result := 2;
  cetInt, cetUInt, cetSingle: Result := 4;
  cetDouble:                  Result := 8;
  else Result := 1;
  END;
  Result := Result * NCoords * Count;
end;

{ TCombinedArray }

function TCombinedArray.DataSize: Integer;
begin
  Result := 0;
  if atVertexes  in fArraysPresent then inc( Result, fVertexNCoords );
  if atTexCoords in fArraysPresent then inc( Result, fTexturesNCoords );
  if atTex1Coords in fArraysPresent then inc( Result, fTexturesNCoords );
  if atTex2Coords in fArraysPresent then inc( Result, fTexturesNCoords );
  if atTex3Coords in fArraysPresent then inc( Result, fTexturesNCoords );
  if atTex4Coords in fArraysPresent then inc( Result, fTexturesNCoords );
  if atNormals   in fArraysPresent then inc( Result, 3 );
  if atColors    in fArraysPresent then inc( Result, fColorNcomponents );
  if atEdgeFlags in fArraysPresent then inc( Result );
  Result := Result * Sizeof( Single ) * Count;
end;

procedure TCombinedArray.SetActiveArrays(const Value: TCombinedElements);
begin
  {$IFDEF CHECKING_STATE}
  if FActiveArrays = Value then Exit;
  {$ENDIF}
  FActiveArrays := Value * fArraysPresent;
end;

procedure TCombinedArray.SetCurrent( Ctx: PContext );
var P: Pointer;
    i: Integer;
    f: TArrayType;
begin
  Assert( Ctx.IsCurrent, 'Context must be current to set current array' );
  P := fData;
  if atVertexes in fArraysPresent then
  begin
    {$IFDEF CHECKING_STATE}
    if Ctx.fCurVrtxArray <> @ Self then
    {$ENDIF}
    begin
      Ctx.fCurVrtxArray := @ Self;
      glVertexPointer( fVertexNCoords, GL_FLOAT, fEltSize, P );
      P := Pointer( Integer( P ) + fVertexNCoords * Sizeof( Single ) );
    end;
  end;
  if atNormals in fArraysPresent then
  begin
    {$IFDEF CHECKING_STATE}
    if Ctx.fCurNormArray <> @ Self then
    {$ENDIF}
    begin
      Ctx.fCurNormArray := @ Self;
      glNormalPointer( GL_FLOAT, fEltSize, P );
      P := Pointer( Integer( P ) + 3 * Sizeof( Single ) );
    end;
  end;
  if atColors in fArraysPresent then
  begin
    {$IFDEF CHECKING_STATE}
    if Ctx.fCurColrArray <> @ Self then
    {$ENDIF}
    begin
      Ctx.fCurColrArray := @ Self;
      glColorPointer( fColorNcomponents, GL_FLOAT, fEltSize, P );
      P := Pointer( Integer( P ) + fColorNcomponents * Sizeof( Single ) );
    end;
  end;
  if atTexCoords in fArraysPresent then
  begin
    {$IFDEF CHECKING_STATE}
    if Ctx.fCurTexCArray <> @ Self then
    {$ENDIF}
    begin
      Ctx.fCurTexCArray[ 0 ] := @ Self;
      {$IFDEF MULTITEXTURING}
      if atTexCoords in FActiveArrays then
      {$ENDIF}
      glTexCoordPointer( fTexturesNCoords, GL_FLOAT, fEltSize, P );
      P := Pointer( Integer( P ) + fTexturesNCoords * Sizeof( Single ) );
    end;
  end;
  {$IFDEF MULTITEXTURING}
  for i := 1 to 4 do
  begin
    CASE i OF
    1: f := atTex1Coords;
    2: f := atTex2Coords;
    3: f := atTex3Coords;
    4: f := atTex4Coords;
    else f := atTexCoords;
    END;
    if f in fArraysPresent then
    begin
      {$IFDEF CHECKING_STATE}
      if Ctx.fCurTexCArray[ i ] <> @ Self then
      {$ENDIF}
      begin
        Ctx.fCurTexCArray[ i ] := @ Self;
        if Ctx.AutoActivateTexture then
          glClientActiveTextureARB( GL_TEXTURE0_ARB + i );
        {$IFDEF MULTITEXTURING}
        if f in FActiveArrays then
        {$ENDIF}
        glTexCoordPointer( fTexturesNCoords, GL_FLOAT, fEltSize, P );
        P := Pointer( Integer( P ) + fTexturesNCoords * Sizeof( Single ) );
        if Ctx.AutoActivateTexture then
          glClientActiveTextureARB( GL_TEXTURE0_ARB );
      end;
    end;
  end;
  {$ENDIF}
  if atEdgeFlags in fArraysPresent then
  begin
    {$IFDEF CHECKING_STATE}
    if Ctx.fCurEdgeArray <> @ Self then
    {$ENDIF}
    begin
      Ctx.fCurEdgeArray := @ Self;
      glEdgeFlagPointer( fEltSize, P );
    end;
  end;
  Assert( CheckErrors );
end;

{ TQuadric }

destructor TQuadric.DESTROY;
begin
  gluDeleteQuadric( fQuadric );
  Assert( CheckErrors );
  inherited;
end;

procedure TQuadric.DrawCylinder(baseradius, topradius, height: Double;
  slices, stacks: Integer);
begin
  gluCylinder( fQuadric, baseradius, topradius, height, slices, stacks );
  Assert( CheckErrors );
end;

procedure TQuadric.DrawDisk(innerradius, outerradius: Double; slices,
  loops: Integer);
begin
  gluDisk( fQuadric, innerradius, outerradius, slices, loops );
  Assert( CheckErrors );
end;

procedure TQuadric.DrawPartialDisk(innerradius, outerradius: Double;
  slices, loops: Integer; startAngle, endAngle: Double);
begin
  gluPartialDisk( fQuadric, innerradius, outerradius, slices, loops,
    startAngle, endAngle );
  Assert( CheckErrors );
end;

procedure TQuadric.DrawSphere(radius: Double; slices, stacks: Integer);
begin
  gluSphere( fQuadric, radius, slices, stacks );
  Assert( CheckErrors );
end;

procedure TQuadric.INIT;
begin
  fQuadric := gluNewQuadric
end;

procedure TQuadric.SetDrawStyle(const Value: TQuadricDrawStyle);
const QuadricDrawStyleAsDWORD: array[ TQuadricDrawStyle ] of DWORD =
      ( {qdsFill}        GLU_FILL,
        {qdsLine}        GLU_LINE,
        {qdsSilhouette}  GLU_SILHOUETTE,
        {qdsPoint}       GLU_POINT );
begin
  {$IFDEF CHECKING_STATE}
  if FDrawStyle = Value then Exit;
  {$ENDIF}
  FDrawStyle := Value;
  gluQuadricDrawStyle( fQuadric, QuadricDrawStyleAsDWORD[ DrawStyle ] );
end;

procedure TQuadric.SetNormalsType(const Value: TQuadricNormals);
const QuadricNormalsAsDWORD: array[ TQuadricNormals ] of DWORD =
      ( {qnSmooth}      GLU_SMOOTH,
        {qnFlat}        GLU_FLAT,
        {qnNone}        GLU_NONE );
begin
  {$IFDEF CHECKING_STATE}
  if FNormalsType = Value then Exit;
  {$ENDIF}
  FNormalsType := Value;
  gluQuadricNormals( fQuadric, QuadricNormalsAsDWORD[ FNormalsType ] );
end;

procedure TQuadric.SetOrientation(const Value: TQuadricOrientation);
const QuadricOrientationAsDWORD: array[ TQuadricOrientation ] of DWORD =
      ( {qoOutside}    GLU_OUTSIDE,
        {qoInside}     GLU_INSIDE );
begin
  {$IFDEF CHECKING_STATE}
  if FOrientation = Value then Exit;
  {$ENDIF}
  FOrientation := Value;
  gluQuadricOrientation( fQuadric, QuadricOrientationAsDWORD[ Orientation ] );
end;

procedure TQuadric.SetTextured(const Value: Boolean);
begin
  {$IFDEF CHECKING_STATE}
  if FTextured = Value then Exit;
  {$ENDIF}
  FTextured := Value;
  gluQuadricTexture( fQuadric, Byte( Textured ) );
end;

initialization
  {$IFDEF GLCONTEXT_MULTITHREADS}
  tc_sema := CreateSemaphore( nil, 1, 1, nil );
  threadContexts := NewList;
  {$ENDIF}

finalization
  {$IFDEF GLCONTEXT_MULTITHREADS}
  CloseHandle( tc_sema );
  threadContexts.Free;
  {$ENDIF}
  CloseOpenGL;

end.
